#!/usr/bin/perl
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# juxtapose_raw_freqs_obs_case_iii.pl:
#	This script inputs RAW OBSERVED Frequences of case-iii gapped segments w/ distinct configurations,
#	and outputs them in a tabular form, where the rows give different #{descendant sites}_fin and the columns give different time-lapses.
#	The script was created by Modifying the output part of 'juxtapose_raw_freqs_obs_case_iv.pl'.
#	[CREATED on 2020/04/12.]
#
# juxtapose_raw_freqs_obs_case_iv.pl:
#	This script inputs RAW OBSERVED Frequences of case-iv gapped segments w/ distinct configurations,
#	and outputs them in a tabular form, where the rows give different "x:y"'s and the columns give different time-lapses.
#	The script was created by DISABLING the function to compute cumulative frequencies
#	in 'compute_dbl_cumulative_frequencies_obs.pl'.
#	[CREATED on 2020/04/11.]
#
# compute_dbl_cumulative_frequencies_obs.pl:
#	This script computes & outputs double-cumulative frequencies
#	from the raw frequencies of the configurations OBSERVED from a set of simulated (or maybe reconstructed) PWAs.
#	The script was created by modifying the basic script, 'compute_dbl_cumulative_frequencies.pl'.
#	[CREATION STARTED on 2020/04/11, ENDED on 2020/04/dd.]
#
# compute_dbl_cumulative_frequencies.pl:
#	This Perl script reads in an input file recording frequencies of configurations,
#	each of which is specified by two non-negative integers,
#	and computes & outputs double-cumulative frequencies.
#	[CREATION STARTED on 2020/04/08, ENDED on 2020/04/dd.]
#


use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	list_sessions_ext
	tfm_labels2indices
	tfm_labels2indices_gen
	);

use MyPWA_Analyses qw(
	readin_2D_array
	cmpt_dbl_cumulative_2Darray
	);


sub create_dir_rw (\$); # IMPORTED from 'lastpiece.alpha.pl' on 2020/04/09. #


$| = 1; # Flush the buffer for standard output as soon as it receives data.

my $program_path = $0;
my ($program_name) = ($program_path =~ m!([^/]+)$!);


	# Controlling the level of test outputs. #
	# (0: silent; 1: very quiet; 2: quiet; 3: medium; 4: verbose; 5: very verbose) 
our $TEST_LEVEL = 3;


		### Reader-Specified Parameters. ###

our @TIME_LAPSES = (0.2, 0.5, 1); # A list of time-lapses to be examined. #

			# (MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' on 2020/04/11.) (GL) #
our $LABEL_CORNER_IN = "\#{residues_in_'anc0001'}\:\#{residues_in_'seq0001'}"; # The label at the top-left corner of the input table.

our @RELV_FLD_LABELS_IN = ('Count(i.e., frequency)');  # A list of labels of relevant fields (except the row-label, i.e., the 0th field). #
our $CT_RLV_FLDS_IN = @RELV_FLD_LABELS_IN;

				# (Modified in creation of 'juxtapose_raw_freqs_obs_case_iii.pl' on 2020/04/12.) (GL) #
##our $LABEL_CORNER_OUT = '#{sites}_Init:#{sites}_Fin\Time_Lapse'; # The label at the top-left corner of the input table. 
our $LABEL_CORNER_OUT = '#{Dec_sites}\Time_Lapse'; # The label at the top-left corner of the input table.
				# End of "(Modified in creation of 'juxtapose_raw_freqs_obs_case_iii.pl' on 2020/04/12.) (GL)" #

our @RELV_FLD_LABELS_OUT = @TIME_LAPSES;
our $CT_RLV_FLDS_OUT = @RELV_FLD_LABELS_OUT;
			# END of "(MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' on 2020/04/11.) (GL)" #


our $FLD_SEP = qr/\t/; # Field separator used for the header. #

our $STR_UNDEF = 'n/a';

##our $MAX_X = 100; # Maximum value for the $x (i.e., the 1st index for the 2D array) to be examined. # THIS IS NO LONGER NECESSARY for the script, 'juxtapose_raw_freqs_obs_case_iii.pl'. #
our $MAX_Y = 100; # Maximum value for the $y (i.e., the 2nd index for the 2D array) to be examined.

=For TESTING the script. (Test1)
  $MAX_X = 10; # Maximum value for the $x (i.e., the 1st index for the 2D array) to be examined.
  $MAX_Y = 10; # Maximum value for the $y (i.e., the 2nd index for the 2D array) to be examined.
=cut # End of "For TESTING the script. (Test1)"


		### Input & Output files. ###

			### RESTART FROM HERE ###

			## MAJOR MODIFICATION in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (I/O-F) ##

my @indir_names = (); # A list of Names of Input Directories. #
foreach my $tlapse (@TIME_LAPSES) { push @indir_names, sprintf ("ancdec_blen%3.1f", $tlapse); }

my $in_topdir  = "../../..";

my $infile_name = 'total_frqs_gseg_configs_bulk.txt';
#my $infile_name = 'total_frqs_gseg_configs_ends.txt';

my @infiles = ();
foreach my $indir_name (@indir_names) { push @infiles, "$in_topdir/$indir_name/$infile_name"; }


##my $topoutdir = "X_le_${MAX_X}_Y_le_$MAX_Y";
my $topoutdir = "Y_le_$MAX_Y"; # MODIFIED in creation of 'juxtapose_raw_freqs_obs_case_iii.pl' (on 2020/04/12). (I/O-F) #

##   $topoutdir = 'TESTs/Test1'; # For TESTING the script (Test1). #


my $outfile_name = $infile_name;

##my $outdir = "$topoutdir/Cum_ObsFreqs";
my $outdir = "$topoutdir/Raw_ObsFreqs";

my $outfile = "$outdir/$outfile_name";

=For the predecessor, 'compute_dbl_cumulative_frequencies.pl'. (I/O-F)

my $infile = '../ancdec_blen0.5/total_frqs_gseg_configs_bulk.txt';

(my $outfile = $infile) =~ s/^(\.\.\/)+//;
$outfile =~ s/frqs/cuml_frqs/;

   $outfile = 'TESTs/Test1/total_cuml_frqs_gseg_configs_bulk.txt'; # For TESTING the script (Test1). #

(my $outdir = $outfile) =~ s/[^\/]+$//;
if ($outdir =~ m/\S/) {
	$outdir =~ s/\/$//;
} else {
	$outdir = '.';
}

if ($TEST_LEVEL >0) {
	print STDOUT "\n[INFORMATION '$program_name'] Output directory: $outdir .\n\n";
}
=cut # END of "For the predecessor, 'compute_dbl_cumulative_frequencies.pl'. (I/O-F)"

			## END of "MAJOR MODIFICATION in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (I/O-F)" ##


		### Check the input file. ###

foreach my $infile (@infiles) { # Added the loop in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (CHK) #
    unless (( -f $infile ) and ( -r $infile )) {
        print STDOUT "[ERROR '$program_name'] CanNOT FIND a readable INPUT FILE, '$infile': $? !\n\n";
        print STDERR "[ERROR '$program_name'] $! \n";
        die ;
    }
} # End of "Added the loop in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (CHK)" #

		## Create the output directory, if it does not exist yet. ##

unless ( create_dir_rw ($outdir) ) { 
    print STDOUT "[ERROR '$program_name'] The subroutine, 'create_dir_rw', FAILED on the  output directory, \"$outdir\" : $?\n\n";
    print STDERR "[ERROR '$program_name'] $!\n";
    die ;
}

		### Declare the Beginning of the Script ###

print STDOUT "\n<<< Program name: $program_name (path: $program_path) >>>\n\n";


			# MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (DC1) #

print STDOUT "Input files: \"$in_topdir\{in_dir_name\}/$infile_name\" ,\n";
print STDOUT "            where {in_dir_name} = '", join ("', '", @indir_names), "' .\n\n\n";

print STDOUT "Output file:\n\n";
print STDOUT "  (Observed Frequencies)  {out_dir}/$outfile_name ,\n";
print STDOUT "      where {out_dir} = $outdir .\n\n";
print STDOUT "\n";

=For the predecessor, 'compute_dbl_cumulative_frequencies.pl'. (DC1)
print STDOUT "Input file: $infile .\n\n";
print STDOUT "Output file: $outfile .\n\n\n";
=cut # End of "For the predecessor, 'compute_dbl_cumulative_frequencies.pl'. (DC1)"

			# END of "MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (DC1)" #

print STDOUT "<< Important Parameters: >>\n\n";


			# MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (DC2) #
print STDOUT '@TIME_LAPSES = (', join (', ', @TIME_LAPSES), ") .\n\n"; 

print STDOUT '$LABEL_CORNER_IN = ', "'$LABEL_CORNER_IN' .\n\n";
print STDOUT '@RELV_FLD_LABELS_IN = (', join (', ', @RELV_FLD_LABELS_IN), ") .\n\n";

print STDOUT '$LABEL_CORNER_OUT = ', "'$LABEL_CORNER_OUT' .\n\n";
print STDOUT '@RELV_FLD_LABELS_OUT = (', join (', ', @RELV_FLD_LABELS_OUT), ") .\n\n";

=For the predecessor, 'compute_dbl_cumulative_frequencies.pl'. (DC2)
print STDOUT '$LABEL_CORNER = ', "'$LABEL_CORNER' .\n\n";

print STDOUT '@RELV_FLD_LABELS = (', join (', ', @RELV_FLD_LABELS), ") .\n\n";
=cut # END of "For the predecessor, 'compute_dbl_cumulative_frequencies.pl'. (DC2)"
			# END of "MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (DC2)" #

print STDOUT '$STR_UNDEF = ', "'$STR_UNDEF' .\n\n";

				# (Modified in creation of 'juxtapose_raw_freqs_obs_case_iii.pl' on 2020/04/12). (DC) #
##print STDOUT '($MAX_X, $MAX_Y) = ', "($MAX_X, $MAX_Y) .\n\n\n";
print STDOUT '$MAX_Y = ', "$MAX_Y .\n\n\n";
				# End of "(Modified in creation of 'juxtapose_raw_freqs_obs_case_iii.pl' on 2020/04/12). (DC)" #


				# MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' on 2020/04/11. (0) #
my $infile0 = $infiles[0];
print STDOUT "\n<<< (0) Find Indexes Corresponding to the Field Labels in the input file ('$infile0'). >>>\n\n";
##print STDOUT "\n<<< (0) Find Indexes Corresponding to the Field Labels in the input file ('$infile'). >>>\n\n";
				# End of "MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' on 2020/04/11. (0)" #

local $/ = "\n"; # Set the input-record-separator to be the newline character. #

my $fh_in0 = open_input_file2 ($infile0);
unless (defined $fh_in0) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile0\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

my $len_lcorner_in = length ($LABEL_CORNER_IN);


my $tbl_header;
while (my $line0 = <$fh_in0>) {
    my $sbstr = substr ($line0, 0, $len_lcorner_in) ;
    if ($LABEL_CORNER_IN eq $sbstr) {
	$tbl_header = $line0;
	last;
    }
}
close ($fh_in0);

unless (defined $tbl_header) {
	print STDOUT "[ERROR '$program_name'] The input file \"$infile0\", has NO LINES starting with '$LABEL_CORNER_IN': $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}
$tbl_header =~ s/\s+$//;

if ($TEST_LEVEL >1) { # (0a) #
	print STDOUT "\n";
	print STDOUT "[INFORMATION '$program_name'] Retrieved the Table Header from the input file \"$infile0\":\n\n";
	print STDOUT "[INFORMATION '$program_name'] \"$tbl_header\" .\n";
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL >1) { # (0a) # ...}"

my $indexes_relv_flds = tfm_labels2indices_gen (@RELV_FLD_LABELS_IN, $tbl_header, $FLD_SEP);
unless (defined $indexes_relv_flds) {
	print STDOUT "[ERROR '$program_name'] The subroutine, 'tfm_labels2indices_gen' FAILED on the Table Header of the input file \"$infile0\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}

if ($TEST_LEVEL >1) { # (0b) #
	print STDOUT "\n[ Output of 'tfm_labels2indices_gen': ]\n\n";
	print STDOUT join ("\t", "'Label'", 'Index'), "\n\n";

	for (my $f=0; $f <$CT_RLV_FLDS_IN; $f++) {
	    my $label = $RELV_FLD_LABELS_IN[$f];
	    my $indx = $indexes_relv_flds->[$f];
	    print STDOUT join ("\t", "'$label'", $indx), "\n";
	}
	print STDOUT "\n\n";

} # End of "if ($TEST_LEVEL >1) { # (0b) # ...}"

			# ADDED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (Pre2) #

my $indx_rlv_fld_in = $indexes_relv_flds->[0];
 
if ($TEST_LEVEL >1) { # (0c) #

	print STDOUT "\n", '$indx_rlv_fld_in = ', $indx_rlv_fld_in, " .\n\n\n";

} # End of "if ($TEST_LEVEL >1) { # (0c) # ...}"
			# End of "ADDED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (Pre2)" #

			# MODIFIED in creation of 'juxtapose_raw_freqs_obs_case_iv.pl' (on 2020/04/11). (1a) #
##print STDOUT "\n\n<<< (1) Read in Relevant 2D arrays & Compute Their Double-Cumulative Counterparts ... >>>\n\n";
##my @set_dcum_arrays_2D = ();

print STDOUT "\n\n<<< (1) Read in Relevant 2D arrays & Store them in a 3D array ... >>>\n\n";
my @set_raw_arrays_2D = (); 

			# END of "MODIFIED in creation of 'juxtapose_raw_freqs_obs_case_iv.pl' (on 2020/04/11). (1a)" #


##for (my $f=0; $f <$CT_RLV_FLDS; $f++) { # 1st Main Outer for-loop (over the relevant fields). #
for (my $f=0; $f <@infiles; $f++) { # 1st Main Outer for-loop (over the input files). # MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (1st-Main) #


			# MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (1st-Main-a) #
	my $infile = $infiles[$f];
	my $tlapse = $TIME_LAPSES[$f];

	print STDOUT "< The $f th input file (for the time-lapse = $tlapse; path = \"$infile\") >\n\n";

=For the predecessor, 'compute_dbl_cumulative_frequencies.pl'. (1st Main)
	my $label_fld = $RELV_FLD_LABELS[$f];
	my $indx_fld = $indexes_relv_flds->[$f];

	print STDOUT "< The $f th field (label = '$label_fld', index = $indx_fld) >\n\n";
	
=cut # END of "For the predecessor, 'compute_dbl_cumulative_frequencies.pl'. (1st Main)"
			# End of "MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (1st-Main-a)" #

	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
		print STDOUT "[ERROR '$program_name'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	##my $len_lcorner = length ($LABEL_CORNER);

	my $dummy_str_end ;

	##my $array_2D = readin_2D_array (*${$fh_in}, $indx_fld, $STR_UNDEF, $dummy_str_end);
	my $array_2D = readin_2D_array (*${$fh_in}, $indx_rlv_fld_in, $STR_UNDEF, $dummy_str_end); # Modified in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (1st-Main-b) #
	unless (defined $array_2D) {
		##print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_2D_array', FAILED to retrieve the field of '$label_fld' in \"$infile\": $? !\n";
		print STDOUT "[ERROR '$program_name'] The subroutine, 'readin_2D_array', FAILED to retrieve the frequencies from the $f th input file, \"$infile\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	close ($fh_in);

	$set_raw_arrays_2D[$f] = $array_2D;

			# MODIFIED in creation of 'juxtapose_raw_freqs_obs_case_iv.pl' (on 2020/04/11). (1b) #

=DISABLED in creation of  'juxtapose_raw_freqs_obs_case_iv.pl'.

	my $dcum_array_2D = cmpt_dbl_cumulative_2Darray (@{$array_2D}, $MAX_X, $MAX_Y);
	unless (defined $dcum_array_2D) {
		##print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_dbl_cumulative_2Darray', FAILED on the 2D array constructed from the field of '$label_fld' in \"$infile\": $? !\n";
		print STDOUT "[ERROR '$program_name'] The subroutine, 'cmpt_dbl_cumulative_2Darray', FAILED on the 2D array constructed from the frequencies from the $f th input file, \"$infile\": $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		die ;
	}

	$set_dcum_arrays_2D[$f] = $dcum_array_2D;
=cut # END of "DISABLED in creation of  'juxtapose_raw_freqs_obs_case_iv.pl'."

			# END of "MODIFIED in creation of 'juxtapose_raw_freqs_obs_case_iv.pl' (on 2020/04/11). (1b)" #

}  # End of the "1st Main Outer for-loop (over the input files). # MODIFIED in creation of 'compute_dbl_cumulative_frequencies_obs.pl' (on 2020/04/11). (1st-Main) #"
##} # End of the 1st Main outer for-loop (over the relevant fields). #



##print STDOUT "\n\n<<< (2) Output the Computed Double-Cumulative 2D arrays to a single file (= \"$outfile\") ... >>>\n\n";
print STDOUT "\n\n<<< (2) Output the RAW 2D arrays to a single file (= \"$outfile\") ... >>>\n\n";

my $fh_out = open_output_file ($outfile, 0);
unless (defined $fh_out) {
	print STDOUT "[ERROR '$program_name'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$outfile\": $? !\n";
	print STDERR "[ERROR '$program_name'] $!\n";
	die ;
}


##print $fh_out join ("\t", $LABEL_CORNER, @RELV_FLD_LABELS), "\n\n";
print $fh_out join ("\t", $LABEL_CORNER_OUT, @RELV_FLD_LABELS_OUT), "\n\n";

####for (my $x=0; $x <$MAX_X; $x++) { # 2nd Main outer for-loop (over $x). #
##for (my $x=0; $x <= $MAX_X; $x++) { # 2nd Main outer for-loop (over $x). # (CORRECTED on 2020/04/10, in creation of 'compute_dbl_cumulative_frequencies_theory.pl'.) (2nd Main outer)
##for (my $x=1; $x <= $MAX_X; $x++) { # 2nd Main outer for-loop (over $x). # (=> Further Modified on 2020/04/11, in creation of 'compute_dbl_cumulative_frequencies_theory.pl'.) (2nd Main outer)
for (my $x=0; $x < 1; $x++) { # 2nd Main outer for-loop (over $x). # (=> MAJOR MODIFICATION in creation of 'juxtapose_raw_freqs_obs_case_iii.pl' (on 2020/04/12).  (2nd Main outer)

				# MODIFIED in creation of 'juxtapose_raw_freqs_obs_case_iv.pl' (on 2020/04/11). (2a) #
	##my @set_rows_dcum = ();
	##for (my $f=0; $f <$CT_RLV_FLDS_OUT; $f++) { $set_rows_dcum[$f] = $set_dcum_arrays_2D[$f]->[$x]; } # The 1st middle for-loop (over relevant fields). #

	my @set_rows_raw = ();
	for (my $f=0; $f <$CT_RLV_FLDS_OUT; $f++) {
	    my $row_raw = $set_raw_arrays_2D[$f]->[$x];
	    unless (defined $row_raw) { $row_raw = []; }
	    $set_rows_raw[$f] = $row_raw;
	} # The 1st middle for-loop (over relevant fields). #
				# End of "MODIFIED in creation of 'juxtapose_raw_freqs_obs_case_iv.pl' (on 2020/04/11). (2a)" #

	####for (my $y=0; $y <$MAX_Y; $y++) { # The 2nd Middle for-loop (over $y). #
	##for (my $y=0; $y <= $MAX_Y; $y++) { # The 2nd Middle for-loop (over $y). # (CORRECTED on 2020/04/10, in creation of 'compute_dbl_cumulative_frequencies_theory.pl'.) (2nd Middle) #
	for (my $y=1; $y <= $MAX_Y; $y++) { # The 2nd Middle for-loop (over $y). # (=> Further Modified on 2020/04/10, in creation of 'compute_dbl_cumulative_frequencies_theory.pl'.) (2nd Middle) #

	    my @out_flds = ();

				# MODIFIED in creation of 'juxtapose_raw_freqs_obs_case_iv.pl' (on 2020/04/11). (2b) #
	    ##for (my $f=0; $f <$CT_RLV_FLDS_OUT; $f++) { $out_flds[$f] = $set_rows_dcum[$f]->[$y]; } # The inner for-loop (over relevant fields). #
	    for (my $f=0; $f <$CT_RLV_FLDS_OUT; $f++) {
		my $raw = $set_rows_raw[$f]->[$y];
		unless (defined $raw) { $raw = 0; }
		$out_flds[$f] = $raw;
	    }
				# End of "MODIFIED in creation of 'juxtapose_raw_freqs_obs_case_iv.pl' (on 2020/04/11). (2b)" #

			# MAJOR MODIFICATION in creation of 'juxtapose_raw_freqs_obs_case_iii.pl' (on 2020/04/12). (MAIN 2) #
	    ####print $fh_out join ("$x:$y", @out_flds), "\n"; # A BUG Fixed on 2020/04/11. #
	    ##print $fh_out join ("\t", "$x:$y", @out_flds), "\n";
	    print $fh_out join ("\t", $y, @out_flds), "\n";
			# End of "MAJOR MODIFICATION in creation of 'juxtapose_raw_freqs_obs_case_iii.pl' (on 2020/04/12). (MAIN 2)" #

	} # End of the 2nd Middle for-loop (over $y). #

}  # 2nd Main outer for-loop (over $x). #

close ($fh_out);


	##### DEFINITIONS of the Subroutines. #####


# create_dir_rw:
#	This subroutine receives the path to a directory,
#	examines (1) whether the directory exists.
#	Otherwise, it creates the directory.
#	Then, it examines (2) whether the directory is readable & writable.
#	Otherwise, it changes the permission mode of the directory to make it readable & writable.
#	[CREATED on 2020/01/29; IMPORTED from 'lastpiece.alpha.pl' on 2020/04/09.]
#
sub create_dir_rw (\$) {

	my ($rf_dir_path) = @_;
	my $dir_path = ${$rf_dir_path};

	unless ( -d $dir_path ) {
	   unless (0 == system ("mkdir -p $dir_path") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to CREATE the OUTPUT DIRECTORY, '$dir_path/': $? !\n\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		return 0 ;
	   }
	}
	unless (( -r $dir_path ) and ( -w $dir_path )) {
	    unless (0 == system ("chmod u+rw $dir_path") ) {
		print STDOUT "[ERROR '$program_name'] FAILED to CHANGE permission MODES of the output directory, '$dir_path': $? !\n";
		print STDERR "[ERROR '$program_name'] $!\n";
		return 0 ;
	    }
	}

	return 1;

}	# End of the subroutine, 'create_dir_rw'. #


