package MyArrayManipu;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyArrayManipu.pm: a package for various array manipulations
#
# REVISED on April 24 (Thu), 2003: I added a subroutine "lin_search",
#				which performs a linear search to see if
#				a scalar belongs to an array.
#

use strict;
use warnings;
use Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw( diff_array union_array nr_array lin_search);
our @EXPORT_OK = qw(
	nr_array_ii append_nr_array
	group_equal_measures multiple_sort one2two_rowsize one2two_rowct two2one
	f_rotate_array b_rotate_array swap_elements interdigitate_arrays
	compare_arrays
	diff_array_of_arrays
	diff_arrays_of_compo_arrays
	union_array_of_arrays
	nr_array_of_arrays
	binsearch_array binsearch_string
	sort_mixed_elements merge_intervals
	merge_integer_intervals compare_interval_sets
	pos_consist_intvls total_length_intvls max_length_intvls
	linsearch_closest_intervals linsearch_ovlp_intervals
	screen_intvls_by_len
	set_max_nonovlp_regions
	transpose_matrix
		);

sub compare_arrays (\@\@\@\@);
sub group_equal_measures (\@\@\@);
sub multiple_sort (\@\@\@);
sub one2two_row (\@$);


# diff_array: a subroutine which receives as arguments two references of arrays,
#		and returns the references of THREE arrays,
#		one is the intersection of the argument arrays,
#		and the other two are their non_overlapping parts.
#
sub diff_array (\@\@) {

	my ($rf_a1,$rf_a2) = @_;
	my @common = my @ext_a1 = my @ext_a2 = ();

	my @sarray1 = sort @{$rf_a1};
	my @sarray2 = sort @{$rf_a2};

	my $item1 = shift @sarray1;
	my $item2 = shift @sarray2;

	while (1) {					# the 'infinite' while loop #

		unless (defined $item1) {
			if (defined $item2) {
				push @ext_a2, $item2;
				@sarray2 and (push @ext_a2, @sarray2); # cram the remaining elements #
			}
			last;
		}

		unless (defined $item2) {
			if (defined $item1) {
				push @ext_a1, $item1;
				@sarray1 and (push @ext_a1, @sarray1); # cram the remaining elements #
			}
			last;
		}

		if ($item1 eq $item2) {			# if-elsif-else #
			push @common, $item1;
			$item1 = shift @sarray1;
			$item2 = shift @sarray2;

		} elsif ($item1 lt $item2) {
			push @ext_a1, $item1;
			$item1 = shift @sarray1;

		} else {
			push @ext_a2, $item2;
			$item2 = shift @sarray2;
		}					# end of the if_elsif_else #

	}						# end of the 'infinite' while loop #

	return (\@common,\@ext_a1,\@ext_a2);

}							# end of the module diff_array #

# union_array: a subroutine which receives as arguments two references of arrays,
#		and returns the reference of  their union.
#
sub union_array (\@\@) {

	my ($rf_a1,$rf_a2) = @_;
	my @union = ();

	my @sarray1 = sort @{$rf_a1};
	my @sarray2 = sort @{$rf_a2};

	my $item1 = shift @sarray1;
	my $item2 = shift @sarray2;

	while (1) {					# the 'infinite' while loop #

		unless (defined $item1) {
			if (defined $item2) {
				push @union, $item2;
				@sarray2 and (push @union, @sarray2); # cram the remaining elements #
			}
			last;
		}

		unless (defined $item2) {
			if (defined $item1) {
				push @union, $item1;
				@sarray1 and (push @union, @sarray1); # cram the remaining elements #
			}
			last;
		}

		if ($item1 eq $item2) {			# if-elsif-else #
			push @union, $item1;
			$item1 = shift @sarray1;
			$item2 = shift @sarray2;

		} elsif ($item1 lt $item2) {
			push @union, $item1;
			$item1 = shift @sarray1;

		} else {
			push @union, $item2;
			$item2 = shift @sarray2;
		}					# end of the if_elsif_else #

	}						# end of the 'infinite' while loop #

	return \@union;

}							# end of the module union_array #


# nr_array: A subroutine which removes the redundancy of a list.
#
sub nr_array (\@) {

	my ($rf_array) = @_;

	(@{$rf_array} < 2) and (return $rf_array);

	my @sorted_array = sort @{$rf_array};
	my $last = shift @sorted_array;
	my @output = ($last);

	while (my $pres = shift @sorted_array) {

		($last eq $pres) and next;

		$last = $pres;
		push @output, $pres;
	}

	return \@output;
}				# end of the subroutine 'nr_array' #


# nr_array_ii:
#	This subroutine removes the redundancy of an array
#	WHILE RESPECTING THE ORIGINAL ORDER of
#	the elements when they first appear.
#
sub nr_array_ii (\@) {

	my ($in_array) = @_;

	my $size = @{$in_array};
	($size <2) and (return $in_array);

	my %already; # ($element => 1, ...) for those which already appeared.
	my @out_array = ();

	foreach my $ele (@{$in_array}) { # the foreach-loop #
	    (defined $already{$ele}) and next;

	    push @out_array, $ele;
	    $already{$ele} = 1;
	}			# end of the foreach-loop #

	return \@out_array;

}			# end of the subroutine 'nr_array_ii' #


# append_nr_array:
#	This subroutine recieves 2 arrays,
#	and appends the second array to the 1st one
#	in an NON-REDUNDANT manner.
#	And the order of the elements are respected.
#	AS a side-effect, the subroutine takes only a non-redundant
#	subset of the 1st array.
#
# INPUT: (\@array1, \@array2, \@indices), where
#	the elements @{$array1/2[$m]}[@indices] are used for judging the redundancy.
#	If @indices is empty, compare $array1/2[$m] themselves.
#
# OUTPUT: \@appended, which is @array1 with @array2 NONREDUNDANTLY appended.
#
sub append_nr_array (\@\@\@) {

	my ($array1, $array2, $indices) = @_;

	my $fl_total = (@{$indices}>0) ? 0 : 1;
	my %already;

	my @appended = ();

	foreach my $ele (@{$array1}, @{$array2}) { # The main foreach-loop #
	    my $sbjct = $fl_total ? $ele : join ('\\', @{$ele}[@{$indices}]);
	    (defined $already{$sbjct}) and next;

	    push @appended, $ele;
	    $already{$sbjct} = 1;
	}		# End of the main foreach-loop #

	return \@appended;

}		# end of the subroutine 'append_nr_array' #


# lin_search: This subroutine sees if a scalar belongs to an array
#		by performing a linear search.
#		The subroutine returns the relevant index if the array contains the scalar,
#		and returns -1 otherwise.
#
sub lin_search ($\@) {

	my ($scalar, $rf_array) = @_;

	for (my $i=0; $i<@{$rf_array}; $i++) {

		($scalar eq $rf_array->[$i]) and (return $i);

	}

	return -1;
}							# end of the subroutine lin_search #


# group_equal_measures: This subroutine receives an array of arrays as the 0th argument and
#			clusters the element arrays with the equal elements specified by the 1st argument.
#			The 2nd argument specifies the comparison method:
#			'+', '-' or undef --- numerical;
#			'a' or 'd'        ---- according to  the letter code.
#
sub group_equal_measures (\@\@\@) {

	my ($rf_array, $rf_indices, $rf_methods) = @_;

	my @array = @{$rf_array};
	(@array)  or (return []);
	(@array == 1) and (return [\@array]);

	my @indices = @{$rf_indices};
	my $ct_indices = @indices;
	my @methods = @{$rf_methods};
	my $last = shift @array;
	my @grouped = ([$last]);

	while (my $pres = shift @array) {

		my $fl_identical = 1;
		foreach my $i (0 ..  $ct_indices -1) {

			my $index = $indices[$i];
			my $method = $methods[$i];
			unless ( (defined $pres->[$index]) and (defined $last->[$index]) ) {
				$fl_identical = 0;
				last;
			} 

			if (defined $method) {		# the outer if-else #
				if ( ($method eq '+') or ($method eq '-') ) {

					unless ($pres->[$index] == $last->[$index]) {
						$fl_identical = 0;
						last;
					}

				} elsif ( ($method eq 'a') or ($method eq 'd') ) {

					unless ($pres->[$index] eq $last->[$index]) {
						$fl_identical = 0;
						last;
					}

				} else {
					print STDERR "Illegal method specifier for 'group_equal_measures'.\n";
					return undef;
				}

			} else {
				unless ($pres->[$index] == $last->[$index]) {
					$fl_identical = 0;
					last;
				} 
			}			# end of the outer if-else #

		}			# end of foreach #

		if ($fl_identical) {
			push @{$grouped[$#grouped]}, $pres;
		} else {
			push @grouped, [$pres];
		}

		$last = $pres;

	}			# end of while #

	return \@grouped;

}			# end of the subroutine group_equal_measures #


# multiple_sort: This subroutine  sorts an array of arrays recieved as the 0th argument,
#		by using multiple elements in the element arrays.
#		The 1st argument specifies the order of preference of the elements. 
#		The 2nd argument specifies the sorting method:
#		'+' (or undef) --- in the numerally ascending order;
#		'-' --- in the numerally descending order;
#		'a' --- in the ascending order of the letter code;
#		'd' --- in the descending order of the letter code.
#
sub multiple_sort (\@\@\@) {

	my ($rf_array, $rf_indices, $rf_methods) = @_;

	(scalar @{$rf_array} < 2) and (return $rf_array);
	my @indices = @{$rf_indices};
	scalar(@indices) or (return $rf_array);
	my @methods = @{$rf_methods};
	my $index = shift @indices;
	my $method = shift @methods;

	my @pre_sorted = ();
	if (defined $method) {

		if ($method eq '+') {
			@pre_sorted = sort { $a->[$index] <=> $b->[$index] } @{$rf_array};
		} elsif ($method eq '-') {
			@pre_sorted = sort { $b->[$index] <=> $a->[$index] } @{$rf_array};
		} elsif ($method eq 'a') {
			@pre_sorted = sort { $a->[$index] cmp $b->[$index] } @{$rf_array};
		} elsif ($method eq 'd') {
			@pre_sorted = sort { $b->[$index] cmp $a->[$index] } @{$rf_array};
		} else {
			print STDERR "Illegal method-specifier for 'multiple_sort: $method.\n";
			return undef;
		}

	} else {

		@pre_sorted = sort { $a->[$index] <=> $b->[$index] } @{$rf_array};
	}

	scalar(@indices) or (return \@pre_sorted);

	my @tmp_indices = ($index);
	my @tmp_methods = ($method);
	my $rf_grouped = group_equal_measures (@pre_sorted, @tmp_indices, @tmp_methods);

	my @sorted_fin = ();
	foreach my $rf_subset (@{$rf_grouped}) {

		if (@{$rf_subset} < 2)  {

			push @sorted_fin, @{$rf_subset};
			next;
		}

		my $rf_substd = multiple_sort (@{$rf_subset}, @indices, @methods);
		push @sorted_fin, @{$rf_substd};

	}		# end of foreach #

	return \@sorted_fin;

}				# end of the subroutine multiple_sort #

# one2two_rowsize: This subroutine receives a 1-dimensional array
#		and converts it to a 2-dimensional array
#		with its 'row-size' specified by the 1st argument.
#
sub one2two_rowsize (\@$) {

	my ($rf_1dim, $row_size) = @_;

	my $size_1dim = @{$rf_1dim};
	$size_1dim or (return []);

	my @two_dim = ();
	my $init_indx = 0;

	while ( (my $end_indx = $init_indx +$row_size -1) < $size_1dim) {

		my @tmp = @{$rf_1dim}[$init_indx .. $end_indx];
		push @two_dim, \@tmp;

		$init_indx += $row_size;
	}

	if ($init_indx < $size_1dim) {

		my @tmp = @{$rf_1dim}[$init_indx .. $size_1dim -1 ];
		push @two_dim, \@tmp;
	} 

	return \@two_dim;

}		# end of the subroutine one2two_rowsize #

# one2two_rowct: This subroutine receives a 1-dimensional array,
#		divides it into a specified number of rows which are
#		distributed as equally as possible,
#		and returns the resulting 2-dimensional array.
#		
sub one2two_rowct (\@$) {

	my ($rf_1dim, $ct_rows) = @_;

	my $ct_1dim = @{$rf_1dim};
	my @output = ();

	if ($ct_1dim < $ct_rows) {
		print STDERR "CAUTION: The specified row-count ($ct_rows) is greater than the size of the input array ($ct_1dim).\n";

		for my $ele (@{$rf_1dim}) { push @output, [$ele]; }
		return \@output;
	}

	my $size_row = int ($ct_1dim/ $ct_rows);

	my $residue = ($ct_1dim % $ct_rows);
	my $flank_init = int ( ($ct_rows - $residue) / 2);

		# make rows of size $size_row + 1 #
	my $init_indx = 0;
	for (my $i=0; $i<$flank_init; $i++) {

		my $end_indx = $init_indx + $size_row - 1;
		my @tmp = @{$rf_1dim}[$init_indx .. $end_indx ];
		push @output , \@tmp;
		$init_indx = $end_indx + 1;
	}

		# make rows of size $size_row #
	for (my $i = $flank_init; $i<$flank_init+$residue; $i++) {

		my $end_indx = $init_indx + $size_row;
		my @tmp = @{$rf_1dim}[$init_indx .. $end_indx ];
		push @output , \@tmp;
		$init_indx = $end_indx + 1;
	}

	for (my $i=$flank_init+$residue; $i<$ct_rows; $i++) {

		my $end_indx = $init_indx + $size_row - 1;
		my @tmp = @{$rf_1dim}[$init_indx .. $end_indx ];
		push @output , \@tmp;
		$init_indx = $end_indx + 1;
	}

	return \@output;

}		# end of the subroutine one2two_rowct #

# two2one: This subroutine transforms a 2-dim array into 
#		a 1-dim array by "graft"ing the rows in the 2-dim array.
#
sub two2one (\@) {

	my ($rf_2dim) = @_;

	my $ct_rows = @{$rf_2dim};

	my @one_dim = ();

	for (my $i=0; $i<$ct_rows; $i++) { push @one_dim, @{$rf_2dim->[$i]}; }

	return \@one_dim;

}		# end of the subroutine two2one #


# b_rotate_array: This subroutine moves the m elements at the tail of the input array
#		  to its head.
#
sub b_rotate_array (\@$) {

	my ($in_array, $ct) = @_;
	my @out_array = @{$in_array};
	(@out_array < 2) and (return \@out_array);
	
	for (my $i=0; $i<$ct; $i++) { 
		my $tail = pop @out_array;
		unshift @out_array, $tail;
	}
	return \@out_array;

}		# end of the subroutine b_rotate_array #


# f_rotate_array: This subroutine moves the m elements at the head of the input array
#		  to its tail.
#
sub f_rotate_array (\@$) {

	my ($in_array, $ct) = @_;
	my @out_array = @{$in_array};
	(@out_array < 2) and (return \@out_array);
	
	for (my $i=0; $i<$ct; $i++) { 
		my $head = shift @out_array;
		push @out_array, $head;
	}
	return \@out_array;

}		# end of the subroutine f_rotate_array #


# swap_elements: This subroutine returns an array which is produced by
#		 swappint 2 elements of the input array.
#
sub swap_elements (\@$$) {

	my ($in_array, $indx1, $indx2) = @_;

	my $ct_elements  = @{$in_array};
	if ( ($indx1 <0) or ($indx1 >= $ct_elements) or ($indx2 <0) or ($indx2 >= $ct_elements) ) {
		print STDERR "Illegal input for \'swap_elements\': (array_size, indx1, indx2) = ($ct_elements, $indx1, $indx2).\n";
		return undef; 
	}

	my @out_array = @{$in_array};

	@out_array[$indx1, $indx2] = @out_array[$indx2, $indx1];

	return \@out_array;
}


# interdigitate_arrays: This subroutine incorporates elements of an array just behind
#			elements of another array.
#			The latter is specified  by a set of indices.
#			
#
# INPUT: (\@array1, \@array2, \@indices) 
# suppose that @indices are not redundant.
#
sub interdigitate_arrays (\@\@\@) {

	my ($array1, $array2, $indices) = @_;

	my $ct_ele1 = @{$array1};
	my $ct_ele2 = @{$array2};
	my $ct_indices = @{$indices};
	unless ($ct_indices ) {
		my @cp_array1 = @{$array1};
		return \@cp_array1;
	}
	my @std_indices = sort { $a <=> $b } @{$indices};

	if ( ($std_indices[0] < 0) or ($std_indices[$ct_indices -1] >= $ct_ele1 )  ) {
		printf STDERR "Illegal input for \'interdigitate_arrays\': (smallest index, largest index, size_array1) = (%d, %d, %d) .\n", @std_indices[0, $ct_indices -1], $ct_ele1;
		return undef;
	}

	my @out_array = ();
	my $index = shift @std_indices;
	my @cp_array2 = @{$array2};

	for (my $i=0; $i <$ct_ele1; $i++) {

		push @out_array, $array1->[$i];
		((defined $index) and ($index == $i)) or next;

		push @out_array, (shift @cp_array2);
		$index = shift @std_indices;
	}

	return \@out_array;

}		# end of the subroutine interdigitate_arrays #


# compare_arrays: This subroutine compares 2 arrays (the 0th and 1st arguments)
#		according to the elements listed in the  2nd argument.
#		The  3rd argument is a list of methods used for the comparison of each element:
#		'+' (or undef) --- in the numerally ascending order;
#		'-' --- in the numerally descending order;
#		'a' --- in the ascending order of the letter code;
#		'd' --- in the descending order of the letter code.
#
# OUTPUT: 0 if @array1 is 'as large as' @array2,
#	 -1 if @array1 is 'smaller than' @array2,
#	 +1 if @array1 is 'larger than' @array2,
#	 undef if some error has occurred.
#
sub compare_arrays (\@\@\@\@) {

	my ($array1, $array2, $indices, $methods) = @_;

	my $ct_indices = @{$indices};

	if (@{$methods} > $ct_indices) {
		print  STDERR "Illegal input for 'compare_arrays': (#indices, #methods) = (%d, $ct_indices).\n",
			scalar (@{$methods});
		return undef;
	}

	for (my $i=0; $i <$ct_indices; $i++) { 

		my $indx = $indices->[$i];
		my $method = $methods->[$i];
		my $ele1 = $array1->[$indx];
		my $ele2 = $array2->[$indx];

		unless ( (defined $ele1) and (defined $ele2) ) {
			print STDERR "The $indx -th element is undefined for the array(s) to be compared via 'compare_arrays'.\n";
			return undef;
		}

		if ( !(defined $method) or ($method eq '+') ) {

			if ($ele1 < $ele2) { return -1; }
			elsif ($ele1 > $ele2) { return 1; }

		} elsif ($method eq '-')  {

			if ($ele1 > $ele2) { return -1; }
			elsif ($ele1 < $ele2) { return 1; }

		} elsif ($method eq 'a') {

			if ($ele1 lt $ele2) { return -1; }
			elsif ($ele1 gt $ele2) { return 1; }

		} elsif ($method eq 'd') {

			if ($ele1 gt $ele2) { return -1; }
			elsif ($ele1 lt $ele2) { return 1; }
		}
	}

	return 0; # @{$array1} is regarded as 'equal to' @{$array2} according to this comparison

} 	# end of the subroutine compare_arrays #


# diff_array_of_arrays: a subroutine which receives as arguments
#	 two references of arrays whose elements  being array references,
#		compares each element of an array with each of the other one,
#		and returns the references of FOUR arrays,
#		two are for the intersection of the input arrays
#		(each of them is judged as 'common' part of each input array),
#		and the other two are their non_overlapping parts.
#
# INPUT: (\@array1, \@array2, \@indices, \@methods),
#	where @indices and @methods are in accord with  those in 'multiple_sort' or 'group_equal_measures'.
#	Namely,  @indices pecifies the order of the elements which are used for the sorting. 
#		@methods specifies the sorting method:
#		'+' (or undef) --- in the numerally ascending order;
#		'-' --- in the numerally descending order;
#		'a' --- in the ascending order of the letter code;
#		'd' --- in the descending order of the letter code.
#
# PREREQUISITE: The input arrays are presorted with 'sort' or 'multiple_sort' or something.
#
sub diff_array_of_arrays (\@\@\@\@) {

	my ($rf_a1,$rf_a2, $indices, $methods) = @_;
	my @common1 = my @common2 = my @ext_a1 = my @ext_a2 = ();

	unless (@{$indices} == @{$methods}) {
		print STDERR "Illegal input for 'diff_array_of_arrays': (#indices, #methods) = (%d, %d) .\n",
			scalar (@{$indices}) , scalar (@{$methods}); 
		return (undef, undef, undef, undef);

	} 
	#my @sarray1 = sort { $a->[$indx_cmp] <=> $b->[$indx_cmp]}
	#		 @{$rf_a1};
	#my @sarray2 = sort { $a->[$indx_cmp] <=> $b->[$indx_cmp] }
	#		 @{$rf_a2};
	my @sarray1 = @{$rf_a1};
	my @sarray2 = @{$rf_a2};

	my $item1 = shift @sarray1;
	my $item2 = shift @sarray2;

	while (1) {					# the 'infinite' while loop #

		unless (defined $item1) {
			if (defined $item2) {
				push @ext_a2, $item2;
				@sarray2 and (push @ext_a2, @sarray2); # cram the remaining elements #
			}
			last;
		}

		unless (defined $item2) {
			if (defined $item1) {
				push @ext_a1, $item1;
				@sarray1 and (push @ext_a1, @sarray1); # cram the remaining elements #
			}
			last;
		}

		my $out_comp = compare_arrays (@{$item1}, @{$item2}, @{$indices}, @{$methods});

		if (!(defined $out_comp) ) {

			print "[ERROR 'diff_array_of_arrays'] 'compare_arrays' failed .\n";
			return (undef, undef, undef, undef);

		#if ($item1->[$indx_cmp] == $item2->[$indx_cmp]) {# if-elsif-else #
		} elsif ($out_comp == 0) { # $item1  is 'as large as' $item2
			push @common1, $item1;
			push @common2, $item2;
			$item1 = shift @sarray1;
			$item2 = shift @sarray2;

		#} elsif ($item1->[$indx_cmp] <  $item2->[$indx_cmp]) {
		} elsif ($out_comp < 0) {  # $item1 is 'smaller than' $item2 #
			push @ext_a1, $item1;
			$item1 = shift @sarray1;

		} else { # $item1 is 'larger than' $item2
			push @ext_a2, $item2;
			$item2 = shift @sarray2;
		}			# end of the if_elsif_else #

	}			# end of the 'infinite' while loop #

	return (\@common1, \@common2, \@ext_a1, \@ext_a2);

}		# end of the subroutine diff_array_of_arrays #


# diff_arrays_of_compo_arrays:
#	This revised version of 'diff_array_of_arrays'
#	compares two arrays, the elements of each of which are
#	the references to 'composite' arrays.
#	The comparison criteria are basically the same as those in 'diff_array_of_arrays'.
#
# PREREQUISITE: The input arrays @array1 and @array2
#	are presorted with 'sort' or 'multiple_sort' or something.
#
# INPUT: (\@array1, \@array2, $indx_sb1, $indx_sb2, \@indices, \@methods), where
#	@array1 and @array2 are the two arrays of composite arrays to be compared;
#	$indx_sb1/2 specifies the element to be compared.
#	More specifically, @{$array1[$k]->[$indx_sb1]} and @{$array2[$l]->[$indx_sb2]}
#	are compared.
#	If $indx_sb1 is undefined, @{$array1[$k]} itself is compared,
#	and @{$array2[$l]} is compared when $indx_sb2 is undefined;
#	@indices and @methods are in accord with  those in 'multiple_sort' or 'group_equal_measures'.
#	Namely,  @indices pecifies the order of the elements which are used for the sorting. 
#		@methods specifies the sorting method:
#		'+' (or undef) --- in the numerally ascending order;
#		'-' --- in the numerally descending order;
#		'a' --- in the ascending order of the letter code;
#		'd' --- in the descending order of the letter code.
#
# OUTPUT: (\@common1, \@common2, \@ext_a1, \@ext_a2), where 
#	@common1/2 contains the elements of @array1/2 'shared' by both;
#	@ext_a1/2 contains the elements 'specific' to  @array1/2.
#
sub diff_arrays_of_compo_arrays (\@\@$$\@\@) {

	my ($rf_a1,$rf_a2, $indx_sb1, $indx_sb2, $indices, $methods) = @_;
	my @common1 = my @common2 = my @ext_a1 = my @ext_a2 = ();

	unless (@{$indices} == @{$methods}) {
		print STDERR "[ERROR 'diff_arrays_of_compo_arrays'] ILLEGAL INPUT: (#indices, #methods) = (%d, %d) .\n",
			scalar (@{$indices}) , scalar (@{$methods}); 
		return (undef, undef, undef, undef);
	} 

	my @sarray1 = @{$rf_a1};
	my @sarray2 = @{$rf_a2};

	my $item1 = shift @sarray1;
	my $item2 = shift @sarray2;

	while (1) {					# the 'infinite' while loop #

		unless (defined $item1) {
			if (defined $item2) {
				push @ext_a2, $item2;
				@sarray2 and (push @ext_a2, @sarray2); # cram the remaining elements #
			}
			last;
		}

		unless (defined $item2) {
			if (defined $item1) {
				push @ext_a1, $item1;
				@sarray1 and (push @ext_a1, @sarray1); # cram the remaining elements #
			}
			last;
		}

			## REVISED in creation of 'diff_arrays_of_compo_arrays'
			## (on May 12 (Wed), 2010)
		my $sb1 = (defined $indx_sb1) ? $item1->[$indx_sb1] : $item1;
		my $sb2 = (defined $indx_sb2) ? $item2->[$indx_sb2] : $item2;
		my $out_comp = compare_arrays (@{$sb1}, @{$sb2}, @{$indices}, @{$methods});
		##my $out_comp = compare_arrays (@{$item1}, @{$item2}, @{$indices}, @{$methods});

		if (!(defined $out_comp) ) {

			print "[ERROR 'diff_arrays_of_compo_arrays'] 'compare_arrays' failed .\n";
			return (undef, undef, undef, undef);

		#if ($item1->[$indx_cmp] == $item2->[$indx_cmp]) {# if-elsif-else #
		} elsif ($out_comp == 0) { # $item1  is 'as large as' $item2
			push @common1, $item1;
			push @common2, $item2;
			$item1 = shift @sarray1;
			$item2 = shift @sarray2;

		#} elsif ($item1->[$indx_cmp] <  $item2->[$indx_cmp]) {
		} elsif ($out_comp < 0) {  # $item1 is 'smaller than' $item2 #
			push @ext_a1, $item1;
			$item1 = shift @sarray1;

		} else { # $item1 is 'larger than' $item2
			push @ext_a2, $item2;
			$item2 = shift @sarray2;
		}			# end of the if_elsif_else #

	}			# end of the 'infinite' while loop #

	return (\@common1, \@common2, \@ext_a1, \@ext_a2);

}		# end of the subroutine diff_arrays_of_compo_arrays #


# union_array_of_arrays: a subroutine which receives as arguments
#	 two references of arrays whose elements  being array references,
#		compares each element of an array with each of the other one,
#		and returns the reference of an array of arrays
#		which is the union of the two input composite arrays.
#	When encoutering elements of THE EQUAL MEASURE,
#	the subroutine incorporates the element of the 1st array.
#	I therefore recommend that the ID, if any, should be used at least
#	as the last element to be compared.
#
# INPUT: (\@array1, \@array2, \@indices, \@methods),
#	where @indices and @methods are in accord with  those in 'multiple_sort' or 'group_equal_measures'.
#	Namely,  @indices pecifies the order of the elements which are used for the sorting. 
#		@methods specifies the sorting method:
#		'+' (or undef) --- in the numerally ascending order;
#		'-' --- in the numerally descending order;
#		'a' --- in the ascending order of the letter code;
#		'd' --- in the descending order of the letter code.
#
# PREREQUISITE: The input arrays are presorted with 'sort' or 'multiple_sort' or something.
#
sub union_array_of_arrays (\@\@\@\@) {

	my ($rf_a1,$rf_a2, $indices, $methods) = @_;
	my @union = ();

	unless (@{$indices} == @{$methods}) {
		print STDERR "[ERROR 'union_array_of_arrays'] Found an ILLEGAL input: (#indices, #methods) = (%d, %d) .\n",
			scalar (@{$indices}) , scalar (@{$methods}); 
		return undef;
	} 
	my @sarray1 = @{$rf_a1};
	my @sarray2 = @{$rf_a2};

	my $item1 = shift @sarray1;
	my $item2 = shift @sarray2;

	while (1) {			# the 'infinite' while loop #

		unless (defined $item1) {
			if (defined $item2) {
				push @union, $item2;
				@sarray2 and (push @union, @sarray2); # cram the remaining elements #
			}
			last;
		}

		unless (defined $item2) {
			if (defined $item1) {
				push @union, $item1;
				@sarray1 and (push @union, @sarray1); # cram the remaining elements #
			}
			last;
		}

		my $out_comp = compare_arrays (@{$item1}, @{$item2}, @{$indices}, @{$methods});

		if (!(defined $out_comp) ) {

			print "[ERROR 'union_array_of_arrays'] 'compare_arrays' failed .\n";
			return undef;

		#if ($item1->[$indx_cmp] == $item2->[$indx_cmp]) {# if-elsif-else #
		} elsif ($out_comp == 0) { # $item1  is 'as large as' $item2
			push @union, $item1;
			$item1 = shift @sarray1;
			$item2 = shift @sarray2;

		} elsif ($out_comp < 0) {  # $item1 is 'smaller than' $item2 #
			push @union, $item1;
			$item1 = shift @sarray1;

		} else { # $item1 is 'larger than' $item2
			push @union, $item2;
			$item2 = shift @sarray2;
		}			# end of the if_elsif_else #

	}			# end of the 'infinite' while loop #

	return \@union;

}		# end of the subroutine union_array_of_arrays #


# nr_array_of_arrays: a subroutine which removes redundancy of an array of arrays.
#
# PREREQUISITE: The input array is pre-sorted with multiple_sort or sort.
#
sub nr_array_of_arrays (\@\@\@) {

	my ($rf_array, $indices, $methods) = @_;

	(@{$rf_array} < 2) and (return $rf_array);

	my @sorted_array = @{$rf_array};
	my $last = shift @sorted_array;
	my @output = ($last);

	while (my $pres = shift @sorted_array) {

	    ##($last eq $pres) and next;
	    my $out_comp = compare_arrays (@{$pres}, @{$last}, @{$indices}, @{$methods});
	    if (!(defined $out_comp) ) {
		print "'compare_arrays' failed as a subroutine of 'nr_array_of_arrays'.\n";
		return undef;
	    }
	    ($out_comp ==  0) and next;

	    $last = $pres;
	    push @output, $pres;
	}

	return \@output;
}				# end of the subroutine nr_array_of_arrays #


# binsearch_array:
#	This subroutine searches a target set of arrays (@set_arrays)
#	for a query array (@query) using a binary search algorithm.
#
# INPUT: (\@query, \@set_arrays, $indx_array, \@indices_relv, \@indices_cmp, \@methods_cmp), where
#	@query is the array to be searched for;
#	$set_arrays[$k]->[$indx_array] if $indx_array is defined,
#	otherwise $set_arrays[$k] itself, is the $k th array to be compared with the query;
#	@indices_rlv gives indices in a limited set to which the search will be narrowd down;
#	@indices_cmp and @methods_cmp are used to compare the arrays in the set with the query array.
#
#	PREREQUISITE: Elements in @set_arrays must be pre-sorted according to @indices_cmp and @methods_cmp,
#	and @indices_relv must be sorted in numerically ascending order.
#
# OUTPUT: ($fl_exist, $indx2_qu, $indx_qu), where
#	$fl_exist = 1 if the query exists in the set,
#		  = 0 otherwise;
#	$indx_qu = $indices_relv[$indx2_qu] gives the index of the query
#	in @set_arrays (if $fl_exist = 1),
#	the index of the left neighbor of the query in the relevant subset of @set_array (if $fl_exist = 0)
#	or -1 if the query is 'smaller' than any arrays in @set_array[@indices_relv].
#
#	THe subroutine returns (undef, undef, undef) if the process failed.
#
sub binsearch_array (\@\@$\@\@\@) {

	my ($query, $set_arrays, $indx_array, $indices_relv, $indices_cmp, $methods_cmp) = @_;
	my $ct_arrays = @{$set_arrays};
	unless ($ct_arrays>0) {
	    print STDERR "[ERROR 'binsearch_array'] ILLEGAL INPUT: NO ARRAYS in the set to be searched: $! !\n";
	    return (undef, undef, undef);
	}
	my $ct_rlv = @{$indices_relv};
	unless ($ct_rlv <= $ct_arrays) {
	    print STDERR "[ERROR 'binsearch_array'] ILLEGAL INPUT: The number of relevant entries ($ct_rlv) EXCEEDS the number of arrays ($ct_arrays): $! !\n";
	    return (undef, undef, undef);
	}

		# Compare the leftmost entry with the query #
	my $indx2_left = 0;
	my $indx_left = $indices_relv->[$indx2_left];
	my $entry_left = $set_arrays->[$indx_left];
	my $array_left = (defined $indx_array) ? $entry_left->[$indx_array] : $entry_left;
	my $sgn_left = compare_arrays (@{$query}, @{$array_left}, @{$indices_cmp}, @{$methods_cmp});
	unless (defined $sgn_left) {
	    print STDERR "[ERROR 'binsearch_array'] The subroutine 'compare_arrays' FAILED to compare the query array with the $indx_left th array in the set: $! !\n";
	    return (undef, undef, undef);
	}
	if ($sgn_left < 0) { # The query is smaller than the smallest relevant array #
		return (0, -1, -1);
	} elsif ($sgn_left > 0) { # The query is larger than the smallest entry #
	    if ($ct_rlv == 1) { # Only the smallest entry is in the set of relevant arrays
		return (0, $indx2_left, $indx_left);
	    }
	} else { # The query equals the smallest array #
	    return (1, $indx2_left, $indx_left);
	}

		# Compare the rightmost entry with the query #
	my $indx2_right = $ct_rlv - 1;
	my $indx_right = $indices_relv->[$indx2_right];
	my $entry_right = $set_arrays->[$indx_right];
	my $array_right = (defined $indx_array) ? $entry_right->[$indx_array] : $entry_right;
	my $sgn_right = compare_arrays (@{$query}, @{$array_right}, @{$indices_cmp}, @{$methods_cmp});
	unless (defined $sgn_right) {
	    print STDERR "[ERROR 'binsearch_array'] The subroutine 'compare_arrays' FAILED to compare the query array with the $indx_right th array in the target set: $! !\n";
	    return (undef, undef, undef);
	}
	if ($sgn_right > 0) { # The query is larger than the largest relevant array #
	    return (0, $indx2_right, $indx_right);

	} elsif ($sgn_right == 0) { # The query equals the largest array #
	    return (1, $indx2_right, $indx_right);
	}

		# Now, the query should be between the smallest and the largest relevant arrays #

		## Main Process:  Binary search ##

	while ($indx2_left + 1 < $indx2_right) { # Main while-loop # continues as long as there are indices in between $indx2_left and $indx2_right

		# Compare the middle entry with the query #

	    my $indx2_mid = int (($indx2_left + $indx2_right)/2);
	    my $indx_mid = $indices_relv->[$indx2_mid];
	    my $entry_mid = $set_arrays->[$indx_mid];
	    my $array_mid = (defined $indx_array) ? $entry_mid->[$indx_array] : $entry_mid;
	    my $sgn_mid = compare_arrays (@{$query}, @{$array_mid}, @{$indices_cmp}, @{$methods_cmp});
	    unless (defined $sgn_mid) {
		print STDERR "[ERROR 'binsearch_array'] The subroutine 'compare_arrays' FAILED to compare the query array with the $indx_mid th array in the terget set: $! !\n";
		return (undef, undef, undef);
	    }

	    if ($sgn_mid < 0) { # The query is smaller than the middle array #
		$indx2_right = $indx2_mid; # Narrow the search down to the left half

	    } elsif ($sgn_mid > 0) { # The query is larger than the middle array # 
		$indx2_left = $indx2_mid; # Narraw the search down to the right half

	    } else { # The query equals the middle array #
		return (1, $indx2_mid, $indx_mid);
	    }
	}			# End of the main while-loop #

	$indx_left = $indices_relv->[$indx2_left];

	return (0, $indx2_left, $indx_left); # Found no relevant array that is equal to the query #

}	# End of the subroutine 'binsearch_array' #


# binsearch_string:
#	This subroutine searches a target set of strings @set_strings
#	for a query string $query using a binary search algorithm.
#	It was created by slightly modifying the subroutine 'binsearch_array'.
#	(CREATED on Mar 9 (Wed), 2011)
#
# binsearch_array:
#	This subroutine searches a target set of arrays (@set_arrays)
#	for a query array (@query) using a binary search algorithm.
#
# INPUT: (\$query, \@set_strings, $indx_string, \@indices_relv), where
#	$query is the string to be searched for;
#	${$set_strings[$k]->[$indx_array]} if $indx_array is defined,
#	otherwise ${$set_arrays[$k]}, is the $k th string to be compared with the query;
#	@indices_rlv gives indices in a limited set to which the search will be narrowd down.
#
#	PREREQUISITE: Elements in @set_strings must be pre-sorted in lexically ascending order (using e.g. cmp),
#	and @indices_relv must be sorted in numerically ascending order.
#
# OUTPUT: ($fl_exist, $indx2_qu, $indx_qu), where
#	$fl_exist = 1 if the query string exists in the set,
#		  = 0 otherwise;
#	$indx_qu = $indices_relv[$indx2_qu] gives the index of the query
#	in @set_strings (if $fl_exist = 1),
#	the index of the left neighbor of the query in the relevant subset of @set_strings (if $fl_exist = 0),
#	or -1 if the query is 'smaller' than any arrays in @set_array[@indices_relv] (with $fl_exist = 0 also in this case).
#
#	THe subroutine returns (undef, undef, undef) if the process failed.
#
sub binsearch_string (\$\@$\@) {

	my ($query, $set_strings, $indx_string, $indices_relv) = @_;
	my $ct_strings = @{$set_strings};
	unless ($ct_strings>0) {
	    print STDERR "[ERROR 'binsearch_string'] ILLEGAL INPUT: NO STRINGS in the set to be searched: $! !\n";
	    return (undef, undef, undef);
	}
	my $ct_rlv = @{$indices_relv};
	unless ($ct_rlv <= $ct_strings) {
	    print STDERR "[ERROR 'binsearch_string'] ILLEGAL INPUT: The number of relevant entries ($ct_rlv) EXCEEDS the number of strings ($ct_strings): $! !\n";
	    return (undef, undef, undef);
	}

		# Compare the leftmost entry with the query #
	my $indx2_left = 0;
	my $indx_left = $indices_relv->[$indx2_left];
	my $entry_left = $set_strings->[$indx_left];
	my $string_left = (defined $indx_string) ? $entry_left->[$indx_string] : $entry_left;

	my $sgn_left = ${$query} cmp ${$string_left};
	##my $sgn_left = ${query} cmp ${$string_left}; # A BUG fixed on Mar 10 (Thu), 2011 #
	unless (defined $sgn_left) {
	    print STDERR "[ERROR 'binsearch_string'] The operator 'cmp' FAILED to compare the query array with the $indx_left th array in the set: $! !\n";
	    return (undef, undef, undef);
	}
	if ($sgn_left < 0) { # The query is smaller than the smallest relevant array #
		return (0, -1, -1);
	} elsif ($sgn_left > 0) { # The query is larger than the smallest entry #
	    if ($ct_rlv == 1) { # Only the smallest entry is in the set of relevant arrays
		return (0, $indx2_left, $indx_left);
	    }
	} else { # The query equals the smallest array #
	    return (1, $indx2_left, $indx_left);
	}

		# Compare the rightmost entry with the query #
	my $indx2_right = $ct_rlv - 1;
	my $indx_right = $indices_relv->[$indx2_right];
	my $entry_right = $set_strings->[$indx_right];
	my $string_right = (defined $indx_string) ? $entry_right->[$indx_string] : $entry_right;
	my $sgn_right = ${$query} cmp ${$string_right};
	unless (defined $sgn_right) {
	    print STDERR "[ERROR 'binsearch_string'] The operator 'cmp' FAILED to compare the query array with the $indx_right th array in the target set: $! !\n";
	    return (undef, undef, undef);
	}
	if ($sgn_right > 0) { # The query is larger than the largest relevant array #
	    return (0, $indx2_right, $indx_right);

	} elsif ($sgn_right == 0) { # The query equals the largest array #
	    return (1, $indx2_right, $indx_right);
	}

		# Now, the query should be between the smallest and the largest relevant arrays #

		## Main Process:  Binary search ##

	while ($indx2_left + 1 < $indx2_right) { # Main while-loop # continues as long as there are indices in between $indx2_left and $indx2_right

		# Compare the middle entry with the query #

	    my $indx2_mid = int (($indx2_left + $indx2_right)/2);
	    my $indx_mid = $indices_relv->[$indx2_mid];
	    my $entry_mid = $set_strings->[$indx_mid];
	    my $string_mid = (defined $indx_string) ? $entry_mid->[$indx_string] : $entry_mid;
	    my $sgn_mid = ${$query} cmp ${$string_mid};
	    unless (defined $sgn_mid) {
		print STDERR "[ERROR 'binsearch_string'] The subroutine 'cmp' FAILED to compare the query array with the $indx_mid th array in the terget set: $! !\n";
		return (undef, undef, undef);
	    }

	    if ($sgn_mid < 0) { # The query is smaller than the middle array #
		$indx2_right = $indx2_mid; # Narrow the search down to the left half

	    } elsif ($sgn_mid > 0) { # The query is larger than the middle array # 
		$indx2_left = $indx2_mid; # Narraw the search down to the right half

	    } else { # The query equals the middle array #
		return (1, $indx2_mid, $indx_mid);
	    }
	}			# End of the main while-loop #

	$indx_left = $indices_relv->[$indx2_left];

	return (0, $indx2_left, $indx_left); # Found no relevant array that is equal to the query #

}	# End of the subroutine 'binsearch_string' #


# sort_mixed_elements:
#	This subroutine sorts array elements 'fairly correctly'
#	even when they are a mixture of numerals and other characters.
#
sub sort_mixed_elements (\@) {

	my ($array) = @_;

	my $ct_elements = @{$array};

	($ct_elements < 2) and (return $array);

		# The 1st preliminary process #
	my $max_len = 0;
	for (my $i=0; $i< $ct_elements; $i++) {
	   my $len = length $array->[$i];
	   ($max_len < $len) and ($max_len = $len); 
	}

		# The 2nd preliminary process #
	my @aux_array = (); # element = [original element, transformed element]
	for (my $i=0; $i<$ct_elements; $i++) {
	    my $element = $array->[$i];
	    if ($element =~ m/^\d/ ) { # begins with a numeral #
		my $tmp = '0' x ($max_len - length $element) . $element;
		push @aux_array , [$element, $tmp];
	    } else {		# otherwise #
		push @aux_array, [$element, $element];
	    }
	}

		# The main process #
	my @std_aux_array =  sort {$a->[1] cmp $b->[1] } @aux_array;

		# Finish #
	my @std_array = ();
	for (my $i=0; $i<$ct_elements; $i++)  {
	    push @std_array , $std_aux_array[$i]->[0];
	}

	return \@std_array;

}			# end of the subroutine sort_mixed_elements #


# merge_intervals:
#	This subroutine receives a set of intervals as an array of arrays,
#	then constructs a set of NON-OVERLAPED merged intervals.
#
# INPUT: (\@intervals, $indx_beg, $indx_end), where
#	@{$intervals[m-1]}[$indx_beg, $indx_end] = ($begin, $end)
#	gives the boundaries of the m-th interval.
#
# OUTPUT: \@merged, where
#	@merged = ([$begin, $end], ...),
#	which are arranged in the ascending order of $begin.
#
sub merge_intervals (\@$$) {

	my ($intervals, $indx_beg, $indx_end) = @_;

	(@{$intervals} > 0) or (return []); 

	my @indices = ($indx_beg, $indx_end); my @methods = ('+', '-');

	my $std_intervals  = (@{$intervals} < 2) ? $intervals :
 		multiple_sort (@{$intervals}, @indices, @methods) ;
	my @reservoir = @{$std_intervals};

	my @merged = ();

	while (my $last = shift @reservoir) {

	    my $size_reservoir = @reservoir;

	    my ($last_beg, $last_end) = @{$last}[$indx_beg, $indx_end];
	    my $orig_last_end = $last_end;
	    ##my @leftover = ();

	    if ( ($size_reservoir > 0) and ($last_end < $reservoir[0]->[$indx_beg]) ) {
		  # There should be no overlap between the 'last' interval and the rest
		push @merged, [$last_beg, $last_end];
		next;
	    }

		# There are some overlaps #
	    while (my $pres = shift @reservoir) {

		my ($pres_beg, $pres_end) = @{$pres}[$indx_beg, $indx_end];

		if ($pres_beg < $last_beg) {
		    print STDERR "It's IMPOSSIBLE! the present interval \[$pres_beg, $pres_end\] comes to the left of the last interval \[$last_beg, $last_end\] !\n";
		    return undef;
		} elsif ($pres_beg <= $last_end) { # overlap between the intervals #
		    ($last_end < $pres_end) and ($last_end = $pres_end);
		} else { # no overlap #
		    unshift @reservoir, $pres;
		    last;
		}
	    }

	    push @merged, [$last_beg, $last_end];
	}

	return \@merged;

}		# end of the subroutine merge_intervals #


# merge_integer_intervals:
#	This subroutine receives a set of intervals as an array of arrays,
#	then constructs a set of NON-OVERLAPED merged intervals.
#	It works only within an integer coordinate system ,
#	so it also merges 2 contiguous intervals, like
#	[1,10][11,20], to form a smoothened interval, like [1,20].
#
# INPUT: (\@intervals, $indx_beg, $indx_end), where
#	@{$intervals[m-1]}[$indx_beg, $indx_end] = ($begin, $end)
#	gives the boundaries of the m-th interval.
#
# OUTPUT: \@merged, where
#	@merged = ([$begin, $end], ...),
#	which are arranged in the ascending order of $begin.
#
sub merge_integer_intervals (\@$$) {

	my ($intervals, $indx_beg, $indx_end) = @_;

	(@{$intervals} > 0) or (return []); 

	my @indices = ($indx_beg, $indx_end); my @methods = ('+', '-');

	my $std_intervals  = (@{$intervals} < 2) ? $intervals :
 		multiple_sort (@{$intervals}, @indices, @methods) ;
	my @reservoir = @{$std_intervals};

	my @merged = ();

	while (my $last = shift @reservoir) {

	    my $size_reservoir = @reservoir;

	    my ($last_beg, $last_end) = @{$last}[$indx_beg, $indx_end];
	    my $orig_last_end = $last_end;
	    ##my @leftover = ();

	    if ( ($size_reservoir > 0) and ($last_end +1 < $reservoir[0]->[$indx_beg]) ) {
		  # There should be no overlap or contact
		  # between the 'last' interval and the rest
		push @merged, [$last_beg, $last_end];
		next;
	    }

		# There are some overlaps #
	    while (my $pres = shift @reservoir) {

		my ($pres_beg, $pres_end) = @{$pres}[$indx_beg, $indx_end];

		if ($pres_beg < $last_beg) {
		    print STDERR "It's IMPOSSIBLE! the present interval \[$pres_beg, $pres_end\] comes to the left of the last interval \[$last_beg, $last_end\] !\n";
		    return undef;
		} elsif ($pres_beg <= $last_end +1) { # overlap or contact between the intervals #
		    ($last_end < $pres_end) and ($last_end = $pres_end);
		} else { # no overlap or contact #
		    unshift @reservoir, $pres;
		    last;
		}
	    }

	    push @merged, [$last_beg, $last_end];
	}

	return \@merged;

}		# end of the subroutine merge_integer_intervals #


# compare_interval_sets:
#	This subroutiine compares 2 sets of intervals.
#
# INPUT: (\@intervals1, \@intervals2), where
#	@{$intervals1/2[$m-1]} = ($beg, $end) are boundaries of the m-th interval.
#
# OUTPUT: (\@shared, \@ext1, \@ext2), which respectively store
#	shared intervals, intervals excessive in @intervals1,
#	and those excessive in @intervals2.
#
# PREREQUISITE: Each input set has to contain
#		intervals which DO NOT OVERLAP with one another.
#
sub compare_interval_sets (\@\@) {

	my ($intervals1, $intervals2) = @_;

	my @std_intervals1 = sort {$a->[0] <=> $b->[0] } @{$intervals1};
	my @std_intervals2 = sort {$a->[0] <=> $b->[0] } @{$intervals2};
	my @shared = my @ext1 = my @ext2 = ();

	while ( (@std_intervals1 > 0) and (@std_intervals2 > 0) ) {
			# the outermost while-loop #

	    my $int1 = shift @std_intervals1;
	    my $int2 = shift @std_intervals2;
	    my ($beg1, $end1) = @{$int1};
	    my ($beg2, $end2) = @{$int2};

	    if ($beg1 < $beg2) {	# the outer if-elsif-else #

		if ($end1 < $beg2) {
		    push @ext1, [$beg1, $end1];
		    unshift @std_intervals2, $int2;
		} elsif ($end1 <= $end2) {
		    push @ext1, [$beg1, $beg2 -1]; 
		    push @shared, [$beg2, $end1];
		    ($end1 == $end2) or (unshift @std_intervals2, [$end1+1, $end2]);
		} else { # $end1 > $end2
		   push @ext1, [$beg1, $beg2-1];
		   push @shared, [$beg2, $end2];
		   unshift @std_intervals1, [$end2+1, $end1];
		}

	    } elsif ($beg2 < $beg1) {

		if ($end2 < $beg1) {
		    push @ext2, [$beg2, $end2];
		    unshift @std_intervals1, $int1;
		} elsif ($end2 <= $end1) {
		   push @ext2, [$beg2, $beg1-1];
		   push @shared, [$beg1, $end2];
		   ($end2 == $end1) or (unshift @std_intervals1, [$end2+1, $end1]);
		} else { # $end2 > $end1
		   push @ext2, [$beg2, $beg1-1];
		   push @shared, [$beg1, $end1];
		   unshift @std_intervals2, [$end1+1, $end2];
		}

	    } else {  # $beg1 == $beg2

		if ($end1 < $end2) {
		   push @shared, [$beg1, $end1];
		   unshift @std_intervals2, [$end1+1, $end2];
		} elsif ($end2 < $end1) {
		   push @shared, [$beg2, $end2];
		   unshift @std_intervals1, [$end2+1, $end1];
		} else { # $end2 = $end1
		   push @shared, [$beg1, $end1];
		}
	    }		# end of the outer if-elsif-else #
	}			# end of the outermost while-loop #

	(@std_intervals1 > 0) and (push @ext1, @std_intervals1);
	(@std_intervals2 > 0) and (push @ext2, @std_intervals2);

	return (\@shared, \@ext1, \@ext2) ;

}		# end of the subroutine compare_interval_sets #

# pos_consist_intvls:
#	This subroutine examines whether 2 sets of interval are consistent or not,
#	according to the following criteria:
#	(i) The intervals, one from a set and the other from the other set,
#	 that have the same ordering number must agree with each other completely.
#	Namely, $beg1 == $beg2 AND $end1 == $end2 if $order1 == $order2;
#	(ii) The intervals with different ordering numbers must NOT overlap with
#	each other, and the coordinates must be consistent with the ordering number.
#	Namely, $end1 < $beg2 if $order1 < $order2.
#
#	The intervals in each set are supposed to be PRE-SORTED in ascending order
#	of the ordering number, which is consistent with the ascending order
#	of the beginning coordinate.
#
# INPUT: (\@intervals1, \@intervals2, \@indices_pos, $index_order), where
#	($beg_k, $end_k) = @{$intervals1/2[$k]}[@indices_pos] gives
#	the beginning and ending positions of the $k th interval in the 1st/2nd set;
#	$index_order specifies where the odering number of each interval is stored.
#	If it is negative, the subroutine regards the two sets as consistent only if
#	there is NO overlap between them.
#
# OUTPUT: 1 if the sets are consistent,
#	  0 otherwise.
#
sub pos_consist_intvls (\@\@\@$) {

	my ($intvls1, $intvls2, $indices_pos, $indx_order) = @_;

	($indx_order < 0) and ($indx_order = $indices_pos->[0]);

	my $ct_int1 = @{$intvls1};
	my $ct_int2 = @{$intvls2};
	my $i1= my $i2 = 0;

	while (($i1<$ct_int1) and ($i2<$ct_int2)) {
	   my ($order1, $beg1, $end1) = @{$intvls1->[$i1]}[$indx_order, @{$indices_pos}];
	   my ($order2, $beg2, $end2) = @{$intvls2->[$i2]}[$indx_order, @{$indices_pos}];

	   if ($order1 < $order2) {
		($end1 < $beg2) or (return 0);
		$i1++;  # examine the next interval in the set 1, while keeping the set2 intact.
	   } elsif ($order2 < $order1) {
		($end2 < $beg1) or (return 0);
		$i2++; # go to the next interval in the set 2, while staying at the same interval in the set2.
	   } else {
		(($beg1 == $beg2) and ($end1 == $end2)) or (return 0);
		   # move to the next interval in either set.
		$i1++;
		$i2++;
	   }
	}

	   # Reaching here indicates that the two sets are consistent with each other.
	return 1;

}		# end of the subroutine 'pos_consist_intvls' #


# total_length_intvls:
#	This subroutine recieves a set of intervals,
#	and calculates their total length.
#
# INPUT: \@intvl_set, where
#	@{$intvl_set[$m]} = ($beg, $end} gives the boundary coordinates
#	of the $m th interval (in terms of an integer frame).
#
# OUTPUT: $total_length,
#	which is just the summation of ($end - $beg + 1).
#	It does NOT CARE whether there are overlaps between the intervals,
#	or whether some intervals are inverted, i.e. $end < $beg.
#
sub total_length_intvls (\@) {

	my ($intvl_set) = @_;

	my $total_length = 0;
	for my $intvl (@{$intvl_set}) {
	    my ($beg, $end) = @{$intvl};

	    unless (defined $end) { # A measure for debugging
		print STDERR "ILLEGAL INPUT FOR THE INTERVALS:\n";
		for (my $k=0; $k<@{$intvl_set}; $k++) {
		   print STDERR "Interval $k:   [" . join (',', @{$intvl_set->[$k]}) . "]\n";
		}
		print STDERR  "\n";
		return undef;
	    }		# End of the measure for debugging #

	    $total_length += $end - $beg + 1;
	}

	return $total_length;

}		# end of the subroutine 'total_length_intvls' #


# max_length_intvls:
#	This subroutine recieves a set of intervals,
#	and calculates their maximum length.
#
# INPUT: \@intvl_set, where
#	@{$intvl_set[$m]} = ($beg, $end} gives the boundary coordinates
#	of the $m th interval (in terms of an integer frame).
#
# OUTPUT: ($indx_max, $max_length), where
#	$max_length is just the maximum among ($end - $beg + 1) for the intervals,
#	and $indx_max is the corresponding index.
#	It does NOT CARE whether there are overlaps between the intervals,
#	or whether some intervals are inverted, i.e. $end < $beg.
#
sub max_length_intvls (\@) {

	my ($intvl_set) = @_;

	my $ct = @{$intvl_set};
	my $max_length = 0;
	my $indx_max = -1;
	for (my $i=0; $i<$ct; $i++) {
	    my ($beg, $end) = @{$intvl_set->[$i]};

	    unless (defined $end) { # A measure for debugging
		print STDERR "ILLEGAL INPUT FOR THE INTERVALS:\n";
		for (my $k=0; $k<@{$intvl_set}; $k++) {
		   print STDERR "Interval $k:   [" . join (',', @{$intvl_set->[$k]}) . "]\n";
		}
		print STDERR  "\n";
		return undef;
	    }		# End of the debugging measure #

	    my $length = $end - $beg + 1;
	    ($max_length < $length) and (($indx_max, $max_length) = ($i, $length));
	}

	return ($indx_max, $max_length);

}		# end of the subroutine 'max_length_intvls' #


# linsearch_closest_intervals:
#	This subroutine examines spatial relationships
#	of an interval and a set of intervals
#	in order to look for intervals
#	that are the closest to the former among the latter set.
#
# PREREQUISITE:
#	The subject intervals are pre-sorted in ascending order of the beginning position.
#	And the subject intervals do NOT overlap with each other.
#
# INPUT: ($qu_beg, $qu_end, \@sb_intvl_set, \@indices_pos, $indx_id), where
#	($sb_beg, $sb_end) = @{$sb_intvl_set[$k]}[@indices_pos]
#	for the $k th interval in the subject set;
#	$sb_intvl_set[$k]->[$indx_id]  (or $k if $indx_id <0)
#	is used as an ID of the subject interval.
#
# OUTPUT: ($minleft_dist, $minright_dist, $id_left_closest, $id_right_closest, \@ids_included), where
#	$minleft_dist = {distance}  if the left-closest interval is separated from the query,
#		= - {size_overlap} if it overlaps the query,
#		= undef (along with $id_left_closest = undef)
#		if there is no 'left-closest' interval;
#	$minright_dist follows the same rule;
#	@ids_included contains IDs of the subject intervals that are completely
#	included in the query interval.
#	We DEFINE the 'left-closest' interval as the interval with the smallest
#	distance among those which begin at or to the left of
#	the beginning position of the query;
#	Similarly, the 'right-closest' interval is defined as that with the smallest
#	distance among those which end at or to the right of the query's ending position.
#
sub linsearch_closest_intervals ($$\@\@$) {

	my ($qu_beg, $qu_end, $sb_intvl_set, $indices_pos, $indx_id) = @_;

	my $ct_sbs = @{$sb_intvl_set};
	($ct_sbs>0) or (return (undef,undef, undef, undef, []));

	my ($leftmost_beg, $rightmost_end) = @{$sb_intvl_set->[0]}[@{$indices_pos}];
	my $id_leftmost = ($indx_id < 0) ? 0 : $sb_intvl_set->[0][$indx_id];

	if ($qu_end <= $leftmost_beg) {
		# There is NEITHER 'left-closest' intervals NOR included ones #
	    return (undef, $leftmost_beg - $qu_end - 1, undef, $id_leftmost, []);
	}

	my ($id_lclosest, $min_ldist, $id_rclosest, $min_rdist);
	my @included = ();

	for (my $k=0; $k<$ct_sbs; $k++) { # the outermost for-loop #

 	    my ($sb_beg, $sb_end) = @{$sb_intvl_set->[$k]}[@{$indices_pos}];
	    my $id_sb = ($indx_id < 0) ? $k : $sb_intvl_set->[$k][$indx_id];

	   if ($qu_end <= $sb_beg) {

		my $dist = $sb_beg - $qu_end - 1;
		(!(defined $min_rdist) or ($dist < $min_rdist) ) and 
		   ( ($id_rclosest, $min_rdist) = ($id_sb, $dist) ) ; 

		last; # The rest will no longer give the 'closest' intervals. 

	    } elsif ($qu_beg >= $sb_end) {

		my $dist = $qu_beg - $sb_end - 1;
		(!(defined $min_ldist) or ($dist < $min_ldist) ) and
		   ( ($id_lclosest, $min_ldist) = ($id_sb, $dist) ) ;

	    } elsif ($qu_end <= $sb_end) {
		if ($qu_beg >= $sb_beg) { # The subject completely embraces the query
		    $min_ldist = $min_rdist = $qu_beg - $qu_end - 1;
		    $id_lclosest = $id_rclosest = $id_sb;

		    last; # This is the 'closest' possible interval #

		} else { # The subject is 'to the right'

		    my $dist = $sb_beg - $qu_end - 1;
		    (!(defined $min_rdist) or ($dist <$min_rdist) ) and
			( ($id_rclosest, $min_rdist) = ($id_sb, $dist) );
		}
	    } elsif ($qu_beg >= $sb_beg) { # The subject is 'to the left'

		my $dist = $qu_beg - $sb_end - 1;
		(!(defined $min_ldist) or ($dist <$min_ldist) ) and
			( ($id_lclosest, $min_ldist) = ($id_sb, $dist) ) ;

	    } else {  # The subject is completely included in the query 
		push @included, $id_sb;
	    }
	}			# end of the outermost for-loop #

	return ($min_ldist, $min_rdist, $id_lclosest, $id_rclosest, \@included);

}		# end of the subroutine 'linsearch_closest_intervals' #


# linsearch_ovlp_intervals:
#	This subroutine is slightly different from
#	the above 'linsearch_closest_intervals'
#	in that it focuses on whether the subject intervals
#	overlap the query or not.
#
# PREREQUISITE:
#	The subject intervals are pre-sorted in ascending order of the beginning position.
#	And the subject intervals do NOT overlap with each other.
#
# INPUT: ($qu_beg, $qu_end, \@sb_intvl_set, \@indices_pos, $indx_id), where
#	($sb_beg, $sb_end) = @{$sb_intvl_set[$k]}[@indices_pos]
#	for the $k th interval in the subject set;
#	$sb_intvl_set[$k]->[$indx_id]  (or $k if $indx_id <0)
#	is used as an ID of the subject interval.
#
# OUTPUT: (\@ovlp_intvls, $left_beg, $left_end, $right_beg, $right_end), where
#	@ovlp_intvls is a list of IDs of the subject intervals
#	that overlap, or are completely included in, the query interval;
#	$left/right_beg/end is the beginning/ending coordinate of
#	the leftmost/rightmost overlapping subject interval
#	($left/right_beg/end = undef if there is no overlapping interval.)
#
sub linsearch_ovlp_intervals ($$\@\@$) {

	my ($qu_beg, $qu_end, $sb_intvl_set, $indices_pos, $indx_id) = @_;

	my $ct_sbs = @{$sb_intvl_set};
	($ct_sbs>0) or (return ([], undef, undef, undef, undef));

	my ($indx_beg, $indx_end) = @{$indices_pos};

	if ($qu_end < $sb_intvl_set->[0][$indx_beg]) {
		# All subject intervals are to the right of the query #
	    return ([], undef, undef, undef, undef);
	} elsif ($qu_beg > $sb_intvl_set->[$ct_sbs-1][$indx_end]) {
		# All subject intervals are to the left of the query #
	    return ([], undef, undef, undef, undef);
	}

	my $k =0;

	while ($k <$ct_sbs) {  # The 1st outer while-loop #
	    ($qu_beg <= $sb_intvl_set->[$k][$indx_end]) and last;
	    $k++;
	}		# End of the 1st outer while-loop #

	my $indx_left = $k; # For the leftmost interval

	my @ovlp_intvls = ();

	my $indx_right = $indx_left; # For the rightmost interval

	while ($k<$ct_sbs) { # The 2nd outer while-loop #
	    ($qu_end < $sb_intvl_set->[$k][$indx_beg]) and last;
	    push @ovlp_intvls, (($indx_id<0) ? $k : $sb_intvl_set->[$k][$indx_id]);
	    $indx_right = $k;
	    $k++;
	}		# End of the 2nd outer while-loop #


	my ($left_beg, $left_end) = @{$sb_intvl_set->[$indx_left]}[@{$indices_pos}];
	my ($right_beg, $right_end) = @{$sb_intvl_set->[$indx_right]}[@{$indices_pos}];

	return (\@ovlp_intvls, $left_beg, $left_end, $right_beg, $right_end) ;

} 		# end of the subroutine 'linsearch_ovlp_intervals' #


# find_max_elements:
#	This satellite subroutine of 'set_max_nonovlp_regions' examines an array,
#	and looks for elements with the maximum value of a specified quantity.
#
# INPUT: (\@array, $indx_qt),  where
#	$array[$m]->[$indx_qt] is the compared quantity of the $m th element if $indx_qt >=0,
#	$array[$m] is such quantity if $indx_qt < 0.
#
# OUTPUT: (\@max_indices, $max_qt), where
#	$max_qt is the maximum value;
#	@max_indices  stores indices of the elements with the maximum value.
#
sub find_max_elements (\@$) {

	my ($array, $indx_qt) = @_;

=for DEBUGGING the script
	print STDERR "\n< Subroutine 'find_max_elements' >\n\n";
	print STDERR "Index = $indx_qt \n\n";
	print STDERR "{elements}\n";
	foreach my $element (@{$array}) {
	   print STDERR (($indx_qt<0) ? $element : join ("\t", @{$element})) , "\n";
	}
	print STDERR "\n";
=cut

	my $size_array = @{$array};
	($size_array>0) or (return ([], undef) );

	my $max_qt = ($indx_qt <0) ? $array->[0] : $array->[0][$indx_qt];
	my @max_indices = (0);

	for (my $m=1; $m<$size_array; $m++) { # The main for-loop #
	    my $element = $array->[$m];
	    my $qt = ($indx_qt <0) ? $element : $element->[$indx_qt];
	    if ($qt > $max_qt) {
			# Reset the set of maximum elements #
		@max_indices = ($m);
		$max_qt = $qt;
	    } elsif ($qt == $max_qt) {
		push @max_indices, $m;
	    }
	}		# End of the main for-loop #

=for DEBUGGING the script
	print STDERR "Maximum value = $max_qt\n";
	print STDERR "Maximum indices = ", join (', ', @max_indices), "\n\n";
=cut

	return (\@max_indices, $max_qt);

}	# End of the subroutine 'find_max_elements' #


# incorporate_new_region:
#	This satellite subroutine of 'set_max_nonovlp_regions'
#	tries to incorporate the $k region into the sets of non-overlapping regions.
#	The two regions are regarded as 'overlapping' if the overlap is larger than
#	min [$thrsh_size, min (size_region1, size_region2) * $thrsh_prpt /100].
#
# INPUT: (\@set_novlp_sets, $k, \@set_regions, $thrsh_size, $thrsh_prpt), where
#	@{$set_novlp_sets[$n]} = (\@indices_regions, $total_score, $end_set, $last_size)
#	   for the $n th non-overlapping set;
#	$k is the index of the region to be incorporated;
#	@{$set_regions[$k]} = ($begin, $end, $score, ...) for the $k th region.
#	
# OUTPUT: \@new_set_novlp_sets,
#	which only stores the non-overlapping sets that incorporated the $k th region.
#
sub incorporate_new_region (\@$\@$$) {

	my ($set_novlp_sets, $k, $set_regions, $thrsh_size, $thrsh_prpt) = @_;

=for DEBUGGING the script
	print STDERR "\n< Subroutine 'incorporate_new_region' >\n\n";
	print STDERR "{Currently I am examining the $k th set}\n\n";
	printf STDERR "Begin = %d, End = %d, Score = %f\n\n", @{$set_regions->[$k]}[0,1,2];
	print STDERR "{Input non_overlapping regions}\n\n";
	print STDERR join ("\t", 'Indices', 'TScore', 'End_set', 'Last_size'), "\n";
	foreach my $novlp_set (@{$set_novlp_sets}) { print STDERR join (',', @{$novlp_set->[0]}) , "\t", join ("\t", @{$novlp_set}[1,2,3]), "\n"; }
	print STDERR "\n";
=cut

	my ($beg, $end, $score) = @{$set_regions->[$k]}[0,1,2];
	my $size = $end - $beg + 1;
	
	my @new_set_novlp_sets = ();
	foreach my $set (@{$set_novlp_sets}) { # The main foreach-loop #

	    my ($indices_regions, $tot_score, $end_set, $last_size) = @{$set};
	    my $min_size = ($last_size < $size) ? $last_size : $size;
	    my $thrsh = (100 * $thrsh_size < $thrsh_prpt * $min_size) ?
	    	$thrsh_size : $thrsh_prpt * $min_size /100;
	    my $prev_indx = $indices_regions->[$#{$indices_regions}];
	    my ($prev_beg, $prev_end) = @{$set_regions->[$prev_indx]}[0,1];
	    my $ovlp_beg = ($beg > $prev_beg) ? $beg : $prev_beg;
	    my $ovlp_end = ($end < $prev_end) ? $end : $prev_end;
	    ($ovlp_end - $ovlp_beg +1 > $thrsh) and next; # Skip overlapping region #

	    	# Incorporate the current region #
	    push @{$indices_regions}, $k;
	    $set->[1] += $score;
	    ($end_set < $end) and ($set->[2] = $end);
	    $set->[3] = $size;
	    push @new_set_novlp_sets, $set;
	}				# End of the main foreach-loop #

	return \@new_set_novlp_sets;

}		# End of the subroutine 'incorporate_new_region' #


# recede_incp_regions:
#	This satellite subroutine of 'set_max_nonovlp_regions'
#	tries to incorporate the $k region into the sets of non-overlapping regions
#	while 'receding' the non-overlapping sets.
#	The two regions are regarded as 'overlapping' if the overlap is larger than
#	min [$thrsh_size, min (size_region1, size_region2) * $thrsh_prpt /100].
#
# INPUT: (\@set_novlp_sets, $k, \@set_regions, $thrsh_size, $thrsh_prpt), where
#	@{$set_novlp_sets[$n]} = (\@indices_regions, $total_score, $end_set, $last_size)
#	   for the $n th non-overlapping set;
#	$k is the index of the region to be incorporated;
#	@{$set_regions[$k]} = ($begin, $end, $score, ...) for the $k th region.
#	
# OUTPUT: \@new_set_novlp_sets,
#	which only stores the non-overlapping sets that incorporated the subject region.
#
sub recede_incp_regions (\@$\@$$) {

	my ($set_novlp_sets, $k, $set_regions, $thrsh_size, $thrsh_prpt) = @_;

		# Initialize the new set #
	my @new_set_novlp_sets = ();
	foreach my $novlp_set (@{$set_novlp_sets}) { # The 1st outer foreach-loop #
	    my @cp_novlp_set = @{$novlp_set};
	    my @cp_rg_indices = @{$novlp_set->[0]};
	    $cp_novlp_set[0] = \@cp_rg_indices;
	    push @new_set_novlp_sets, \@cp_novlp_set;
	}				# End of the 1st outer foreach-loop # 

	while (1) { # The outer while-loop #

		# Recede the non-overlapping sets #

	     my %already; # ( "concatenated:indices:of:regions" => 1, ...)
			  # for the region that already exists. 
	     my @reservoir = ();
	     foreach my $novlp_set (@new_set_novlp_sets) { # The inner foreach-loop #
		my ($indices_rg, $tot_score) = @{$novlp_set}[0,1];
		(@{$indices_rg}>1) or next; # Skip if the set becomes empty #

		my $rm_indx = pop @{$indices_rg};
		my $last_indx = $indices_rg->[$#{$indices_rg}];
		my $concate_indices = join (':', @{$indices_rg});
		(defined $already{$concate_indices}) and next; # Avoid redundancy 
		my $rm_score = $set_regions->[$rm_indx][2];
		my ($last_beg, $last_end) = @{$set_regions->[$last_indx]}[0,1];
		my $last_size = $last_end - $last_beg + 1;
		push @reservoir, [$indices_rg, $tot_score - $rm_score, $last_end, $last_size];
		$already{$concate_indices} = 1;
	     }					# End of the inner foreach-loop #

	     (@reservoir>0) or (return []); # No sets are left any longer #

	     my $new_set_novlp_sets = incorporate_new_region (@reservoir, $k, @{$set_regions}, $thrsh_size, $thrsh_prpt);
	    (@{$new_set_novlp_sets}>0) and (return $new_set_novlp_sets);
	    
	     @new_set_novlp_sets = @reservoir; # Reset the new set #

	}		# End of the outer while-loop #

	print STDERR "[CAUTION 'recede_incp_regions'] An UNEXPECTED SITUATION encountered !!\n\n";

	return [];

}		# End of the subroutine 'recede_incp_regions' #


# set_max_nonovlp_regions:
#	This subroutine receives a set of possible overlapping regions,
#	and returns the set of non-overlapping regions that maximizes
#	the total score.
#	The 2 regions are regarded as 'overlapping' if
#	the size of their overlap is larger than:
#	min [$thrsh_size, min (size_region1, size_region2) * $thrsh_prpt /100].
#
# INPUT: (\@set_regions, $thrsh_size, $thrsh_prpt), where
#	@{$set_regions[$k]} = ($begin, $end, $score, ...) for the $k th region.
#
# OUTPUT: \@maxsets_nr_regions, where
#	@{$maxsets_nr_regions[$k]}, which follows the same format as @set_regions,
#	stores the $k th of the maximum sets of non-overlapping regions.
#
sub set_max_nonovlp_regions (\@$$) {

	my ($set_regions, $thrsh_size, $thrsh_prpt) = @_;

	my $ct_regions = @{$set_regions};
	($ct_regions>0) or (return []);

=for DEBUGGING the script
	print STDERR "\n< Subroutine 'set_max_nonovlp_regions' >\n\n";
	print STDERR "[Input regions]\n";
	print STDERR join ("\t", 'Indx', 'Beg', 'End', 'Score'), "\n";
	for (my $i=0; $i<@{$set_regions}; $i++) {
	    print STDERR join ("\t", $i, @{$set_regions->[$i]}[0,1,2]), "\n";
	}
	print STDERR "\n";
=cut

	my @sort_indices = (0, 1); my @sort_methods = ('+', '+');
	$set_regions = multiple_sort (@{$set_regions}, @sort_indices, @sort_methods);

	my @maxsets_nr_regions = ([]);
		# element = [a list of indices of regions in the maximum set] #

		# Initialize #
	my ($init_beg, $init_end, $init_score) = @{$set_regions->[0]}[0,1,2];
	my $init_size = $init_end - $init_beg + 1;
	my @set_manip_sets = ( [[0], $init_score, $init_end, $init_size] );
		# element = [\@indices_regions, $total_score, $end_set, $last_size];

	my ($prev_rend, $lcl_max_size) = ($init_end, $init_size) ;

	for (my $k=1; $k<$ct_regions; $k++) { # The outermost for-loop #

=for DEBUGGING the script
	    print STDERR "[ The $k th round ]\n\n";
	    print STDERR "{Maximum sets}\n";
	    foreach my $maxset (@maxsets_nr_regions) { print STDERR join (',', @{$maxset}), "\n"; }
	    print STDERR "\n{Input sets}\n";
	    print STDERR join ("\t", 'Indices', 'TScore', 'end', 'last_size'), "\n";
	    for my $man_set (@set_manip_sets) {
		print STDERR join ("\t", join (':', @{$man_set->[0]}), @{$man_set}[1,2,3]) , "\n";
	    }
	    print STDERR "\n";
=cut
	    my $curr_rg = $set_regions->[$k];
	    my ($beg, $end, $score) = @{$curr_rg}[0,1,2];
	    my $size = $end - $beg + 1;
	    my $min_size = ($lcl_max_size < $size)  ? $lcl_max_size : $size;

	    my $ovall_thrsh = (100 * $thrsh_size < $thrsh_prpt * $min_size) ?
			$thrsh_size : $thrsh_prpt * $min_size /100 ;
=for DEBUGGING the script
	    print STDERR "{ Current interval }\n";
	    print STDERR "Begin = $beg, End = $end, Score = $score\n\n";
	    print STDERR "Rightmost end of the manipulable set = $prev_rend\n";
	    print STDERR "Overall threshold = $ovall_thrsh\n\n";
=cut
	    if ($prev_rend - $beg + 1 <= $ovall_thrsh) { # current region does not overlap any of the sets #
			# Only incorporate the best-scoring sets #
		my ($max_indices, $max_score) = find_max_elements (@set_manip_sets, 1);

		my @new_maxsets = ();
		foreach my $maxset (@maxsets_nr_regions) { # The middle foreach-loop #
		    foreach my $max_indx (@{$max_indices}) { # The inner foreach-loop #
			push @new_maxsets, [@{$maxset}, @{$set_manip_sets[$max_indx]->[0]} ];
		    }					 #  End of the inner foreach-loop #
		}				# End of the middle foreach-loop #
		@maxsets_nr_regions = @new_maxsets; # Update the maximum sets

			# Reset the set of manipulable sets of regions #
		@set_manip_sets = ( [[$k], $score, $end, $size] );
		($prev_rend, $lcl_max_size) = ($end, $size) ;
		next;
	    }
	    ($prev_rend < $end) and ($prev_rend = $end);
	    ($lcl_max_size < $size) and ($lcl_max_size = $size);

		# First, try to incorporate the region into intact sets #
	    my $new_set_manip_sets = incorporate_new_region (@set_manip_sets, $k, @{$set_regions}, $thrsh_size, $thrsh_prpt);
	    (@{$new_set_manip_sets}>0) and next; # Some sets have incorporated the region #

		# Second, try to incorporate the region while 'receding the sets #
	    $new_set_manip_sets = recede_incp_regions (@set_manip_sets, $k, @{$set_regions}, $thrsh_size, $thrsh_prpt);

	    if (@{$new_set_manip_sets}>0) {
		push @set_manip_sets, @{$new_set_manip_sets};
		next;
	    }

		# If the region cannot be incorporated by any means #
	    push @set_manip_sets, [[$k], $score, $end, $size];

	}			# End of the outermost for-loop #

=for DEBUGGING the script
	print STDERR "[ After the for-loop ]\n\n";
	print STDERR "{Manipulative sets}\n\n";
	print STDERR join ("\t", 'Indices', 'TScore', 'end', 'last_size'), "\n";
	for my $man_set (@set_manip_sets) {
		print STDERR join ("\t", join (':', @{$man_set->[0]}), @{$man_set}[1,2,3]) , "\n";
	}
	print STDERR "\n";
=cut

	if (@set_manip_sets>0) {
	    ## my $max_indices = max_scoring_regions (@set_manip_sets, 1);
	    my ($max_indices, $max_score) = find_max_elements (@set_manip_sets, 1);
=for DEBUGGING the script
	    print STDERR "Indicec of maximum sets:  ", join (',', @{$max_indices}), "\n\n";
=cut
	    my @new_maxsets = ();
	    foreach my $maxset (@maxsets_nr_regions) { # The 2nd outer for-loop #
		foreach my $max_indx (@{$max_indices}) { # The 2nd inner for-loop #
		   push @new_maxsets, [@{$maxset}, @{$set_manip_sets[$max_indx]->[0]} ];
		}				# End of the 2nd inner for-loop #
	    }					# End of the 2nd outer for-loop #
	    @maxsets_nr_regions = @new_maxsets; # Update the maximum sets # 
	}

=for DEBUGGING the script
	print STDERR "Final output:\n";
	print STDERR "{Maximum sets}\n";
	foreach my $maxset (@maxsets_nr_regions) { print STDERR join (',' , @{$maxset}), "\n"; }
	print STDERR "\n";
=cut

		## Add the finishing touch ##
	my @output = ();
	foreach my $maxset (@maxsets_nr_regions) { # The final outer foreach-loop #
	    my @nr_regions = ();
	    foreach my $indx (@{$maxset}) { # The final inner foreach-loop #
		push @nr_regions, $set_regions->[$indx];
	    }				# End of the final inner foreach-loop #
	    push @output, \@nr_regions;
	}				# End of the final outer foreach-loop #

	return \@output;
	###return \@maxsets_nr_regions;

}		# End of the subroutine 'set_max_nonovlp_regions' #


# screen_intvls_by_len:
# 	This subroutine receives a set of intervals,
#	and screens them according to the condition
#	  $end - $beg + 1 >= $thrsh_len.
#
# INPUT: (\@set_intvls, $indx_beg, $indx_end, $thrsh_len), where
#	@{$set_intvls[$r]}[$indx_beg, $indx_end] = ($beg, $end)
#	are the beginning and ending coordinates of the intervals.
#
# OUTPUT: (\@long_intvls, \@short_intvls), where
# 	@long_intvls stores intervals which are longer than or as long as $thrsh_len,
#	@short_intvls stores those which are shorter than $thrsh_len.
#
sub screen_intvls_by_len (\@$$$) {

	my ($set_intvls, $indx_beg, $indx_end, $thrsh_len) = @_;

	my @long_intvls = my @short_intvls = ();

	foreach my $intvl (@{$set_intvls}) {
	    my ($beg, $end) = @{$intvl}[$indx_beg, $indx_end];
	    ($end - $beg + 1 >= $thrsh_len) ?
		(push @long_intvls, $intvl) : (push @short_intvls, $intvl) ;
	}

	return (\@long_intvls, \@short_intvls);

}		# End of the subroutine 'screen_intvls_by_len' #


# transpose_matrix:
#	This subroutine receives a rectangular matrix (as a 2-dimentional array)
#	and returns the 'transposed' matrix, in which the rows and the columns are swapped
#	from the original.
#
# INPUT: \@matrix, where
#	$matrix[$i]->[$j] gives the element in the intersection of 
#	the $i th row and the $j th column (with 0<= $i <$ct_row, 0<= $j < $ct_clm).
#
# OUTPUT: \@transposed, where
#	$transposed[$j]->[$i] = $matrix[$i]->[$j].
#
sub transpose_matrix (\@) {

	my ($matrix) = @_;

	my $ct_row = @{$matrix};
	unless ($ct_row >0) {
	    print STDERR "[ERROR 'transpose_matrix'] Received an EMPTY ARRAY : $! !\n";
	    return  undef;
	}

	my $ct_clm = @{$matrix->[0]};

	my @transposed = ();
	for (my $j=0; $j<$ct_clm; $j++) { # The outer for-loop #

	    my @tr_row = ();

	    for (my $i=0; $i<$ct_row; $i++) { # The inner for-loop #
		push @tr_row, $matrix->[$i][$j];
	    }				# End of the inner for-loop # 

	    push @transposed, \@tr_row;

	}			# End of the outer for-loop #

	return \@transposed;

}		# End of the subroutine 'transpose_matrix' #


1;	# a 'spell' which makes the module 'use'-able and 'require'-able #


