package MyMalnIO;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyMalnIO.pm: This module reads in output files of multiple alignment,
#		processes them, and print-outs an output in a specific format.
#
# This module was forged based on a part of the Perl script named "malign_fractionate_ms.pl" (on May 9 (Fri), 2003).
#

use strict;
use warnings;
use Exporter;

our @ISA = qw( Exporter);

#our @EXPORT = qw( extract_clw pr_phylip_like );
our @EXPORT_OK = qw(
	id2seq  batch_id2seq br_print_fasta extract_fasta readin_mfasta
	extract_clw extract_in_oden extract_phylip_like extract_phylip_int
	extract_alignment
	process_seqID extract_paraln align_refinery_aa align_refinery_nt
        trim_outgroup rm_monomorphic rm_in_monomorphic pairwise_difference
	pr_phylip_aln  pr_phylip_like  pr_ODEN_input  pr_ODEN_input_lite
	pr_clustal_like pr_clustal_like_lite pr_clustal_like_aa
	pr_clustal_like_lite_aa pr_selected_sites
        print_aln_decorated monomorphic_counter print_bl_aln_decorated
		);

use MyFile_manipu qw( open_input_file open_input_file2 );
use MyMalnManipu qw( id_indicator id_indicator_aa );
use MyMalnManipu_pep_spe qw( consv_indicator_aa );

# id2seq: This subroutine receives a database name and a sequence ID,
#	conducts a databasse search via 'fastacmd',
#	and returns a reference to the resulting sequence with the title (seq ID)  part removed.
#	(NOTE: This subroutine strips out the white spaces trailing the sequence.)
#
sub id2seq (\$$) {

	my ($database, $seqid) = @_; 

	##my $fastacmd = '/home/ezawa/Bioinfo_tools/blast-2.2.6-ia32/fastacmd';
	my $fastacmd = '/usr/local/Bioinfo_tools/blast-2.2.6-ia32/fastacmd';

	my $input = ` $fastacmd -d ${$database} -s $seqid ` ;

	unless ($input =~ m/^\>/ ) {
		print STDERR $input;
		return undef;
	}

	my ($title, $seq) = split /\n/, $input , 2;

	(defined $seq) or (return undef);

	$seq =~ s/\s+$//;

	return \$seq;
	

}		# end of the subroutine id2seq #


# batch_id2seq: This subroutine conducts the subroutine id2seq multiple times.
#	We ASSUME that the sequence IDs are in the form registerd in the NCBI database
#	used by 'fastacmd'.
#
sub batch_id2seq (\@\@) {

	my ($databases, $seqids) = @_;

	my $ct_seqs = @{$seqids};
	my @db_list = @{$databases};
	if ( ($ct_seqs > 1) and (@db_list == 1) ) {
		for (2 .. $ct_seqs) { push @db_list, $db_list[0]; }
	}

	unless (@db_list == $ct_seqs) {
		print STDERR "Illegal input for \'batch_id2seq\'.\n";
		return undef;
	}

	my @seqs = ();
	for (my $i=0; $i< $ct_seqs; $i++) {

		my $seq = id2seq (${$db_list[$i]}, $seqids->[$i]);
		push @seqs, $seq;
	}

	return \@seqs;

}		# end of the subroutine batch_id2seq # 


# br_print_fasta: This subroutine just concatenates the ">"-prepended sequence IDs and 
#		the corresponding  outputs of "id2seq" (or the like),
#		and prints them out to the file specified by the 2nd argument.
#
sub br_print_fasta (\@\@\$) {

	my ($seqids, $seqs, $outfile ) =@_;

	my $ct_seqs = @{$seqs};
	($ct_seqs) or (return 1);

	unless (@{$seqids} == $ct_seqs) {
		print STDERR "Illegal input for \'br_print_fasta\': the number of sequence IDs does NOT agrees with that of sequences.\n";
		return 0;
	}

	my $fh;
	open ($fh, "> ${$outfile}" ) or die "\nCannot open ${$outfile}: $!\n";

	for (my $i=0; $i< $ct_seqs; $i++) {
		print $fh "\>$seqids->[$i]\n";
		print $fh "${$seqs->[$i]}\n";
	}

	close ($fh);

	return 1;
}			# end of the subroutine br_print_fasta #


# extract_fasta: This subroutine extracts a list of sequence IDs  and thier associated sequences
#		from a file in the fasta format.
#
sub extract_fasta (\$) {

	my ($rf_infile) = @_;

	local $/ = "\n";

	unless ( open (INTMP, "< ${$rf_infile}") ) {
		open (INTMP, " gunzip -c ${$rf_infile}.gz | ") or die "\nCannot folk or open ${$rf_infile}(.gz): $!\n";
	}

	my $seqID = '';
	my %seqid2desc; # ($seqID => \$desc, ...)

	while (my $header = <INTMP>) {
		($header =~ s/^>\s*//) or next;
		my ($tmp)  = ($header =~ m/^(\S+)/ );
		##my ($tmp) = ( $header =~ m/^>(\S+)/ ); 
		(defined $tmp) or next;
		$seqID = $tmp;
		$header =~ s/\s+$//;
		my ($desc) = ($header =~ m/^\S+\s+(\S+.*)/ );
		(defined $desc) and ($seqid2desc{$seqID} = \$desc);
		last;
	}
	$seqID or (return ([],[], [], {}) );

	my @seqIDs = ($seqID);
	my @seqs = my @seqlens = ();
	my $seq = '';

	while (my $line = <INTMP>) {

		## my ($seqID) = ($line =~ m/^>(\S+)/ );
		##if (defined $seqID) {
		if ($line =~ s/^>\s*//) {
			$line =~ s/\s+$//;
			my ($seqID, $desc) = split /\s+/, $line, 2;
			unless (defined $seqID) {
			    print STDERR "[ERROR 'extract_fasta'] NO SEQUENCE ID !\n";
			    return (undef, undef, undef, undef);
			}
			my $length = ($seq =~ tr/a-zA-Z\*// );
			push @seqlens, $length;
			push @seqIDs, $seqID;
			my $cp_seq = $seq;
			$seq = '';
			push @seqs, \$cp_seq;
			((defined $desc) and ($desc ne '')) and ($seqid2desc{$seqID} = \$desc);
		} else {
			$line =~ s/\s+$//;
			#$line =~ s/\s+//g;
			$seq .= $line;
		}
	}

	close (INTMP);

	my $length = ($seq =~ tr/a-zA-Z\*//);
	push @seqlens, $length;
	push @seqs, \$seq;

	return (\@seqs, \@seqIDs, \@seqlens, \%seqid2desc);

}			# end of the file extract_fasta #


# readin_mfasta:
#	This subroutine reads in a list of sequences in the multi-FASTA format,
#	then stores the retrieved information into two hashes,
#	%id2seq and %id2desc.
#
# INPUT: (\$infile, \%id2seq, \%id2desc), where
#	$infile is the path to the input file;
#	%id2seq = ($seq_id => \$sequence, ...);
#	%id2desc = ($seq_id => \$desc, ...) for the sequences whose descriptions are available.
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub readin_mfasta (\$\%\%) {

	my ($infile, $id2seq, $id2desc) = @_;
	local $/ = "\n";

	my $fh = open_input_file (${$infile}); 
	unless (defined $fh) {
	    print STDERR "[ERROR 'readin_mfasta'] 'open_input_file' FAILED !!\n";
	    return 0;
	}

	my $seqid;
	my $seq; # = $id2seq->{$seqid};
	while (my $in = <$fh>) {

	    if ($in =~ s/^>\s*//) { # The header part
		$in =~ s/\s+$//;
		($seqid, my $desc) = split /\s+/, $in;
		unless ((defined $seqid) and ($seqid ne '')) {
		    print STDERR "[CAUTION 'readin_mfasta'] ANOMALOUS HEADER containing NO SEQUENCE ID:\n";
		    print STDERR " '>$in' !\n\n";
		    $seqid = undef;
		    next;
		}
		$seqid =~ s/^lcl\|//; # ADDED on Monday, March 26, 2007 #

		my $init_seq = '';
		$seq = $id2seq->{$seqid} = \$init_seq;

		(defined $desc) and ($id2desc->{$seqid} = \$desc);

	    } elsif (defined $seqid) {
		$in =~ s/\s+$//;
		${$seq} .= $in;
	    }

	}		# End of the subroutine 

	return 1;

}		# End of the subroutine 'readin_mfasta' #


# extract_alignment:
#	This subroutine extracts an alignment,
#	either in the FASTA format or in the CLUSTAL format (with out without a header).
#
# INPUT: \$inaln_file.
#
#  OUTPUT: (\@align, \@seqids, \@seqlens), where
#	${$align[$k]} is the $k th sequence embedded in the alignment;
#	$seqids[$k] is the ID of the $k th sequence;
#	$seqlens[$k] is the length of the $k th sequence;
#
sub extract_alignment (\$) {

	my ($inaln_file) = @_;
	local $/ = "\n";

	my $fh_tmp = open_input_file2 (${$inaln_file});
	unless (defined $fh_tmp) {
	    print STDERR "[ERROR 'extract_alignment'] The subroutine 'open_input_file2' FAILED on the input file '${$inaln_file}': $! !\n";
	    return (undef, undef, undef);
	}
	my $header;
	while (my $in = <$fh_tmp>) { # Leading while-loop #
	    ($in =~ m/\S+/) or next;
	    ($header = $in) =~ s/^\s+//;
	    last;
	}		# End of the leading while-loop #
	close ($fh_tmp);

	unless (defined $header) {
	    print STDERR "[ERROR 'extract_alignment'] NO UNBLANK LINES in the input alignment '${$inaln_file}': $! !\n";
	    return (undef, undef, undef);
	}
	$header = uc $header;

	if ($header =~ m/^CLUSTAL/) {
	    my ($align, $seqids, $seqlens, $labels) = extract_clw (${$inaln_file}, 1);
	    unless ((defined $align) and (@{$align}>0)) {
		print STDERR "[ERROR 'extract_alignment'] Subroutine 'extract_clw' FAILED with \$fl_header = 1 on the  input alignment '${$inaln_file}': $! !\n";
		return (undef, undef, undef);
	    }
	    return ($align, $seqids, $seqlens);

	} elsif ($header =~ m/^\>/) {
	    my ($align, $seqids, $seqlens, $seqid2desc) = extract_fasta (${$inaln_file});
	    unless ((defined $align) and (@{$align}>0)) {
		print STDERR "[ERROR 'extract_alignment'] Subroutine 'extract_fasta' FAILED on the input alignment '${$inaln_file}': $! !\n";
		return (undef, undef, undef);
	    }
	    return ($align, $seqids, $seqlens);
	
	} else { # Try 'extract_clw' with $fl_header = 0.
	    my ($align, $seqids, $seqlens, $labels) = extract_clw (${$inaln_file}, 0);
	    unless ((defined $align) and (@{$align}>0)) {
		print STDERR "[ERROR 'extract_alignment'] Subroutine 'extract_clw' FAILED with \$fl_header = 0 on the  input alignment '${$inaln_file}': $! !\n";
		return (undef, undef, undef);
	    }
	    return ($align, $seqids, $seqlens);
	}

}	# End of the subroutine 'extract_alignment' #


# align_extract_str: a subroutine which receives lines of a chunk of alignment,
#			extracts the whole alignment,
#			and saves the result in the array OF STRINGS "@align"
#
sub align_extract_str (\@\@\@) {

	my  ($rf_lines, $rf_align, $rf_seq_lens) = @_;
	my @characters;

	my $i =0;
	foreach  my $line (@{$rf_lines}) {

		my ($aln) = ($line =~ m/^\S+\s+(\S+)/ );
		(defined $aln) or next;
		my $part_len = ($aln =~ tr/a-zA-Z\*//);
		($rf_seq_lens->[$i]) ? ($rf_seq_lens->[$i] += $part_len) : ($rf_seq_lens->[$i] = $part_len);
		(defined $rf_align->[$i]) ? (${$rf_align->[$i]} .= $aln) : ($rf_align->[$i] = \$aln);
		$i++;
	}

}				# end of the subroutine align_extract_str #


#my %clusters;		# key = (sub)cluster number, value = [list of sequence IDs ] #

# process_seqID ($\%): This subroutine receives the name of an input file, 
#			processes the information,
#			and stores the resulting list of sequences belonging to each cluster.
#
sub process_seqID ($\%) {

	my ($infile, $rf_clusters) = @_;

	#local $/ = "";	# use a blank line as the separator #

	open (INTMP, "<$infile") or die "\nCannot open $infile: $!\n";
	while (my $cluster_rec = <INTMP> ) {
		 $cluster_rec =~ s/^\s+//;
		($cluster_rec =~ m/^cluster/) or ($cluster_rec =~ m/^subcl/) or ($cluster_rec =~ m/^original/ ) or next;

		my ($header, $sequences) = split /\n/, $cluster_rec;
		my ($cluster_num) = ($header =~ m/^\D*([\d_]+)/ );

		$sequences =~ s/^\[//;
		$sequences =~ s/\]\.$//;
		my @list_seq = split /,/ , $sequences;
		${$rf_clusters}{$cluster_num} = \@list_seq;
	}
	close (INTMP);
	
}				# end of the subroutine process_seqID # 


# extract_paraln : This subroutine reads in an output file of the parallel computing
#			multiple-alignment program, extracts sequences embedded in the alignment,
#			and outputs the result into an array of strings.
#
sub extract_paraln ($\%) {

	my ($infile, $rf_clusters) = @_;

	my ($cluster_num) = ($infile =~ m/CL([\d_]+)\.out$/o);
	unless (defined $rf_clusters->{$cluster_num}) {
		print STDERR "Sequence ID list is not available for cluster no.$cluster_num.\n";
		return ([], [], []);
	}
	my $rf_seq_IDs = $rf_clusters->{$cluster_num};
	local $/ = "\n";

	my @align = ();
	my @seq_lengths = ();

 	open(INTMP, "<$infile") or die "\nCannot open $infile: $!\n";
	while(my $input = <INTMP>) {

		($input =~ s/^\"//) or next;
		chomp $input;
		$input =~ s/\"$//;
		push @align, \$input;

		my $length = ($input =~ tr/a-zA-Z\*// );
		push @seq_lengths, $length;
	}
	close(INTMP);

	if ( scalar(@{$rf_seq_IDs}) != scalar(@align) ) {
		print STDERR "Cluster no.$cluster_num does not have as many seq_ID\'s as sequences.\n";
		return ([], [], []);
	}

	my $max_len = 0;
	foreach my $seqid (@{$rf_seq_IDs}) {

		($max_len < length($seqid) ) and ($max_len = length($seqid) );
	}

	my @labels = ();
	foreach my $seqid (@{$rf_seq_IDs}) {

		my $pad_len = 5 + $max_len - length($seqid);
		my $label = $seqid . ' ' x $pad_len;
		push @labels, $label;
	}

	return (\@align, \@seq_lengths, \@labels);

 }					# end of the subroutine extract_paraln #


# extract_clw : This subroutine reads in a multiple alignment
#			from the file of clustal format.
# 			And store the alignment in an array of STRINGS.
#
sub extract_clw ($$) {

	my ($infile, $fl_header) = @_;

	local $/ = '';		# use blank lines as input_record_sepqrators #

	open (INTMP, "<$infile") or die "\nCannot open $infile: $!\n";

	if ($fl_header) { my $header = <INTMP>;}

	my @first_lines = ();

	while (my $in = <INTMP> ) { 	# the 1st while #

	 	@first_lines = split /\n/, $in;
		@first_lines and last;
	}				# end of the 1st while #

	(@first_lines) or ( return ( [], [], [], [] ) );
 
	my @labels = ();
	my @seq_IDs = ();
	foreach my $line (@first_lines) {

		my ($label, $seqid) = ($line =~ m/^((\S+)\s+)/ ) or next;
		push @labels, $label;
		push @seq_IDs, $seqid;
	}

	my $fl_ref = ($first_lines[$#first_lines] =~ m/^\s/ );

	my @align = ();
	my @seq_lengths = ();

	for (my $i = 0; $i < @labels; $i++) {
		my $temp = '';
		$align[$i] = \$temp;
	}

	align_extract_str ( @first_lines, @align, @seq_lengths);

	while (my $in = <INTMP> ) {	# the 2nd while #

		my @in_lines = split /\n/, $in;
		@in_lines or next;

		align_extract_str (@in_lines, @align,@seq_lengths);

	}				# end of the 2nd while #

	close (INTMP);

	#$fl_ref and (pop @seq_lengths);

	return (\@align, \@seq_IDs, \@seq_lengths, \@labels);

}				# end of the subroutine extract_clw #


# align_extract_str_II: a subroutine which receives lines of a chunk of alignment,
#			extracts the whole alignment,
#			and saves the result in the array OF STRINGS "@align".
#			This subroutine differs from 'align_extract_str' in that
#			is discards lines with a specified "comment character" prepended.
#
sub align_extract_str_II (\@\@\@$) {

	my  ($rf_lines, $rf_align, $rf_seq_lens,$comchar) = @_;
	my @characters;

	my $len_comchar = length $comchar;

	my $i =0;
	foreach  my $line (@{$rf_lines}) {

		my ($aln) = ($line =~ m/^\S+\s+(\S+)/ );
		(defined $aln) or next;
		( ($len_comchar <= length $line )
			and (substr ($line, 0, $len_comchar) eq $comchar) ) and next;
		my $part_len = ($aln =~ tr/a-zA-Z\*//);
		($rf_seq_lens->[$i]) ? ($rf_seq_lens->[$i] += $part_len) : ($rf_seq_lens->[$i] = $part_len);
		(defined $rf_align->[$i]) ? (${$rf_align->[$i]} .= $aln) : ($rf_align->[$i] = \$aln);
		$i++;
	}

}				# end of the subroutine align_extract_str_II #


# extract_in_oden : This subroutine reads in a multiple alignment
#			in a format specially prepared for ODEN,
# 			And store the alignment in an array of STRINGS.
#
sub extract_in_oden ($$) {

	my ($infile, $fl_header) = @_;

	local $/ = '';		# use blank lines as input_record_sepqrators #

	open (INTMP, "<$infile") or die "\nCannot open $infile: $!\n";

	if ($fl_header) { my $header = <INTMP>; }

	my @first_lines = ();

	while (my $in = <INTMP> ) { 	# the 1st while #

	 	@first_lines = split /\n/, $in;
		@first_lines and last;
	}				# end of the 1st while #

	(@first_lines) or ( return ( [], [], [], [] ) );
 
	my @labels = ();
	my @seq_IDs = ();
	foreach my $line (@first_lines) {

		my ($label, $seqid) = ($line =~ m/^((\S+)\s+)/ ) or next;
		($label =~ m/^\>/ ) and next;	# skip the comment line
		push @labels, $label;
		push @seq_IDs, $seqid;
	}

	#my $fl_ref = ($first_lines[$#first_lines] =~ m/^\s/ );

	my @align = ();
	my @seq_lengths = ();

	for (my $i = 0; $i < @labels; $i++) {
		my $temp = '';
		$align[$i] = \$temp;
	}

	align_extract_str_II ( @first_lines, @align, @seq_lengths, '>');

	while (my $in = <INTMP> ) {	# the 2nd while #

		my @in_lines = split /\n/, $in;
		@in_lines or next;

		align_extract_str_II (@in_lines, @align,@seq_lengths, '>');

	}				# end of the 2nd while #

	close (INTMP);

	#$fl_ref and (pop @seq_lengths);

	return (\@align, \@seq_IDs, \@seq_lengths, \@labels);

}				# end of the subroutine extract_in_oden #


# extract_phylip_like: This subroutine reads in a multiple-alignment
#			in a Phylip-like format,
#			and returns lists of sequence IDs, sequences, lengths, and labels.
#
sub extract_phylip_like (\$$) {

	my ($rf_in, $fl_header) = @_;

	my @chunks = split /\n\n+/, ${$rf_in};

	my $header = $fl_header ? (shift @chunks) : undef ;
	@chunks  or (return (undef, undef, undef, undef) );

	my $labels = shift @chunks;
	my @labels = split /\n/, $labels;

	my @seqids = ();
	foreach my $label (@labels) {

		my ($seqid) = ($label =~ m/^\>?(\S+)/ );
		(defined $seqid) or ($seqid = $label);
		push @seqids, $seqid;
		my $tmp = '';
	}			# end of the foreach #

	my @seqs = ();
	while (my $sub_aln = shift @chunks) { # the outer while #

		my @subseqs = split /\n/, $sub_aln;

		for (my $i=0; $i<@subseqs; $i++) { # the inner for #

			$subseqs[$i] =~ s/\s+//g;  # remove white spaces #
			(defined $seqs[$i]) ? (${$seqs[$i]} .= $subseqs[$i]) :
						($seqs[$i] = \$subseqs[$i]) ;	

		}				# end of the inner for #
	}				# end of the outer while #

	unless (@seqs == @labels) {

		printf STDERR "Discrepancy between the sequence count %d and the number of labels %d\n", scalar(@seqs), scalar(@labels);
		return (undef, undef, undef, undef);
	}

	my @lengths = ();
	for (my $i=0; $i<@seqs; $i++) { 

		my $len =  (${$seqs[$i]} =~ tr/a-zA-Z\*// );
		push @lengths, $len;
	}

	return  (\@seqs, \@seqids, \@lengths, \@labels);

}			# end of the subroutine extract_phylip_like # 


# extract_phylip_int:
#	This subroutine extracts an alignment in the Phylip interleaved format.
#
# INPUT: \$infile.
#
# OUTPUT: (\@seqs, \@seqids), where
#	${$seqs[$m]} is the $m th sequence embedded in the alignment;
#	$seqids[$m] is the ID of the $m th sequence.
#
sub extract_phylip_int (\$) {

	my ($infile) = @_;
	local $/ = '';

	my $fh_tmpin ;
	unless ( open ($fh_tmpin, "< ${$infile}") ) {
	   print STDERR "\n[ERROR 'extract_phylip_int'] CanNOT open '${$infile}' : $! .\n\n";
	   return (undef, undef);
	}

	my $header = <$fh_tmpin>;
	my @headlines = split /\n/, $header;
	unless (@headlines>1) {
	    $header =~ s/\s+$//;
	    print STDERR "[ERROR 'extract_phylip_int'] The input file '${$infile}' has an UNEXPECTEDLY SHORT HEADER:\n";
	    print STDERR "'$header' .\n\n";
	    return (undef, undef);
	}

	my $init_info = shift @headlines;
	unless ($init_info =~ m/\s+\d+\s+\d+/) {
	    $header =~ s/\s+$//;
	    print STDERR "[ERROR 'extract_phylip_int'] The input file '${$infile}' has an UNEXPECTED form of the HEADER:\n";
	    print STDERR "'$header' .\n\n";
	    return (undef, undef);
	}

	my @seqs = my @seqids = ();

	while (my $line = shift @headlines) { # The 1st while-loop #
	    my ($seqid, $seq) = split /\s+/, $line, 2;
	    if (defined $seq) {
		$seq =~ s/\s+//g;
		push @seqs, \$seq;
		push @seqids, $seqid;
	    } else {
		($seqid = $line) =~ s/\s+$//;
		$seq = '';
		push @seqs, \$seq;
		push @seqids, $seqid;
	    }
	}			# End of the 1st while-loop # 
	my $ct_seqs = @seqs;

	while (my $in = <$fh_tmpin>) { # The 2nd while-loop #
	    my @lines = split /\n/, $in;
	    unless (@lines == $ct_seqs) {
		print STDERR "[ERROR 'extract_phylip_int'] The number of the lines (", scalar (@lines), ") CLASHES with that of the sequences ($ct_seqs) for the alignment '${$infile}' !\n";
		return (undef, undef);
	    }

	    for (my $m=0; $m<$ct_seqs; $m++) { # The inner for-loop #
		my $line = $lines[$m];
		$line =~ s/\s+//g;
		${$seqs[$m]} .= $line;
	    }			# End of the inner for-loop #

	}		# End of the 2nd while-loop #

	close ($fh_tmpin);

	return (\@seqs, \@seqids);

}			# End of the subroutine 'extract_phylip_int' #


# align_refinery_aa: This subroutine extracts alignment columns which consist 
#		    only of unambiguous amino acid residues.
#
#
sub align_refinery_aa (\@$) {

	my ($rf_align, $sw_amb) = @_;

	@{$rf_align} or (return ([], [])  );

	my @refined = ();
	my @site_list = ();
	#my @copy = ();

	for (my $i=0; $i<@{$rf_align}; $i++) {	# the 1st for #

		my $tmp = '';
		push @refined, \$tmp;
		#push @copy, ${$rf_align->[$i]};
	}					# end of the 1st for #


	my $aln_len = length(${$rf_align->[0]});
	for (my $l=0; $l<$aln_len; $l++) {# the 2nd for #

		my @reservoir = ();
		my $ct_amb = 0;

		foreach my $rf_seq (@{$rf_align}) {	# the middle foreach #

			my $tmp = substr(${$rf_seq}, 0, 1, ''); 
			##($tmp =~ m/[ac-ik-np-tvwy]/i ) or ($ct_amb++ );
			my $tmp2 = uc $tmp;
			($tmp2 =~ tr/ACDEFGHIKLMNPQRSTUVWY\*//) or ($ct_amb++);
				# U: selenocystein, *: stop codon #
			push @reservoir, $tmp;

		}				# end of the middle foreach #

		if ($ct_amb) {

			my $site = $l +1;
			$sw_amb and (print STDERR "site no.= $site ($ct_amb ambiguous sites):\t@reservoir\n");
			 next;
		}

		for (my $i=0; $i<@refined; $i++) { ${$refined[$i]} .= $reservoir[$i];}
		push @site_list , $l +1;


	}					# end of the 2nd for #

	return (\@refined, \@site_list) ;

}		# end of the subroutine align_refinery_aa #


# align_refinery_nt: This subroutine extracts alignment columns which consist 
#		    only of unambiguous nucleotide residues.
#
#
sub align_refinery_nt (\@$) {

	my ($rf_align, $sw_amb) = @_;

	@{$rf_align} or (return ([], [])  );

	my @refined = ();
	my @site_list = ();
	#my @copy = ();

	for (my $i=0; $i<@{$rf_align}; $i++) {	# the 1st for #

		my $tmp = '';
		push @refined, \$tmp;
		#push @copy, ${$rf_align->[$i]};
	}					# end of the 1st for #


	my $aln_len = length(${$rf_align->[0]});
	for (my $l=0; $l<$aln_len; $l++) {# the 2nd for #

		my @reservoir = ();
		my $ct_amb = 0;

		foreach my $rf_seq (@{$rf_align}) {	# the middle foreach #

			my $tmp = substr(${$rf_seq}, 0, 1, ''); 
			($tmp =~ m/[acgtuACGTU]/ ) or ($ct_amb++ );
			push @reservoir, $tmp;

		}				# end of the middle foreach #

		if ($ct_amb) {

			my $site = $l +1;
			$sw_amb and (print STDERR "site no.= $site ($ct_amb ambiguous sites):\t@reservoir\n");
			 next;
		}

		for (my $i=0; $i<@refined; $i++) { ${$refined[$i]} .= $reservoir[$i];}
		push @site_list , $l +1;


	}					# end of the 2nd for #

	return (\@refined, \@site_list) ;

}		# end of the subroutine align_refinery_nt #


# trim_outgroup : This subroutine removes all outgroup sequences except the best one.
#
sub trim_outgroup (\@\@$$) {

	my ($rf_align, $rf_seqIDs, $bestout, $pre_out) = @_;

	my $pre_len = length $pre_out;
	@{$rf_align} or (return ($rf_align, $rf_seqIDs) );

	my @out_align = my @out_seqIDs = ();

	for (my $i=0; $i<@{$rf_seqIDs}; $i++) {

		#($rf_seqIDs->[$i] =~ m/^$pre_out/o ) and ($rf_seqIDs->[$i] ne $bestout) and next;
		(substr($rf_seqIDs->[$i], 0, $pre_len) eq $pre_out) and ($rf_seqIDs->[$i] ne $bestout) and next;

		push @out_align, $rf_align->[$i];
		push @out_seqIDs, $rf_seqIDs->[$i];
	}

	return (\@out_align, \@out_seqIDs);

}			# end of the subroutine trim_outgroup #


# rm_monomorphic: This subroutine receives a multiple alignment
#			and removes columns with identical characters,
#			but it does NOT care if the columns include ambiguous characters or not.
#		If it receives a dummy list as the second argument, the subroutine
#		comstructs a list of selected sites based on the order of the selected sites in the input string.
#
sub rm_monomorphic (\@\@) {

	my ($rf_refined, $rf_site_list) = @_;

	@{$rf_refined} or (return ( [], []) );
	(my $len = length ${$rf_refined->[0]} ) or (return ($rf_refined, []) );
	my $fl_sitelist = ($len == @{$rf_site_list} );
 
	my @selected_sites = ();
	for (my $i =0; $i< @{$rf_refined}; $i++) {
		my $tmp = '';
		push @selected_sites, \$tmp;
	}
	my @sel_site_list = ();

	for (my $l=0; $l < $len; $l++) {

		my $id_ct =1;
		my $first_char = substr( ${$rf_refined->[0]}, 0, 1, '');
		my @reservoir = ($first_char);

		for (my $i=1; $i<@{$rf_refined}; $i++) {

			my $char = substr(${$rf_refined->[$i]}, 0, 1, '');
			($char eq $first_char) and $id_ct++;
			push @reservoir, $char;
		}
		($id_ct == @{$rf_refined}) and next;

		push @sel_site_list, ($fl_sitelist ? $rf_site_list->[$l] : $l+1) ;

		for (my $i=0; $i<@selected_sites; $i++) { ${$selected_sites[$i]} .= $reservoir[$i]; }
	}

	return (\@selected_sites, \@sel_site_list);

}			# end of the subroutine rm_monomorphic #

# rm_in_monomorphic: This subroutine recives a multiple alignment
#			and removes columns in which all INGROUP characters are identical,
#			but it does NOT care if the columns include ambiguous characters or not.
#
#		If it receives a dummy list as the second argument, the subroutine
#		comstructs a list of selected sites based on the order of the selected sites in the input string.
#
sub rm_in_monomorphic (\@\@\@$) {

	my ($rf_refined, $rf_site_list, $rf_seqIDs, $pre_out) = @_;
	my $fl_report = 0;	# if we want a report of removed sites(=1) or not(=0) #


	@{$rf_refined} or (return ( [], []) );
	(my $len = length ${$rf_refined->[0]} ) or (return ($rf_refined, []) );
	unless (@{$rf_refined} == @{$rf_seqIDs} ) {
		print STDERR "Discrepancy in the number of sequences and that of sewquence IDs !\n";
		return ([], []);
	}
	my $fl_sitelist = ($len == @{$rf_site_list} );

	my @list_in = ();

	for (my $i=0; $i<@{$rf_seqIDs}; $i++) {

		($rf_seqIDs->[$i] =~ m/^$pre_out/o) or (push @list_in, $i);
	}
 
	my @selected_sites = ();
	for (my $i =0; $i< @{$rf_refined}; $i++) {
		my $tmp = '';
		push @selected_sites, \$tmp;
	}
	my @sel_site_list = ();

	for (my $l=0; $l < $len; $l++) {

		my @reservoir = ();

		for (my $i=0; $i<@{$rf_refined}; $i++) {

			my $char = substr(${$rf_refined->[$i]}, 0, 1, '');
			push @reservoir, $char;
		}

		my $id_ct =0;
		my $first_char = $reservoir[ $list_in[0] ];

		foreach my $i (@list_in) {

			($first_char eq $reservoir[$i]) and $id_ct++;
		}

		if ($id_ct == @list_in){

		   if ($fl_report) {
			my $site_no = $fl_sitelist ? $rf_site_list->[$l] : $l+1 ;
			print STDERR "Site no. = $site_no: ";
			print STDERR @reservoir;
			print STDERR "\n";
		   }
			next;
		}


		push @sel_site_list, ($fl_sitelist ? $rf_site_list->[$l] : $l+1) ;

		for (my $i=0; $i<@selected_sites; $i++) { ${$selected_sites[$i]} .= $reservoir[$i]; }
	}

	return (\@selected_sites, \@sel_site_list);

}			# end of the subroutine rm_in_monomorphic #


# pairwise_difference: This subroutine receives a pair of sequences in an alignment 
#			and enumerates the sites where the sequences differ.
#
sub pairwise_difference (\$\$\@) {

	my ($rf_seq1, $rf_seq2, $rf_site_list) = @_;

	my $fl_list = ( (defined $rf_site_list) and (length(${$rf_seq1}) == @{$rf_site_list}) ); 
	my @list_differences = my @list_gaps = ();

	my $seq1 = ${$rf_seq1};
	my $seq2 = ${$rf_seq2};

	## ADDRED on March 16 (Mon), 2009 ##
	my $len1 = length $seq1;
	my $len2 = length $seq2;
	if ($len1 < $len2) {
	    $seq1 .= '-' x ($len2 - $len1);
	} elsif ($len1 > $len2) {
	    $seq2 .= '-' x ($len1 - $len2); 
	}

	my $site_ct = length $seq1;

	for (my $l=0; $l<$site_ct; $l++) {

		my $site1 = substr($seq1, 0, 1, '');
		my $site2 = substr($seq2, 0, 1, '');
		($site1 eq $site2) and next;

		my $site_num = $fl_list ? $rf_site_list->[$l] : $l + 1;
		my @info = ($site_num, $site1, $site2);

		if ( ($site1 eq '-') or ($site2 eq '-') ) {

			push @list_gaps, \@info;
		} else {

			push @list_differences, \@info;
		}
	}

	return (\@list_differences, \@list_gaps);

}			# end of the subroutine pairwise_difference #


# pr_phylip_aln:
#	This subroutine prints out a muliple alignment
#	exacly in the Phylip format, but WITHOUT A HEADER.
#
# INPUT: (\@seqids, \@align, $dis_len, \*${$fh}), where
#	$seqids[$m] and ${$align[$m]} are, respectively,
#	the identity and the sequence, may with gaps inserted,
#	of the $m th sequence.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub pr_phylip_aln (\@\@$\*) {

	my ($seqids, $align, $dis_len, $fh) = @_;

	my $ct_seq = @{$align};
	($ct_seq>0)  or  (return 0);

	my @alignment = ();

	for (my $m=0; $m<$ct_seq; $m++) { # The 1st outer for #

		## Trim the label or pad it with spaces
		## to force it to be 10 characters long.
	    my $aln = $seqids->[$m];
	    my $id_len = length $aln;
	    if($id_len >10) { $aln = substr ($aln, 0, 10);
	    } elsif ($id_len <10) { $aln = sprintf "%-10s", $aln;
	    }
		## concatenate the label with the sequence
	    ##$aln .= ' ' . ${$align->[$m]};
	    $aln .= ' ' x 2 . ${$align->[$m]};
		## REVISED to conform with PAML (on June 23 (sat), 2007) ##

	    push @alignment, \$aln;

	}		# End of the 1st outer for # 

	my $aln_len = length ${$alignment[0]};
	##my $times = int ( ($aln_len -11) /$dis_len);
	##my $residue = ($aln_len -11) % $dis_len;
	my $times = int ( ($aln_len -12) /$dis_len);
	my $residue = ($aln_len -12) % $dis_len;
		## REVISED to conform with PAML (on June 23 (sat), 2007) ##

	for (my $i=0; $i<$times; $i++) { # The 2nd outer for #

	    ($i==0) or (print $fh "\n");
	    for (my $k=0; $k<@alignment; $k++) { # the inner for #
		my $aln = $alignment[$k];
		my $subseq = ($i==0) ?
		   substr(${$aln}, 0, $dis_len +12, '') :
		   substr(${$aln}, 0, $dis_len, '') ;
		   ##substr(${$aln}, 0, $dis_len +11, '') :
		   ##substr(${$aln}, 0, $dis_len, '') ;
			## REVISED to conform with PAML (on June 23 (sat), 2007) ##

		print $fh "$subseq\n";
	    }			# end of the inner for #
	}			# end of the 2nd outer for #

	if ($residue) {
	    ($times == 0) or (print $fh "\n");
	    foreach my $aln (@alignment) { print $fh "${$aln}\n"; }
	}

	return 1;
}		# end of the subroutine pr_phylip_aln #


# pr_phylip_like: This subroutine prints out a muliple alignment in a Phylip-like format WITHOUT A HEADER.
#
sub pr_phylip_like (\@\@$\*) {

	my ($rf_seqids, $rf_align, $dis_len, $fh) = @_;

	(@{$rf_align})  or  (return 0);

	if (@{$rf_seqids}) {

		print $fh join "\n", @{$rf_seqids};
		print $fh "\n\n";
	}

	my $aln_len = length ${$rf_align->[0]};
	my $times = int ($aln_len/$dis_len);
	my $residue = $aln_len % $dis_len;

	for (my $i=0; $i<$times; $i++) {	# the outer for #

		foreach my $rf_aln (@{$rf_align}) {		# the inner for #

			my $subseq = substr(${$rf_aln}, 0, $dis_len, '');
			print $fh "$subseq\n";

		}				# end of the inner for #

		print $fh "\n";
	}					# end of the outer for #

	if ($residue) {

		foreach my $rf_aln (@{$rf_align}) { print $fh "${$rf_aln}\n"; }
		print $fh "\n";
	}

	return 1;
}		# end of the subroutine pr_phylip_like #


# pr_ODEN_input: This subroutine prints out a muliple alignment in a Phylip-like format WITHOUT A HEADER.
#
# PREREQUISITE: The length of seqence IDs must be less than 15.
#
sub pr_ODEN_input (\@\@$\*) {

	my ($rf_seqids, $rf_align, $dis_len, $fh) = @_;

	(@{$rf_align})  or  (return 0);

	unless (@{$rf_seqids} ) {
		print STDERR "Sequence IDs are missing for \"pr_ODEN_unput\"\!\n";
		return 0;
	}
	unless (@{$rf_seqids} == @{$rf_align}) {
		print STDERR "The number of sequences clashes with that of IDs\!\n";
		return 0;
	}

		# REVISED on Friday, July 14, 2006 #
	my $max_id_len = 0;
	foreach my $seqid (@{$rf_seqids}) {
	    my $len = length $seqid;
	    ($max_id_len < $len) and ($max_id_len = $len);
	}
	my $label_width = $max_id_len + 5;

	my @labels = ();
	foreach my $seqid (@{$rf_seqids}) {

		##my $label = $seqid . ' ' x (15 - length($seqid) );
		my $label = $seqid . ' ' x ($label_width - length ($seqid) ); # REVISED on Frideay, July 14, 2006 #
		push @labels, $label;
	}

	my $aln_len = length ${$rf_align->[0]};
	my $times = int ($aln_len/$dis_len);
	my $residue = $aln_len % $dis_len;

	my @cp_aln = ();
	for my $rf_seq (@{$rf_align}) {
		my $cp_seq = ${$rf_seq};
		push @cp_aln, \$cp_seq;
	}

	#for (my $i=0;  $i<$times; $i++) {	# the outer for #
	foreach my $i (1 .. $times) {	# the outer for #
		#foreach my $rf_aln (@{$rf_align}) {		# the inner for #
		#for (my $j=0; $j<@{$rf_align}; $j++) {		# the inner for #
		for (my $j=0; $j<@cp_aln; $j++) {		# the inner for #

			#my $subseq = substr(${$rf_align->[$j]}, 0, $dis_len, '');
			my $subseq = substr(${$cp_aln[$j]}, 0, $dis_len, '');
			print $fh "$labels[$j]$subseq\n";

		}				# end of the inner for #

		($i == $times) or (print $fh "\n");
	}					# end of the outer for #

	if ($residue) {

		$times and (print $fh "\n");
		#for (my $j=0; $j<@{$rf_align}; $j++) {
		#	 print $fh "$labels[$j]${$rf_align->[$j]}\n"; }
		for (my $j=0; $j<@cp_aln; $j++) {
			 print $fh "$labels[$j]${$cp_aln[$j]}\n"; }
	}

	return 1;
}		# end of the subroutine pr_ODEN_input #


# pr_ODEN_input_lite: This subroutine prints out a muliple alignment in a Phylip-like format WITHOUT A HEADER.
#
# PREREQUISITE: The length of seqence IDs must be less than 15.
#
# CAUTION: In order to save the memory consumption, this subroutine chews up the input  alignments, so BEWARE !!
#
sub pr_ODEN_input_lite (\@\@$\*) {

	my ($rf_seqids, $rf_align, $dis_len, $fh) = @_;

	(@{$rf_align})  or  (return 0);

	unless (@{$rf_seqids} ) {
		print STDERR "Sequence IDs are missing for \"pr_ODEN_input_lite\"\!\n";
		return 0;
	}
	unless (@{$rf_seqids} == @{$rf_align}) {
		print STDERR "The number of sequences clashes with that of IDs\!\n";
		return 0;
	}

		# REVISED on Friday, July 14, 2006 #
	my $max_id_len = 0;
	foreach my $seqid (@{$rf_seqids}) {
	    my $len = length $seqid;
	    ($max_id_len < $len) and ($max_id_len = $len);
	}
	my $label_width = $max_id_len + 5;

	my @labels = ();
	foreach my $seqid (@{$rf_seqids}) {

		#my $label = $seqid . ' ' x (15 - length($seqid) );
		##my $label = $seqid . ' ' x (16 - length($seqid) ); # A minor revision on Sep 25 (Sat), 2004
		my $label = $seqid . ' ' x ($label_width - length ($seqid) ); # REVISED on Frideay, July 14, 2006 #
		push @labels, $label;
	}

	my $aln_len = length ${$rf_align->[0]};
	my $times = int ($aln_len/$dis_len);
	my $residue = $aln_len % $dis_len;

	#for (my $i=0;  $i<$times; $i++) {	# the outer for #
	foreach my $i (1 .. $times) {	# the outer for #
		#foreach my $rf_aln (@{$rf_align}) {		# the inner for #
		for (my $j=0; $j<@{$rf_align}; $j++) {		# the inner for #

			my $subseq = substr(${$rf_align->[$j]}, 0, $dis_len, '');
			print $fh "$labels[$j]$subseq\n";

		}				# end of the inner for #

		($i == $times) or (print $fh "\n");
	}					# end of the outer for #

	if ($residue) {

		$times and (print $fh "\n");
		for (my $j=0; $j<@{$rf_align}; $j++) {
			 print $fh "$labels[$j]${$rf_align->[$j]}\n"; }
	}

	return 1;
}		# end of the subroutine pr_ODEN_input_lite #


# pr_clustal_like:
#	This subroutine prints out a muliple alignment
#	in a CLUSTAL W -like format WITH A HEADER
#
# PREREQUISITE: The length of seqence IDs must be less than 15.
#
sub pr_clustal_like (\@\@$\$) {

	my ($rf_seqids, $rf_align, $dis_len, $outfile) = @_;

	(@{$rf_align})  or  (return 0);

	unless (@{$rf_seqids} ) {
		print STDERR "Sequence IDs are missing for \'pr_clustal_like\'\!\n";
		return 0;
	}
	unless (@{$rf_seqids} == @{$rf_align}) {
		print STDERR "The number of sequences clashes with that of IDs for \'pr_clustal_like\'\!\n";
		return 0;
	}

	my $id_ind = id_indicator (@{$rf_align}) ;

	my @seqids = (@{$rf_seqids}, '');
	my @align = (@{$rf_align}, $id_ind);

	my $fh;
	open ($fh, "> ${$outfile}" ) or die "\nCannot open ${$outfile}: $!\n";

	print $fh "CLUSTAL W -like alignment format\n\n\n";
	pr_ODEN_input (@seqids, @align, $dis_len, *${$fh}) ;

	close ($fh);

	return 1;

}		# end of the subroutine pr_clustal_like #


# pr_clustal_like_lite: This subroutine prints out a muliple alignment in a CLUSTAL W -like format WITH A HEADER
#
# PREREQUISITE: The length of seqence IDs must be less than 15.
#
# CAUTION: In order to save memory space, this subroutine CHEWS UP the input sequence, so WATCH OUT !!
#
sub pr_clustal_like_lite (\@\@$\$) {

	my ($rf_seqids, $rf_align, $dis_len, $outfile) = @_;

	(@{$rf_align})  or  (return 0);

	unless (@{$rf_seqids} ) {
		print STDERR "Sequence IDs are missing for \'pr_clustal_like_lite\'\!\n";
		return 0;
	}
	unless (@{$rf_seqids} == @{$rf_align}) {
		print STDERR "The number of sequences clashes with that of IDs for \'pr_clustal_like_lite\'\!\n";
		return 0;
	}

	my $id_ind = id_indicator (@{$rf_align}) ;

	my @seqids = (@{$rf_seqids}, '');
	my @align = (@{$rf_align}, $id_ind);

	my $fh;
	open ($fh, "> ${$outfile}" ) or die "\nCannot open ${$outfile}: $!\n";

	print $fh "CLUSTAL W -like alignment format\n\n\n";
	pr_ODEN_input_lite (@seqids, @align, $dis_len, *${$fh}) ;

	close ($fh);

	return 1;

}		# end of the subroutine pr_clustal_like_lite #


# pr_clustal_like_aa: This subroutine prints out a muliple alignment
#		OF POLYPEPTIDE SEQUENCES in a CLUSTAL W -like format WITH A HEADER.
#
# PREREQUISITE: The length of seqence IDs must be less than 15.
#
sub pr_clustal_like_aa (\@\@$\$) {

	my ($rf_seqids, $rf_align, $dis_len, $outfile) = @_;

	(@{$rf_align})  or  (return 0);

	unless (@{$rf_seqids} ) {
		print STDERR "Sequence IDs are missing for \'pr_clustal_like_aa\'\!\n";
		return 0;
	}
	unless (@{$rf_seqids} == @{$rf_align}) {
		print STDERR "The number of sequences clashes with that of IDs for \'pr_clustal_like_aa\'\!\n";
		return 0;
	}

	#my $id_ind = id_indicator (@{$rf_align}) ;
	##my $id_ind = id_indicator_aa (@{$rf_align}) ; # REVISED on Fri, Sep 29, 2006
	my $id_ind = consv_indicator_aa (@{$rf_align}) ;

	my @seqids = (@{$rf_seqids}, '');
	my @align = (@{$rf_align}, $id_ind);

	my $fh;
	open ($fh, "> ${$outfile}" ) or die "\nCannot open ${$outfile}: $!\n";

	print $fh "CLUSTAL W -like alignment format\n\n\n";
	pr_ODEN_input (@seqids, @align, $dis_len, *${$fh}) ;

	close ($fh);

}		# end of the subroutine pr_clustal_like_aa #


# pr_clustal_like_lite_aa: This subroutine prints out a PEPTIDE muliple alignment
#	 in a CLUSTAL W -like format WITH A HEADER.
#
# PREREQUISITE: The length of seqence IDs must be less than 15.
#
# CAUTION: In order to save memory space,
#	this subroutine CHEWS UP the input sequence, so WATCH OUT !!
#
sub pr_clustal_like_lite_aa (\@\@$\$) {

	my ($rf_seqids, $rf_align, $dis_len, $outfile) = @_;

	(@{$rf_align})  or  (return 0);

	unless (@{$rf_seqids} ) {
		print STDERR "Sequence IDs are missing for \'pr_clustal_like_lite_aa\'\!\n";
		return 0;
	}
	unless (@{$rf_seqids} == @{$rf_align}) {
		print STDERR "The number of sequences clashes with that of IDs for \'pr_clustal_like_lite_aa\'\!\n";
		return 0;
	}

	##my $id_ind = id_indicator (@{$rf_align}) ;
	##my $id_ind = id_indicator_aa (@{$rf_align}) ; ## REVISED on Friday, Sep 29, 2006
	my $id_ind = consv_indicator_aa (@{$rf_align}) ; 

	my @seqids = (@{$rf_seqids}, '');
	my @align = (@{$rf_align}, $id_ind);

	my $fh;
	open ($fh, "> ${$outfile}" ) or die "\nCannot open ${$outfile}: $!\n";

	print $fh "CLUSTAL W -like alignment format\n\n\n";
	pr_ODEN_input_lite (@seqids, @align, $dis_len, *${$fh}) ;

	close ($fh);

	return 1;

}		# end of the subroutine pr_clustal_like_lite_aa #


# pr_selected_sites:
#	This program simply print outs the output of "rm_monomorphic" or "rm_in_monomorphic"
#
sub pr_selected_sites (\@\@\*) {

	my ($rf_sel_sites, $rf_list_sites, $fh) = @_;

	unless (@{$rf_sel_sites})  {

		print STDERR "No elements ?\n";
		return 0;
	}

	my $len = length ${$rf_sel_sites->[0]};
	my $list_ct = @{$rf_list_sites};
	unless ($len == $list_ct ) {

		print STDERR "Discrepancy between the length of the string (= $len) and the number of listed site no.s (= $list_ct) !\n";
		return 0;
	} 

	my @sel_sites = ();

	foreach my $rf_seq (@{$rf_sel_sites}) { push @sel_sites, ${$rf_seq}; }

	for (my $l=0; $l< $list_ct; $l++) {

		my $column = '';
		foreach my $seq (@sel_sites) {

			#my $char = substr ($seq, 0, 1, '');
			#$column .= $char;
			$column .= substr ($seq, 0, 1, '');

		}

		print $fh "Site no. = $rf_list_sites->[$l]:\t$column\n";
	}

	return 1;
}			# end of the subroutine pr_selected_sites #


# print_aln_decorated: This subroutine prints outs a block of multiple alignment
#			  along with frame number, residue numbers on the alignment-edge,
#			  and some other information (an array of arrays of the same size).
#
sub print_aln_decorated (\@\@\@\@\@\*) {

	my ($rf_seqIDs, $rf_columns, $rf_resnums, $rf_pos, $rf_info, $fh) = @_;

	my $display_len = 60;
	my $digit_ct = 6;
	my $label_len = 14;

	my $ct_clms = @{$rf_columns};
	$ct_clms or (return 0);
	my $seq_ct = @{$rf_resnums->[0]};
	my $max_pair_ct = $seq_ct * ($seq_ct -1) / 2;

	my @labels = ();
	for (my $i = 0; $i < $seq_ct; $i++) {

		my $seqID = $rf_seqIDs->[$i];
		if (defined $seqID) {
			my $label = $seqID . ' ' x ($label_len -length($seqID) );
			push @labels, $label;
		} else {
			my $label = ' ' x $label_len;
			push @labels, $label;
		}
	}


	my $bl_len = @{$rf_pos};
	my $ds_start = 0;
	#my $bl_end = $rf_pos->[$bl_len -1];

	my $last_pos = $rf_pos->[$ds_start];
	my @ds_seqs = ();
	for (1 .. $seq_ct) { push @ds_seqs, ''; }
	my @ds_info = ();
	for (1 .. scalar(@{$rf_info}) ) { push @ds_info, ''; }
	my $ds_ct = 0;

    for (my $l=0; $l<$bl_len; $l++) { 		# the outer for #

	my $curr_pos = $rf_pos->[$l];
	if ( ($curr_pos > $last_pos + 1) or ($ds_ct >= $display_len) ) { 

		printf $fh "(Frame: %${digit_ct}d --%${digit_ct}d)\n", $rf_pos->[$ds_start] + 1, $rf_pos->[$l-1] + 1;
		for (my $k=0; $k<$seq_ct; $k++) {

			my $start = $rf_resnums->[$ds_start][$k];
			$start = ($start eq '-') ?
				(' ' x ($digit_ct -3) . '-' x 3) : sprintf ( "%${digit_ct}d", $start);
			my $end = $rf_resnums->[$l-1][$k];
			$end = ($end eq '-' ) ?
				(' ' x ($digit_ct -3) . '-' x 3) : sprintf ( "%${digit_ct}d", $end);
			print $fh "$labels[$k]$start  $ds_seqs[$k] $end\n";
			$ds_seqs[$k] = $rf_columns->[$l][$k];
		}

		for (my $i=0; $i<@{$rf_info}; $i++) {

			print $fh ' ' x ($label_len + $digit_ct +2) . $ds_info[$i] . "\n";
			$ds_info[$i] = $rf_info->[$i][$l];
			(defined $ds_info[$i]) or ($ds_info[$i] = ' ');
		}

		print $fh "\n";
		$ds_ct = 1;
		$ds_start = $l;
			
	} else {

		for (my $k=0; $k<$seq_ct; $k++) { $ds_seqs[$k] .= $rf_columns->[$l][$k]; }
		for (my $i=0; $i< @{$rf_info}; $i++) {
			my $info = $rf_info->[$i][$l];
			$ds_info[$i] .= (defined $info) ? $info : ' ';
		}

		$ds_ct++;

	}

	$last_pos = $rf_pos->[$l];
    }						# end of the outer for #


	printf $fh "(Frame: %${digit_ct}d --%${digit_ct}d)\n", $rf_pos->[$ds_start] + 1, $rf_pos->[$bl_len -1] + 1;
	for (my $k=0; $k<$seq_ct; $k++) {
		my $start = $rf_resnums->[$ds_start][$k] ;
		$start = ($start eq '-') ?
			(' ' x ($digit_ct -3) . '-' x 3) : sprintf ( "%${digit_ct}d", $start);
		my $end = $rf_resnums->[$bl_len -1][$k] ;
		$end = ($end eq '-' ) ?
			(' ' x ($digit_ct -3) . '-' x 3) : sprintf ( "%${digit_ct}d", $end);
		print $fh "$labels[$k]$start  $ds_seqs[$k] $end\n";
	}

	for (my $i=0; $i<@{$rf_info}; $i++) {
		print $fh ' ' x ($label_len + $digit_ct +2) . $ds_info[$i] . "\n";
	}

	print $fh "\n";

}			# end of the subroutine print_aln_decorated #


# monomorphic_counter: This subroutine makes a list of monomorphic sites
#			out of a positional distribution of matched pair counts.
#
sub monomorphic_counter (\@$) {

	my ($matched_pair_distri, $ct_seqs) = @_;

	@{$matched_pair_distri} or (return []);
	$ct_seqs or (return []);
	my $max_pair_ct = $ct_seqs * ($ct_seqs -1) / 2;

	my @list_mono = ();
	foreach my $matched_ct (@{$matched_pair_distri}) {

		my $sign = ($matched_ct == $max_pair_ct) ? '*' : ' ';
		push @list_mono, $sign;
	}

	return \@list_mono;

}			# end of the subroutine monomorphic_counter # 


# print_bl_aln_decorated: This subroutine prints out the alignment,
#			 decorated with fringe information, block by block.
#
sub print_bl_aln_decorated (\@\@\@\@\@\*) {

	my ($rf_seqIDs, $bl_columns, $bl_resnums, $bl_frames, $bl_matched_distri, $fh) = @_;

	my $ct_seqs = @{$rf_seqIDs};
	my $ct_blocks = @{$bl_columns};
	$ct_blocks or (return 0);

	for (my $i =0; $i<$ct_blocks; $i++) { 	# the outer for #

		my $frame_ct = @{$bl_frames->[$i]};
		my $block_start = $bl_frames->[$i][0] + 1;
		my $block_end = $bl_frames->[$i][$frame_ct -1] + 1;


		printf $fh "Block_%d (start=$block_start, end=$block_end, $frame_ct columns)\n\n", $i+1;
		my $list_mono = monomorphic_counter (@{$bl_matched_distri->[$i]}, $ct_seqs);
		my @info = ($list_mono, $bl_matched_distri->[$i]);
 		print_aln_decorated (@{$rf_seqIDs}, @{$bl_columns->[$i]},@{$bl_resnums->[$i]}, @{$bl_frames->[$i]}, @info, *${$fh}) ;

		print $fh "\n";
	}					# end of the outer for #

}				# end of the subroutine print_bl_aln_decorated #



1;		# A "spell" necessary for the module to be "use"-able and/or "require"-able #
