package MyLASTPIECE_IO;
#
###################################################################################
#	
#	LASTPIECE_P (Local Alignment-STate Probabilities that Insertion-type and dEletion-type gaps Co-Exist, type P):
#	    A package of programs to compute the probabilities of gapped segments in each of which an insertion-type gap and a deletion-type gap co-exist, (which were referred to as case-(iv) gapped segments by Ezawa (2016a?)),
#	    under a stochastic model of sequence evolution with biologically realistic insertions/deletions.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyLASTPIECE_IO.pm:
#	This Perl package consists of subroutines for data inputs, (test) outputs,
#	and some other processings, that are used in the main (master) and servant scripts
#	of the package, 'LASTPIECE_P'.
#	[CREATION STARTED on 2020/01/27, ENDED on 2020/mm/dd.]

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	); # ADDED 'open_output_file' in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). #
	# list_sessions_ext
	# tfm_labels2indices


use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	print_indel_length_freqs
	print_indel_params
	readin_indel_params
	print_indel_rates
	readin_indel_rates
	print_muP_case_i
	readin_muP_case_i
	print_muP_case_ii_or_iii
	readin_muP_case_ii_or_iii
	print_muP0s
	readin_muP0s
	readin_mu_Adel
	readin_2B_graft_convoluted_1st
	readin_2B_graft_convoluted_2nd
	readin_mu_Adel_1st
	readin_mu_Adel_2nd
	);

sub print_indel_length_freqs  (\$$$\@\@\@\@);
sub print_indel_params  (\$$$$$); # Added on 2020/01/29. #
sub readin_indel_params (\$); # Added on 2020/01/31. #
##sub print_indel_rates  (\$$$$$\@\@\@\@);
sub print_indel_rates  (\$$$$\@\@\@\@); # Modified on 2020/01/29.
sub readin_indel_rates (\$); # Added on 2020/01/31. #
sub print_muP_case_i  (\$$\@\@); # Added on 2020/02/01.
sub readin_muP_case_i (\$$); # Added on 2020/02/08.
sub print_muP_case_ii_or_iii  (\$$\@\@); # Added on 2020/02/01.
sub readin_muP_case_ii_or_iii (\$$$); # Added on 2020/02/10.
sub print_muP0s  (\$$\@$\@); # Added on 2020/02/04.
sub readin_muP0s (\$$$); # Added on 2020/02/05.
sub readin_mu_Adel (\$$$); # Added on 2020/02/08.

##sub readin_2B_graft_convoluted_1st (\$$); # Added on 2020/02/15.
##sub readin_2B_graft_convoluted_2nd (\$$); # Added on 2020/02/15.
sub readin_2B_graft_convoluted_1st (\$$$); # CORRECTED on 2020/02/16.
sub readin_2B_graft_convoluted_2nd (\$$$); # CORRECTED on 2020/02/16.
sub readin_mu_Adel_1st (\$$$); # Added on 2020/02/17.
sub readin_mu_Adel_2nd (\$$$); # Added on 2020/02/20.

	# This specifies the level of tests on the MAIN subroutine, 'xxx'.
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;



	### DEFINITIONS of the Subroutines. ###

# print_indel_length_freqs:
#	This subroutine prints out the length frequencies of insertions/deletions,
#	both in the bulk and at the boundary.
#	It was created by modifying a block in "calc_log_mfacs_pars_pwa_dawg.pl" (at "~/Ezawa_works_KyuTech/Study/New_Alignment_Programs/OpenSource/Exported/TESTs/Test_MyANEX_Main/Test_nonrec_mlt_shifts_gapblocks/InData/").
#	[CREATED on 2020/01/26; CORRECTED on 2020/01/27; MOVED from "../Main_Scripts/lastpiece.alpha.pl" on 2020/01/27; Minor Modification on 2020/01/29]
#
# INPUTS: (\$outfile, $cutoff_ins_len, $cutoff_del_len, \@freqs_ins, \@freqs_del, \@freqs_ins_bd, \@freqs_del_bd), where
#
#	$outfile is the path to the output file;
#	$cutoff_ins/del_len is the cut-off length for the insertions/deletions;
#
#	$freqs_ins/del[$l] is the frequency of the insertion/deletion of length $l (in the bulk);
#	$freqs_ins/del_bd[$l] is the frequency of the insertion/deletion of length $l (on a boundary);
#
# OUTPUT: $flag, where
#
#	$flag	= 1 if the subroutine ended without any particular problem;
#		= 0 if it failed.
#
sub print_indel_length_freqs  (\$$$\@\@\@\@) {

	my ($rf_ofile, $cutoff_ins_len, $cutoff_del_len, $freqs_ins, $freqs_del, $freqs_ins_bd, $freqs_del_bd) = @_;

	my $ofile = ${$rf_ofile};
	my $fh_out = open_output_file ($ofile, 0);
	unless (defined $fh_out) {
	   print STDOUT "[ERROR 'print_indel_length_freqs'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile\": $! !\n";
	   return 0;
	}

   ##print STDOUT "\n< Indel length frequencies >\n\n";
   print STDOUT "# Indel length frequencies:\n\n"; # ADDED on 2020/01/29.

	print $fh_out join ("\t", 'Length', 'Ins', 'Del', 'Ins(bd)', 'Del(bd)'), "\n\n";

	my $cutoff = ($cutoff_ins_len > $cutoff_del_len) ? $cutoff_ins_len : $cutoff_del_len ;

   for (my $l=1; $l<= $cutoff; $l++) {
	my $f_ins = ($l > $cutoff_ins_len) ? 'undef' : $freqs_ins->[$l] ;
	my $f_del = ($l > $cutoff_del_len) ? 'undef' : $freqs_del->[$l] ;
	my $f_ins_bd = ($l > $cutoff_ins_len) ? 'undef' : $freqs_ins_bd->[$l] ;
	my $f_del_bd = ($l > $cutoff_del_len) ? 'undef' : $freqs_del_bd->[$l] ;

	print $fh_out join ("\t", $l, $f_ins, $f_del, $f_ins_bd, $f_del_bd), "\n";
   }
	## print $fh_out "\n\n"; # DISABLED on 2020/01/29. #

	close ($fh_out);

	return 1;
 
}	# End of the subroutine, 'print_indel_length_freqs'. #


# print_indel_params:
#	This subroutine prints out some important parameters of the indel model used.
#	[CREATED on 2020/01/29.]
#
sub print_indel_params  (\$$$$$) {

	my ($rf_ofile, $totrate_ins, $totrate_del, $cutoff_ins_len, $cutoff_del_len) = @_;

	my $ofile = ${$rf_ofile};
	my $fh_out = open_output_file ($ofile, 0);
	unless (defined $fh_out) {
	   print STDOUT "[ERROR 'print_indel_rates'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile\": $! !\n";
	   return 0;
	}

	print $fh_out "# Indel Model Parameters:\n\n";

	print $fh_out join ("\t", 'TOT_RATE_INS', 'TOT_RATE_DEL', 'CUTOFF_INS_LEN', 'CUTOFF_DEL_LEN'), "\n\n";

	print $fh_out join ("\t", $totrate_ins, $totrate_del, $cutoff_ins_len, $cutoff_del_len), "\n";

	## print $fh_out "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_indel_params'. #


# readin_indel_params:
#	This subroutine reads in the indel model parameters
#	printed out by 'print_indel_params'.
#	[CREATED on 2020/01/31.]
#	[DEBUGGED on 2020/02/02]:
#	Added the preliminary while-loop to skip the header.
#
# INPUT: (\$infile), where
#	$infile is the path to the input file.
#
# OUTPUTS: ($totrate_ins, $totrate_del, $cutoff_ins_len, $cutoff_del_len), where
#
#	$totrate_ins and $totrate_del are the total rates of insertions and deletions, respectively;
#	$cutoff_ins_len and $cutoff_del_len are the cutoff lengths for insertions and deletions, respectively.
#
sub readin_indel_params (\$) {

	my ($rf_infile) = @_;
	local $/ = "\n"; # Use a newline character as the input record separater. #

	my $infile = ${$rf_infile};
	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_indel_params'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $! !\n";
	    return (undef, undef, undef, undef);
	}

			# ADDED in DEBUGGING on 2020/02/02. #
	while (my $line = <$fh_in>) { # Preliminary while-loop for skipping the header part. 
	    if ($line =~ m/^TOT_RATE/) { last; }
	} # End of the preliminary while-loop for skipping the header part.
			# End of "ADDED in DEBUGGING on 2020/02/02." #


	my ($totrate_ins, $totrate_del, $cutoff_ins_len, $cutoff_del_len);

	while (my $line = <$fh_in>) { # Main while-loop to read in the parameters.

	    if ($line !~ m/^\S/) { next; } # Skip if the line does NOT begin with a character other than white spaces.
	    ##if ($line =~ m/^\#/) { next; } # Skip comment lines. # Disabled on 2020/02/02.

	    $line =~ s/\n+$//;

	    ($totrate_ins, $totrate_del, $cutoff_ins_len, $cutoff_del_len) = split /\t/, $line;

	    last;

	} # End of the main while-loop to read in the parameters.

	close ($fh_in);

	return ($totrate_ins, $totrate_del, $cutoff_ins_len, $cutoff_del_len);

}	# End of the subroutine, 'readin_indel_params'. #


# print_indel_rates:
#	This subroutine prints out the rates of insertions/deletions,
#	both in the bulk and at the boundary.
#	It was created by slightly modifying 'print_indel_length_freqs'.
#	[CREATED on 2020/01/27; MODIFIED on 2020/01/29]
#
# print_indel_length_freqs:
#	This subroutine prints out the length frequencies of insertions/deletions,
#	both in the bulk and at the boundary.
#	It was created by modifying a block in "calc_log_mfacs_pars_pwa_dawg.pl" (at "~/Ezawa_works_KyuTech/Study/New_Alignment_Programs/OpenSource/Exported/TESTs/Test_MyANEX_Main/Test_nonrec_mlt_shifts_gapblocks/InData/").
#	[CREATED on 2020/01/26; CORRECTED on 2020/01/27; MOVED to from "../Main_Scripts/lastpiece.alpha.pl" on 2020/01/27]
#	[CORRECTED on 2020/01/31]:
#	Replaced $freqs_xxx with $rates_xxx, and $f_xxx with $r_xxx.
#
# INPUTS: (\$outfile, $type, $cutoff_ins_len, $cutoff_del_len, \@rates_ins, \@rates_del, \@rates_ins_bd, \@rates_del_bd), where
#
#	$outfile is the path to the output file;
#	$type indicates the type of the indel rates (either "Raw" or "Dressed");
#	$cutoff_ins/del_len is the cut-off length for the insertions/deletions;
#
#	$rates_ins/del[$l] is the rate (at a specific position) of the insertion/deletion of length $l (in the bulk);
#	$rates_ins/del_bd[$l] is the rate of the insertion/deletion of length $l (on a boundary);
#
# OUTPUT: $flag, where
#
#	$flag	= 1 if the subroutine ended without any particular problem;
#		= 0 if it failed.
#
# [ sub print_indel_rates  (\$$$$$\@\@\@\@) {
#	my ($rf_ofile, $totrate_ins, $totrate_del, $cutoff_ins_len, $cutoff_del_len, $freqs_ins, $freqs_del, $freqs_ins_bd, $freqs_del_bd) = @_; ] ... OBSOLETE as of 2020/01/29.
#
sub print_indel_rates  (\$$$$\@\@\@\@) {

	my ($rf_ofile, $type, $cutoff_ins_len, $cutoff_del_len, $rates_ins, $rates_del, $rates_ins_bd, $rates_del_bd) = @_;

	my $ofile = ${$rf_ofile};
	my $fh_out = open_output_file ($ofile, 0);
	unless (defined $fh_out) {
	   print STDOUT "[ERROR 'print_indel_rates'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile\": $! !\n";
	   return 0;
	}

=OBSOLETE as of 2020/01/29.
	print $fh_out "# TOT_RATE_INS = $totrate_ins\n";
	print $fh_out "# TOT_RATE_DEL = $totrate_del\n";
	print $fh_out "# CUTOFF_INS_LEN = $cutoff_ins_len\n";
	print $fh_out "# CUTOFF_DEL_LEN = $cutoff_del_len\n\n";

	print $fh_out "\n< Length-Dependent Indel Rates >\n\n";
=cut # END of "OBSOLETE as of 2020/01/29."

	print $fh_out "## Length-Dependent Indel Rates ($type):\n\n"; # ADDED on 2020/01/29. #

	print $fh_out join ("\t", 'Length', 'Ins', 'Del', 'Ins(bd)', 'Del(bd)'), "\n\n";

	my $cutoff = ($cutoff_ins_len > $cutoff_del_len) ? $cutoff_ins_len : $cutoff_del_len ;

   for (my $l=1; $l<= $cutoff; $l++) {
	my $r_ins = ($l > $cutoff_ins_len) ? 'undef' : $rates_ins->[$l] ;
	my $r_del = ($l > $cutoff_del_len) ? 'undef' : $rates_del->[$l] ;
	my $r_ins_bd = ($l > $cutoff_ins_len) ? 'undef' : $rates_ins_bd->[$l] ;
	my $r_del_bd = ($l > $cutoff_del_len) ? 'undef' : $rates_del_bd->[$l] ;

	print $fh_out join ("\t", $l, $r_ins, $r_del, $r_ins_bd, $r_del_bd), "\n";
   }
	## print $fh_out "\n\n"; # DISABLED on 2020/01/29. #

	close ($fh_out);

	return 1;
 
}	# End of the subroutine, 'print_indel_rates'. #


# readin_indel_rates:
#	This subroutine reads in indel rates that were output by 'print_indel_rates'.
#	[CREATED on 2020/01/31.]
#
# INPUT: (\$infile), where
#	$infile is the path to the input file.
#
# OUTPUTS: (\@rates_ins, \@rates_del, \@rates_ins_bd, \@rates_del_bd), where
#
#	$rates_ins/del[$l] is the rate (at a specific position) of the insertion/deletion of length $l (in the bulk);
#	$rates_ins/del_bd[$l] is the rate of the insertion/deletion of length $l (on a boundary);
#
sub readin_indel_rates (\$) {

	my ($rf_infile) = @_;

	local $/ = "\n"; # Use a newline character as the input record separater. #

	my $infile = ${$rf_infile};
	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_indel_rates'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $! !\n";
	    return (undef, undef, undef, undef);
	}

	my @rates_ins = my @rates_del = my @rates_ins_bd = my @rates_del_bd = ();

		# Skip the header part. #
	while (my $line = <$fh_in>) { # Preliminary while-loop. #
	    if ($line =~ m/^Length/) { last; }
	} # End of the preliminary while-loop. #

		# Now, read in the input data. #

	while (my $line = <$fh_in>) { # The main while-loop. #

	    if ($line !~ m/^\d/) { next; } # Skip if the line does NOT begin with a decimal digit. #

	    my @fields = split /\t/, $line;
	    my $l = shift @fields;

	    foreach my $rates (\@rates_ins, \@rates_del, \@rates_ins_bd, \@rates_del_bd) {
		my $rate = shift @fields;
		unless ((defined $rate) and ($rate =~ m/\S/)) { next; } # This is just an insurance. (Probably, this will not happen.)
		if ($rate eq 'undef') { next; }
		$rates->[$l] = $rate;
	    }
	} # End of the main while-loop. #

	close ($fh_in);

	return (\@rates_ins, \@rates_del, \@rates_ins_bd, \@rates_del_bd);

}	# End of the subroutine, 'readin_indel_rates'. #


# print_muP_case_i:
#	This subroutine prints out the multiplication factors for a case-(i) gapped segment,
#	which does NOT have any ancestral or descendant sites in between a pair of PASs.
#	[CREATED on 2020/02/01]
#
# INPUTS: (\$ofile, $label_corner, \@labels_clms, \@muP_case_i), where
#
#	$ofile is the path to the output file;
#
#	$label_corner is the label for the corner of the output table;
#
#	$labels_clm[$jt] is the label for the $jt th column (corresponding to the $jt th time-interval);
#
#	$muP_case_i[$jt] is the multiplication factor of the case-(i) gapped segment at the $jt th time-interval (i.e., time-lapse).
#
# OUTPUT: $flag, where
#
#	$flag	= 1 if the subroutine ended without any particular problem;
#		= 0 if it failed.
#
sub print_muP_case_i  (\$$\@\@) {

	my ($rf_ofile, $label_corner, $labels_clms, $muP_case_i) = @_;

=MOVED to below the input check (on 2020/02/03).
	my $ofile = ${$rf_ofile};
	my $fh_out = open_output_file ($ofile, 0);
	unless (defined $fh_out) {
	   print STDOUT "[ERROR 'print_muP_case_i'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile\": $! !\n";
	   return 0;
	}
=cut # End of "MOVED to below the input check (on 2020/02/03)."

	my $ct_tintvls = @{$muP_case_i};

	unless (@{$labels_clms} == $ct_tintvls) {
	     my $ct_labels = @{$labels_clms};
	   print STDOUT "[ERROR 'print_muP_case_i'] ILLEGAL INPUT: The number of labels (=$ct_labels) CLASHES with the number of time-intervals (=$ct_tintvls): $! !\n";
	   return 0;
	}

		# MOVED from above the input check (on 2020/02/03). #
	my $ofile = ${$rf_ofile};
	my $fh_out = open_output_file ($ofile, 0);
	unless (defined $fh_out) {
	   print STDOUT "[ERROR 'print_muP_case_i'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile\": $! !\n";
	   return 0;
	}
		# End of "MOVED from above the input check (on 2020/02/03)." #

	print $fh_out join ("\t", $label_corner, @{$labels_clms}), "\n\n";

	print $fh_out join ("\t", '0:0', @{$muP_case_i}), "\n";

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_muP_case_i'. #


# readin_muP_case_i:
#	This subroutine reads in the multiplication factors for a case-(i) gapped segment,
#	which does NOT have any ancestral or descendant sites in between a pair of PASs.
#	The factors are typically output by the subroutine, 'print_muP_case_i'.
#	[CREATION STARTED on 2020/02/07, ENDED on 2020/02/08.]
#
# INPUTS: (\$infile, $ct_sb_intvls), where
#
#	$infile is the path to the input file;
#	$ct_sb_intvls is the number of sub-intervals into which the whole time-interval is split;
#
# OUTPUTS: (\@muP_case_i), where
#
#	$muP_case_i[$jt] is the multiplication factor of the case-(i) gapped segment at the $jt th time-interval (i.e., time-lapse).
#
sub readin_muP_case_i (\$$) {

	my ($rf_infile, $ct_sb_intvls) = @_;

	local $/ = "\n"; # Use a newline character as the input record separater. #

	my $infile = ${$rf_infile};
	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_muP_case_i'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $! !\n";
	    return (undef);
	}

	my $ct_tlapses = $ct_sb_intvls +1;

	while (my $line = <$fh_in>) {  # Main while-loop (over input lines). #

	    unless ($line =~ m/^0\:0/) { next; }
	    $line =~ s/\s+$//;

	    my @muP_case_i = split /\t/, $line;
	    shift @muP_case_i; # Remove the row label. #

	    my $ct_flds = @muP_case_i;
	    unless ($ct_flds == $ct_tlapses) {
		print STDOUT "[ERROR 'readin_muP_case_i'] The number of input fields (=$ct_flds) CLASHES with the specified number of Time-Lapses (=$ct_tlapses): $! !\n";
		return (undef);
	    }

	    return \@muP_case_i;

	} # End of the main while-loop (over input lines). #

	print STDOUT "[ERROR 'readin_muP_case_i'] Found NO lines starting with '0:0': $! !\n";
	return (undef);

		### RESTART FROM HERE (1) ###

}	# End of the subroutine, 'readin_muP_case_i'. #


# print_muP_case_ii_or_iii:
#	This subroutine prints out the multiplication factors for case-(ii) gapped segments,
#	which have some ancestral sites but no descendant sites in between a pair of PASs,
#	or for case-(iii) gapped segments,
#	which have some descendant sites but no ancestral sites in between a pair of PASs.
#	[CREATED on 2020/02/01]
#
#	[MODIFIED on 2020/02/02]:
#	Now, the subroutine does NOT print out the 0th element of @muP_case_x,
#	which actually corresponds to the case-(i) gapped segment.
#
#	[NOTE ADDED on 2020/02/05]:
#	This subroutine can also print out the multiplication factors,
#	mu_Dcr [(x, [t_1, t_1 + Delta{}t]) ; (L, t_2) ],
#	for the creation of a "D"-region.
#
# INPUTS: (\$ofile, $label_corner, \@labels_clms, \@muP_case_x), where
#
#	$ofile is the path to the output file;
#
#	$label_corner is the label for the corner of the output table;
#
#	$labels_clm[$jt] is the label for the $jt th column (corresponding to the $jt th time-interval);
#
#	$muP_case_x[$jt]->[$l] (w/ $l >0) is the multiplication factor of the case-(ii) or -(iii) gapped segment
#	at the $jt th time-interval with $l sites in between the PASs.
#
# OUTPUT: $flag, where
#
#	$flag	= 1 if the subroutine ended without any particular problem;
#		= 0 if it failed.
#
sub print_muP_case_ii_or_iii  (\$$\@\@) {

	my ($rf_ofile, $label_corner, $labels_clms, $muP_case_x) = @_;

=MOVED to below the input checks (on 2020/02/03). 
	my $ofile = ${$rf_ofile};
	my $fh_out = open_output_file ($ofile, 0);
	unless (defined $fh_out) {
	   print STDOUT "[ERROR 'print_muP_case_ii_or_iii'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile\": $! !\n";
	   return 0;
	}
=cut # END of "MOVED to below the input checks (on 2020/02/03)."

	unless ((@{$muP_case_x}>0) and (@{$muP_case_x->[0]}>1) ) {
	   print STDOUT "[ERROR 'print_muP_case_ii_or_iii'] ILLEGAL INPUT: NO ELEMENTS to be printed out in the 2D array,", '@{$muP_case_x}', ": $! !\n";
	   return 0;
	}

	my $ct_tintvls = @{$muP_case_x};

	unless (@{$labels_clms} == $ct_tintvls) {
	     my $ct_labels = @{$labels_clms};
	   print STDOUT "[ERROR 'print_muP_case_ii_or_iii'] ILLEGAL INPUT: The number of labels (=$ct_labels) CLASHES with the number of time-intervals (=$ct_tintvls): $! !\n";
	   return 0;
	}

		# MOVED from above the input checks (on 2020/02/03). #
	my $ofile = ${$rf_ofile};
	my $fh_out = open_output_file ($ofile, 0);
	unless (defined $fh_out) {
	   print STDOUT "[ERROR 'print_muP_case_ii_or_iii'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile\": $! !\n";
	   return 0;
	}
		# End of "MOVED from above the input checks (on 2020/02/03)." #

	my $cap_ct_sites = @{$muP_case_x->[0]};

	print $fh_out join ("\t", $label_corner, @{$labels_clms}), "\n\n";


	##for (my $l =0; $l < $cap_ct_sites; $l++) { # Main outer for-loop (over the site-counts). #
	for (my $l =1; $l < $cap_ct_sites; $l++) { # Main outer for-loop (over the site-counts). # (MODIFIED on 2020/02/02). #

	    my @out_muPs = ();

	    for (my $jt = 0; $jt < $ct_tintvls; $jt++) { # Inner for-loop (over the time-intervals). #
		$out_muPs[$jt] = $muP_case_x->[$jt][$l];
	    } # End of the inner for-loop (over the time-intervals). #

	    print $fh_out join ("\t", $l, @out_muPs), "\n";

	} # End of the main outer for-loop (over the site-counts). #


	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_muP_case_ii_or_iii'. #


# readin_muP_case_ii_or_iii:
#	This subroutine reads in the multiplication factors for case-(ii) gapped segments,
#	which have some ancestral sites but no descendant sites in between a pair of PASs,
#	or for case-(iii) gapped segments,
#	which have some descendant sites but no ancestral sites in between a pair of PASs.
#	Typically, these factors are printed out via 'print_muP_case_ii_or_iii'.
#	It was created by modifying the predecessor, 'readin_mu_Adel'.
#	[CREATED on 2020/02/10]
#
#	[NOTE ADDED on 2020/02/10]:
#	This subroutine can also read in the multiplication factors,
#	mu_Dcr [(x, [t_1, t_1 + Delta{}t]) ; (L, t_2) ],
#	for the creation of a "D"-region,
#	which are typically printed out via the script, 'cmpt_mu_Dcr.pl',
#	or the multiplication factors for 'A -> {emptyset}':
#	mu_{A->emptyset} [ (L_I, t_I); (0, 0) ; t_2](sigmaBE=2),
#	which are typically printed out via the script, 'cmpt_mu_A2empty.pl'.
#	
# readin_mu_Adel:
#	This subroutine reads in the multiplication factors, mu_{Adel} [ (L_I, t_I) ; (0, - dL_fD, t) ],
#	that were output by the Perl script, 'primt_muP0s'.
#	It was created by modifying the predecessor subroutine, 'readin_muP0s'.
#	[CREATION STARTED on 2020/02/07, ENDED on 2020/02/08.]
#
# readin_muP0s:
#	This subroutine reads in the multiplication factors, mu_{P0} [ (L, t) | (L_I, t_I) ],
#	that were output by 'primt_muP0s'.
#	[CREATED on 2020/02/05.]
#
# INPUT: (\$infile, $ct_sb_intvls, $ub_ct_sites), where
#
#	$infile is the path to the input file;
#
#	$ct_sb_intvls is the number of sub-intervals into which the whole time-interval is split;
#	$ub_ct_sites is the upper-bound (inclusive) of the number of sites in between the PASs
#		up to which the multiplication factors are computed.
#
# OUTPUTS: (\@muP_case_x), where
#	$muP_case_x[$l]->[$jt] (w/ 0< $l <= $ub_ct_sites  & 0 <= $jt <= $ct_sb_intvls)
#	is the multiplication factor of the case-(ii) or -(iii) gapped segment
#	at the $jt th time-interval with $l sites in between the PASs.
#	[NOTE: The order of indexes, $jt and $l, is the OPPOSITE of that for @muP_case_x that is fed into 'print_muP_case_ii_or_iii'.]
#
sub readin_muP_case_ii_or_iii (\$$$) {

	my ($rf_infile, $ct_sb_intvls, $ub_ct_sites) = @_;

	local $/ = "\n"; # Use a newline character as the input record separater. #

	my $infile = ${$rf_infile};
	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_muP_case_ii_or_iii'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $! !\n";
	    return (undef);
	}

	my $cap_ct_sites = $ub_ct_sites +1;
	my $ct_tlapses = $ct_sb_intvls +1;


		### RESTART FROM HERE (2) ###

	    # Initialize the input 2D array, @muP_case_x. #
	my @muP_case_x = ([]);
	for (my $l = 1; $l < $cap_ct_sites; $l++) {
	   $muP_case_x[$l] = [];
	}

	# $muP_case_x[$l]->[$jt] (w/ 0< $l <= $ub_ct_sites  & 0 <= $jt <= $ct_sb_intvls)
	# is the multiplication factor of the case-(ii) or -(iii) gapped segment
	# at the $jt th time-interval with $l sites in between the PASs.

	while (my $line = <$fh_in>) { # Main outer while-loop (over input lines). #

	   my ($l) = ($line =~ m/^(\d+)\t/);
	   unless (defined $l) { next; } # Skip lines without multiplication factors. #

	   unless ((0 < $l) and ($l < $cap_ct_sites)) {
		print STDOUT "[CAUTION 'readin_muP_case_ii_or_iii'] Found a row whose \$l (=$l) is NOT WITHIN the specified range, [1, $ub_ct_sites]: $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

	   $line =~ s/\s+$//;

	   my @fields = split /\t/, $line;
	   my $label = shift @fields; 
	   unless (@fields == $ct_tlapses) {
		my $ct_flds = @fields;
		print STDOUT "[CAUTION 'readin_muP_case_ii_or_iii'] In the row labelled '$label', the number of fields (=$ct_flds) is NOT EQUAL to the specified number of Time-Lapses (=$ct_tlapses): $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

	   $muP_case_x[$l] = \@fields;
	   ##for (my $jt=0; $jt < $ct_tlapses; $jt++) { # Inner for-loop (over data fields). #
	   ##
	   ##	my $muP = $fields[$jt];
	   ##	$muP_case_x[$l]->[$jt] = $muP;
	   ##
	   ##}	# End of the inner for-loop (over data fields). #

	}	# End of the main outer while-loop (over input lines). #

	close ($fh_in);

	return (\@muP_case_x);

}	# End of the subroutine, 'readin_muP_case_ii_or_iii'. #



# print_muP0s:
#	This subroutine prints out the multiplication factors, mu_{P0} [ (L, t) | (L_I, t_I) ],
#	which were computed under the "base" rate operator.
#	It was created by modifying the predecessor, 'print_muP_case_ii_or_iii'.
#	[CREATED on 2020/02/03; DEBUGGED on 2020/02/04]
#
# print_muP_case_ii_or_iii:
#	This subroutine prints out the multiplication factors for case-(ii) gapped segments,
#	which have some ancestral sites but no descendant sites in between a pair of PASs,
#	or for case-(iii) gapped segments,
#	which have some descendant sites but no ancestral sites in between a pair of PASs.
#	[CREATED on 2020/02/01]
#
#	[MODIFIED on 2020/02/02]:
#	Now, the subroutine does NOT print out the 0th element of @muP_case_x,
#	which actually corresponds to the case-(i) gapped segment.
#
# INPUTS: (\$ofile, $label_corner, \@labels_clms, $ub_ct_sites, \@muP0s), where
#
#	$ofile is the path to the output file;
#
#	$label_corner is the label for the corner of the output table;
#
#	$labels_clm[$jt] is the label for the $jt th column (corresponding to the $jt th time-interval);
#
#	$ub_ct_sites is the upper-bound (inclusive) of the number of sites for which the results will be output; # ADDED in creation of 'print_muP0s' (on 2020/02/03). #
#
#	[ $muP_case_x[$jt]->[$l] (w/ $l >0) is the multiplication factor of the case-(ii) or -(iii) gapped segment 
#	at the $jt th time-interval with $l sites in between the PASs. ] ... for the predecessor, 'print_muP_case_ii_or_iii'.
#
#
#	$muP0s[$L_I]->[$jt][$l] (w/ $L_I > 0 & $l >0) is the multiplication factor computed under the "base" rate operator
#	at the $jt th time-interval with $L_I sites initially and $l sites finally, both in between the PASs.    # (MODIFIED in creation of 'print_muP0s' (on 2020/02/03).) #
#
# OUTPUT: $flag, where
#
#	$flag	= 1 if the subroutine ended without any particular problem;
#		= 0 if it failed.
#
sub print_muP0s  (\$$\@$\@) {

	my ($rf_ofile, $label_corner, $labels_clms, $ub_ct_sites, $muP0s) = @_;

	my $ct_tintvls = @{$labels_clms};

	unless (@{$muP0s}>$ub_ct_sites) {
	   my $size_muP0s = @{$muP0s};
	   print STDOUT "[ERROR 'print_muP0s'] ILLEGAL INPUT: The size of ", '@{$muP0s}', "(=$size_muP0s) is NOT LAEGER THAN the specified upper-bound (=$ub_ct_sites): $! !\n";
	   return 0;
	}

		# A BUG fixed on 2020/02/04. #
	##unless (@{$muP0s->[1]} = $ct_tintvls) {
	unless (@{$muP0s->[1]} == $ct_tintvls) {
		# End of "A BUG fixed on 2020/02/04." #

	   my $ct_tintvls2 = @{$muP0s->[1]};
	   print STDOUT "[ERROR 'print_muP0s'] ILLEGAL INPUT: The number of labels (=$ct_tintvls) CLASHES with the number of time-intervals in the output 3D array (=$ct_tintvls2): $! !\n";
	   return 0;
	}

	unless ( @{$muP0s->[1][0]} > $ub_ct_sites ) {
	   my $size_muP0s_1_0 = @{$muP0s->[1][0]};
	   print STDOUT "[ERROR 'print_muP0s'] ILLEGAL INPUT: The size of ", '@{$muP0s->[1][0]}', "(=$size_muP0s_1_0) is NOT LAEGER THAN the specified upper-bound (=$ub_ct_sites): $! !\n";
	   return 0;
	}


	my $ofile = ${$rf_ofile};
	my $fh_out = open_output_file ($ofile, 0);
	unless (defined $fh_out) {
	   print STDOUT "[ERROR 'print_muP0s'] The subroutine 'open_output_file' FAILED to open a file-handle for the file \"$ofile\": $! !\n";
	   return 0;
	}

	print $fh_out join ("\t", $label_corner, @{$labels_clms}), "\n\n";

	my $cap_ct_sites = $ub_ct_sites +1;

	for (my $L_I = 1; $L_I < $cap_ct_sites; $L_I++) { # Main outer for-loop (over the initial site-count).

	    my $muP0s_fxd_L_I = $muP0s->[$L_I];

	    for (my $l =1; $l < $cap_ct_sites; $l++) { # Main middle for-loop (over the final site-counts). # (MODIFIED on 2020/02/02). #

		my @out_muP0s = ();

		for (my $jt = 0; $jt < $ct_tintvls; $jt++) { # Inner for-loop (over the time-intervals). #
		    $out_muP0s[$jt] = $muP0s_fxd_L_I->[$jt][$l];
		} # End of the inner for-loop (over the time-intervals). #

		print $fh_out join ("\t", "$L_I:$l", @out_muP0s), "\n";

	    } # End of the main middle for-loop (over the final site-counts). #

	} # Main outer for-loop (over the initial site-count).

	close ($fh_out);

	return 1;

}	# End of the subroutine, 'print_muP0s'. #


# readin_muP0s:
#	This subroutine reads in the multiplication factors, mu_{P0} [ (L, t) | (L_I, t_I) ],
#	that were output by 'primt_muP0s'.
#	[CREATED on 2020/02/05.]
#
# INPUT: (\$infile, $ct_sb_intvls, $ub_ct_sites), where
#
#	$infile is the path to the input file;
#
#	$ct_sb_intvls is the number of sub-intervals into which the whole time-interval is split;
#	$ub_ct_sites is the upper-bound (inclusive) of the number of sites in between the PASs
#		up to which the multiplication factors are computed.
#
# OUTPUTS: (\@muP0s), where
#	$muP0s[$L_I]->[$jt][$l] (w/ 0 < $L_I, l <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls) is the multiplication factor computed under the "base" rate operator
#	at the $jt th time-lapse with $L_I sites initially and $l sites finally, both in between the PASs.
#
sub readin_muP0s (\$$$) {

	my ($rf_infile, $ct_sb_intvls, $ub_ct_sites) = @_;

	local $/ = "\n"; # Use a newline character as the input record separater. #

	my $infile = ${$rf_infile};
	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_muP0s'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $! !\n";
	    return (undef);
	}

	my $cap_ct_sites = $ub_ct_sites +1;
	my $ct_tlapses = $ct_sb_intvls +1;

	    # Initialize the input 3D array, @muP0s. #
	my @muP0s = ([]);
	for (my $LI = 1; $LI < $cap_ct_sites; $LI++) {
	    my @array_2D = ();
	    for (my $jt = 0; $jt < $ct_tlapses; $jt++) { $array_2D[$jt] = [0]; }
	    $muP0s[$LI] = \@array_2D;
	}

	while (my $line = <$fh_in>) { # Main outer while-loop (over input lines). #

	   my ($LI, $l) = ($line =~ m/^(\d+)\:(\d+)\t/);
	   unless (defined $LI) { next; } # Skip lines without multiplication factors. #

	   unless ((0 < $LI) and ($LI < $cap_ct_sites) and (0 < $l) and ($l < $cap_ct_sites)) {
		print STDOUT "[CAUTION 'readin_muP0s'] Found a row whose (\$LI, \$l) = ($LI, $l) is NOT WITHIN the specified range, [1, $ub_ct_sites] x [1, $ub_ct_sites]: $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

	   ##$line =~ s/\n$//;
	   $line =~ s/\s+$//; # Modified on 2020/02/08. #

	   my @fields = split /\t/, $line;
	   my $label = shift @fields; 
	   unless (@fields == $ct_tlapses) {
		my $ct_flds = @fields;
		print STDOUT "[CAUTION 'readin_muP0s'] In the row labelled '$label', the number of fields (=$ct_flds) is NOT EQUAL to the specified number of Time-Lapses (=$ct_tlapses): $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

	   for (my $jt=0; $jt < $ct_tlapses; $jt++) { # Inner for-loop (over data fields). #

		my $muP0 = $fields[$jt];

		$muP0s[$LI]->[$jt][$l] = $muP0;

	   }	# End of the inner for-loop (over data fields). #

	}	# End of the main outer while-loop (over input lines). #


	close ($fh_in);

	return (\@muP0s);

}	# End of the subroutine, 'readin_muP0s'. #


# readin_mu_Adel:
#	This subroutine reads in the multiplication factors, mu_{Adel} [ (L_I, t_I) ; (0, - dL_fD, t) ],
#	that were output by the Perl script, 'cmpt_mu_Adel.pl'.
#	It was created by modifying the predecessor subroutine, 'readin_muP0s'.
#	[CREATION STARTED on 2020/02/07, ENDED on 2020/02/08.]
#
# readin_muP0s:
#	This subroutine reads in the multiplication factors, mu_{P0} [ (L, t) | (L_I, t_I) ],
#	that were output by 'primt_muP0s'.
#	[CREATED on 2020/02/05.]
#
# INPUT: (\$infile, $ct_sb_intvls, $ub_ct_sites), where
#
#	$infile is the path to the input file;
#
#	$ct_sb_intvls is the number of sub-intervals into which the whole time-interval is split;
#	$ub_ct_sites is the upper-bound (inclusive) of the number of sites in between the PASs
#		up to which the multiplication factors are computed.
#
# OUTPUTS: (\@mu_Adel), where
#	$mu_Adel[$LI]->[$jt][$dL_fD] (w/ 0 < $LI <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls & 0 <= $dL_fD <= {upper-bound specified by the input})
#	is the multiplication factor that an "A"-region (with $LI initial sites) was deleted
#	at the $jt th time-lapse (actually, between $jt-1 and $jt),
#	in conjunction with $dL_fD sites in the flanking "D"-region(s).
#
sub readin_mu_Adel (\$$$) {

	my ($rf_infile, $ct_sb_intvls, $ub_ct_sites) = @_;

	local $/ = "\n"; # Use a newline character as the input record separater. #

	my $infile = ${$rf_infile};
	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_muP0s'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $! !\n";
	    return (undef);
	}

	my $cap_ct_sites = $ub_ct_sites +1;
	my $ct_tlapses = $ct_sb_intvls +1;


		### RESTART FROM HERE (2) ###

	    # Initialize the input 3D array, @mu_Adel. #
	my @mu_Adel = ([]);
	for (my $LI = 1; $LI < $cap_ct_sites; $LI++) {
	    my @array_2D = ();
	    for (my $jt = 0; $jt < $ct_tlapses; $jt++) { $array_2D[$jt] = []; }
	    $mu_Adel[$LI] = \@array_2D;
	}

	# $mu_Adel[$LI]->[$jt][$dL_fD] (w/ 0 < $LI <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls & 0 <= $dL_fD <= {upper-bound specified by the input})
	# is the multiplication factor that an "A"-region (with $LI initial sites) was deleted
	# at the $jt th time-lapse (actually, between $jt-1 and $jt),
	# in conjunction with $dL_fD sites in the flanking "D"-region(s).

	while (my $line = <$fh_in>) { # Main outer while-loop (over input lines). #

	   my ($LI, $dL_fD) = ($line =~ m/^(\d+)\:(\d+)\t/);
	   unless (defined $LI) { next; } # Skip lines without multiplication factors. #

	   unless ((0 < $LI) and ($LI < $cap_ct_sites)) {
		print STDOUT "[CAUTION 'readin_mu_Adel'] Found a row whose \$LI (=$LI) is NOT WITHIN the specified range, [1, $ub_ct_sites]: $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

	   $line =~ s/\s+$//;

	   my @fields = split /\t/, $line;
	   my $label = shift @fields; 
	   unless (@fields == $ct_tlapses) {
		my $ct_flds = @fields;
		print STDOUT "[CAUTION 'readin_mu_Adel'] In the row labelled '$label', the number of fields (=$ct_flds) is NOT EQUAL to the specified number of Time-Lapses (=$ct_tlapses): $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

	   for (my $jt=0; $jt < $ct_tlapses; $jt++) { # Inner for-loop (over data fields). #

		my $mu_Adel = $fields[$jt];

		$mu_Adel[$LI]->[$jt][$dL_fD] = $mu_Adel;

	   }	# End of the inner for-loop (over data fields). #

	}	# End of the main outer while-loop (over input lines). #

	close ($fh_in);

	return (\@mu_Adel);

}	# End of the subroutine, 'readin_mu_Adel'. #


# readin_2B_graft_convoluted_1st:
#	This subroutine reads in the 1st set of multiplication factors,
#	muP_{1st} [ (L_F, t_F)  | (L_I, t_I) ],
#	which will be graft-convoluted.
#	Here, L_I and L_F (both in {1, 2, ..., $ub_ct_sites)) are the numbers of sites
#	at the initial time (=t_I) and the final time (=t_F), respectively.
#	It was created by modifying the predecessor, 'readin_muP0s'.
#	[CREATED on 2020/02/15.]
#	
# readin_muP0s:
#	This subroutine reads in the multiplication factors, mu_{P0} [ (L, t) | (L_I, t_I) ],
#	that were output by 'primt_muP0s'.
#	[CREATED on 2020/02/05.]
#
# INPUT: (\$infile, $ct_sb_intvls, $ub_ct_sites), where
#
#	$infile is the path to the input file;
#
#	$ct_sb_intvls is the number of sub-intervals into which the whole time-interval is split;
#	$ub_ct_sites is the upper-bound (inclusive) of the number of sites in between the PASs
#		up to which the multiplication factors are computed.
#
# OUTPUTS: (\@muPs_1st), where
#	$muPs_1st[$L_I]->[$L_F][$jt] (w/ 0 < $L_I, $L_F <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls)
#	is the multiplication factor, in the 1st set to be graft-convoluted,
#	at the $jt th time-lapse with $L_I sites initially and $L sites finally, both in between the PASs.
#
##sub readin_2B_graft_convoluted_1st (\$$) { # A BUG fixed on 2020/02/16.
sub readin_2B_graft_convoluted_1st (\$$$) {

	my ($rf_infile, $ct_sb_intvls, $ub_ct_sites) = @_;

	local $/ = "\n"; # Use a newline character as the input record separater. #

	my $infile = ${$rf_infile};
	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_2B_graft_convoluted_1st'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $! !\n";
	    return (undef);
	}

	my $cap_ct_sites = $ub_ct_sites +1;
	my $ct_tlapses = $ct_sb_intvls +1;

	    # Initialize the input 3D array, @muPs_1st. #
	my @muPs_1st = ([]);
	for (my $LI = 1; $LI < $cap_ct_sites; $LI++) {
	    my @array_2D = ([]);
	    for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { $array_2D[$LF] = []; }
=For the predecessor, 'readin_muP0s'. (1)
	    my @array_2D = ();
	    for (my $jt = 0; $jt < $ct_tlapses; $jt++) { $array_2D[$jt] = [0]; } # 
=cut # End of "For the predecessor, 'readin_muP0s'. (1)"
	    $muPs_1st[$LI] = \@array_2D;
	}

	while (my $line = <$fh_in>) { # Main outer while-loop (over input lines). #

	   my ($LI, $LF) = ($line =~ m/^(\d+)\:(\d+)\t/);
	   unless (defined $LI) { next; } # Skip lines without multiplication factors. #

	   unless ((0 < $LI) and ($LI < $cap_ct_sites) and (0 < $LF) and ($LF < $cap_ct_sites)) {
		print STDOUT "[CAUTION 'readin_2B_graft_convoluted_1st'] Found a row whose (\$LI, \$LF) = ($LI, $LF) is NOT WITHIN the specified range, [1, $ub_ct_sites] x [1, $ub_ct_sites]: $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

	   ##$line =~ s/\n$//;
	   $line =~ s/\s+$//; # Modified on 2020/02/08. #

	   my @fields = split /\t/, $line;
	   my $label = shift @fields; 
	   unless (@fields == $ct_tlapses) {
		my $ct_flds = @fields;
		print STDOUT "[CAUTION 'readin_2B_graft_convoluted_1st'] In the row labelled '$label', the number of fields (=$ct_flds) is NOT EQUAL to the specified number of Time-Lapses (=$ct_tlapses): $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

	   $muPs_1st[$LI]->[$LF] = \@fields;

=For the predecessor, 'readin_muP0s'. (2)
	   for (my $jt=0; $jt < $ct_tlapses; $jt++) { # Inner for-loop (over data fields). #
		my $muP0 = $fields[$jt];
		$muP0s[$LI]->[$jt][$l] = $muP0;
	   }	# End of the inner for-loop (over data fields). #
=cut # End of "For the predecessor, 'readin_muP0s'. (2)"

	}	# End of the main outer while-loop (over input lines). #


	close ($fh_in);

	return (\@muPs_1st);

}	# End of the subroutine, 'readin_2B_graft_convoluted_1st'. #


# readin_2B_graft_convoluted_2nd:
#	This subroutine reads in the 2nd set of multiplication factors,
#	muP_{2nd} [ (L_F, t_F)  | (L_I, t_I) ],
#	which will be graft-convoluted.
#	Here, L_I and L_F (both in {1, 2, ..., $ub_ct_sites)) are the numbers of sites
#	at the initial time (=t_I) and the final time (=t_F), respectively.
#	It was created by SLIGHTLY modifying the predecessor, 'readin_2B_graft_convoluted_1st'.
#	[CREATED on 2020/02/15.]
#
# readin_2B_graft_convoluted_1st:
#	This subroutine reads in the 1st set of multiplication factors,
#	muP_{1st} [ (L_F, t_F)  | (L_I, t_I) ],
#	which will be graft-convoluted.
#	Here, L_I and L_F (both in {1, 2, ..., $ub_ct_sites)) are the numbers of sites
#	at the initial time (=t_I) and the final time (=t_F), respectively.
#	It was created by modifying the predecessor, 'readin_muP0s'.
#	[CREATED on 2020/02/15.]
#	
# readin_muP0s:
#	This subroutine reads in the multiplication factors, mu_{P0} [ (L, t) | (L_I, t_I) ],
#	that were output by 'primt_muP0s'.
#	[CREATED on 2020/02/05.]
#
# INPUT: (\$infile, $ct_sb_intvls, $ub_ct_sites), where
#
#	$infile is the path to the input file;
#
#	$ct_sb_intvls is the number of sub-intervals into which the whole time-interval is split;
#	$ub_ct_sites is the upper-bound (inclusive) of the number of sites in between the PASs
#		up to which the multiplication factors are computed.
#
# OUTPUTS: (\@muPs_2nd), where
#	[ $muPs_1st[$L_I]->[$L_F][$jt] (w/ 0 < $L_I, $L_F <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls) ] ... for the predecessor,  'readin_2B_graft_convoluted_1st'. #
#	$muPs_2nd[$L_F]->[$L_I][$jt] (w/ 0 < $L_I, $L_F <= $ub_ct_sites, & 0 <= $jt <= $ct_sb_intvls)
#	is the multiplication factor, in the 1st set to be graft-convoluted,
#	at the $jt th time-lapse with $L_I sites initially and $L sites finally, both in between the PASs.
#
##sub readin_2B_graft_convoluted_2nd (\$$) { # A BUG fixed on 2020/02/16.
sub readin_2B_graft_convoluted_2nd (\$$$) {

	my ($rf_infile, $ct_sb_intvls, $ub_ct_sites) = @_;

	local $/ = "\n"; # Use a newline character as the input record separater. #

	my $infile = ${$rf_infile};
	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_2B_graft_convoluted_2nd'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $! !\n";
	    return (undef);
	}

	my $cap_ct_sites = $ub_ct_sites +1;
	my $ct_tlapses = $ct_sb_intvls +1;

	    # Initialize the input 3D array, @muPs_2nd. #
	my @muPs_2nd = ([]);
	for (my $LI = 1; $LI < $cap_ct_sites; $LI++) {
	    my @array_2D = ([]);
	    for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { $array_2D[$LF] = []; }
=For the predecessor, 'readin_muP0s'. (1)
	    my @array_2D = ();
	    for (my $jt = 0; $jt < $ct_tlapses; $jt++) { $array_2D[$jt] = [0]; } # 
=cut # End of "For the predecessor, 'readin_muP0s'. (1)"
	    $muPs_2nd[$LI] = \@array_2D;
	}

	while (my $line = <$fh_in>) { # Main outer while-loop (over input lines). #

	   my ($LI, $LF) = ($line =~ m/^(\d+)\:(\d+)\t/);
	   unless (defined $LI) { next; } # Skip lines without multiplication factors. #

	   unless ((0 < $LI) and ($LI < $cap_ct_sites) and (0 < $LF) and ($LF < $cap_ct_sites)) {
		print STDOUT "[CAUTION 'readin_2B_graft_convoluted_2nd'] Found a row whose (\$LI, \$LF) = ($LI, $LF) is NOT WITHIN the specified range, [1, $ub_ct_sites] x [1, $ub_ct_sites]: $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

	   ##$line =~ s/\n$//;
	   $line =~ s/\s+$//; # Modified on 2020/02/08. #

	   my @fields = split /\t/, $line;
	   my $label = shift @fields; 
	   unless (@fields == $ct_tlapses) {
		my $ct_flds = @fields;
		print STDOUT "[CAUTION 'readin_2B_graft_convoluted_2nd'] In the row labelled '$label', the number of fields (=$ct_flds) is NOT EQUAL to the specified number of Time-Lapses (=$ct_tlapses): $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

	   $muPs_2nd[$LF]->[$LI] = \@fields;
	   ##$muPs_1st[$LI]->[$LF] = \@fields; # For the predecessor, 'readin_2B_graft_convoluted_1st'.

=For the predecessor, 'readin_muP0s'. (2)
	   for (my $jt=0; $jt < $ct_tlapses; $jt++) { # Inner for-loop (over data fields). #
		my $muP0 = $fields[$jt];
		$muP0s[$LI]->[$jt][$l] = $muP0;
	   }	# End of the inner for-loop (over data fields). #
=cut # End of "For the predecessor, 'readin_muP0s'. (2)"

	}	# End of the main outer while-loop (over input lines). #


	close ($fh_in);

	return (\@muPs_2nd);

}	# End of the subroutine, 'readin_2B_graft_convoluted_2nd'. #


# readin_mu_Adel_1st:
#	This subroutine reads in multiplication factors for the deletion of an "A"-region,
#	mu_{Adel} [ (L_I, t_I) ; (0, - dL_fD, t_F) ],
#	that were typically output by the Perl script, 'cmpt_mu_Adel.pl',
#	and outputs the results in the format different from that of 'readin_mu_Adel' above,
#	(and almost identical to that of 'readin_2B_graft_convoluted_1st').
#	It was created by slightly modifying the predecessor, 'readin_2B_graft_convoluted_1st'.
#	[CREATED on 2020/02/17.]
#	
# readin_2B_graft_convoluted_1st:
#	This subroutine reads in the 1st set of multiplication factors,
#	muP_{1st} [ (L_F, t_F)  | (L_I, t_I) ],
#	which will be graft-convoluted.
#	Here, L_I and L_F (both in {1, 2, ..., $ub_ct_sites)) are the numbers of sites
#	at the initial time (=t_I) and the final time (=t_F), respectively.
#	It was created by modifying the predecessor, 'readin_muP0s'.
#	[CREATED on 2020/02/15.]
#	
# readin_muP0s:
#	This subroutine reads in the multiplication factors, mu_{P0} [ (L, t) | (L_I, t_I) ],
#	that were output by 'primt_muP0s'.
#	[CREATED on 2020/02/05.]
#
# INPUT: (\$infile, $ct_sb_intvls, $ub_ct_sites), where
#
#	$infile is the path to the input file;
#
#	$ct_sb_intvls is the number of sub-intervals into which the whole time-interval is split;
#	$ub_ct_sites is the upper-bound (inclusive) of the number of sites in between the PASs
#		up to which the multiplication factors are computed.
#
# OUTPUTS: (\@mu_Adel_1st), where
#	$mu_Adel_1st[$LI]->[$dL_fD][$jt] (w/ 0 < $LI <= $ub_ct_sites, & 0 <= dL_fD <= {upper-bound specified by the input} & 0 <= $jt <= $ct_sb_intvls)
#	is the multiplication factor, mu_{Adel} [ ($LI, t_I) ; (0, - $dL_fD, t_F) ],
#	that an "A"-region (with $LI initial sites) was deleted
#	at the $jt th time-lapse (actually, between $jt-1 and $jt),
#	in conjunction with $dL_fD sites in the flanking "D"-region(s).
#
sub readin_mu_Adel_1st (\$$$) {

	my ($rf_infile, $ct_sb_intvls, $ub_ct_sites) = @_;

	local $/ = "\n"; # Use a newline character as the input record separater. #

	my $infile = ${$rf_infile};
	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_mu_Adel_1st'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $! !\n";
	    return (undef);
	}

	my $cap_ct_sites = $ub_ct_sites +1;
	my $ct_tlapses = $ct_sb_intvls +1;

	    # Initialize the input 3D array, @mu_Adel_1st. #
	my @mu_Adel_1st = ([]);
	for (my $LI = 1; $LI < $cap_ct_sites; $LI++) {
	    my @array_2D = ();
=For the predecessor, 'readin_2B_graft_convoluted_1st'. (1)
	    my @array_2D = ([]);
	    for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { $array_2D[$LF] = []; }
=cut # End of "For the predecessor, 'readin_2B_graft_convoluted_1st'. (1)"
=For the predecessor, 'readin_muP0s'. (1)
	    my @array_2D = ();
	    for (my $jt = 0; $jt < $ct_tlapses; $jt++) { $array_2D[$jt] = [0]; } # 
=cut # End of "For the predecessor, 'readin_muP0s'. (1)"
	    $mu_Adel_1st[$LI] = \@array_2D;
	}

	while (my $line = <$fh_in>) { # Main outer while-loop (over input lines). #

	   ##my ($LI, $LF) = ($line =~ m/^(\d+)\:(\d+)\t/);
	   my ($LI, $dL_fD) = ($line =~ m/^(\d+)\:(\d+)\t/);

	   unless (defined $LI) { next; } # Skip lines without multiplication factors. #

	   unless ((0 < $LI) and ($LI < $cap_ct_sites) and ($dL_fD >= 0)) {
		print STDOUT "[CAUTION 'readin_mu_Adel_1st'] Found a row whose (\$LI, \$dL_fD) = ($LI, $dL_fD) is NOT WITHIN the specified range, [1, $ub_ct_sites] x [0, {input-specified}]: $! ! ... I will SKIP this row ...\n\n";
		next; # ADDED in fixing a BUG (on 2020/02/20). #
	   }
=For the predecessor, 'readin_2B_graft_convoluted_1st'. (2)
	   unless ((0 < $LI) and ($LI < $cap_ct_sites) and (0 < $LF) and ($LF < $cap_ct_sites)) {
		print STDOUT "[CAUTION 'readin_mu_Adel_1st'] Found a row whose (\$LI, \$LF) = ($LI, $LF) is NOT WITHIN the specified range, [1, $ub_ct_sites] x [1, $ub_ct_sites]: $! ! ... I will SKIP this row ...\n\n";
		next;
	   }
=cut # End of "For the predecessor, 'readin_2B_graft_convoluted_1st'. (2)"

	   ##$line =~ s/\n$//;
	   $line =~ s/\s+$//; # Modified on 2020/02/08. #

	   my @fields = split /\t/, $line;
	   my $label = shift @fields; 
	   unless (@fields == $ct_tlapses) {
		my $ct_flds = @fields;
		print STDOUT "[CAUTION 'readin_mu_Adel_1st'] In the row labelled '$label', the number of fields (=$ct_flds) is NOT EQUAL to the specified number of Time-Lapses (=$ct_tlapses): $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

	   $mu_Adel_1st[$LI]->[$dL_fD] = \@fields;

=For the predecessor, 'readin_muP0s'. (2)
	   for (my $jt=0; $jt < $ct_tlapses; $jt++) { # Inner for-loop (over data fields). #
		my $muP0 = $fields[$jt];
		$muP0s[$LI]->[$jt][$l] = $muP0;
	   }	# End of the inner for-loop (over data fields). #
=cut # End of "For the predecessor, 'readin_muP0s'. (2)"

	}	# End of the main outer while-loop (over input lines). #


	close ($fh_in);

	return (\@mu_Adel_1st);

}	# End of the subroutine, 'readin_mu_Adel_1st'. #


# readin_mu_Adel_2nd:
#	This subroutine reads in multiplication factors for the deletion of an "A"-region,
#	mu_{Adel} [ (L_I, t_I) ; (0, - dL_fD, t_F) ],
#	that were typically output by the Perl script, 'cmpt_mu_Adel.pl',
#	and outputs the results in the format different from that of 'readin_mu_Adel' above,
#	(and almost identical to that of 'readin_2B_graft_convoluted_2nd').
#	It was created by somewhat modifying the predecessor, 'readin_mu_Adel_1st'.
#	[CREATED on 2020/02/20.]
#
# readin_mu_Adel_1st:
#	This subroutine reads in multiplication factors for the deletion of an "A"-region,
#	mu_{Adel} [ (L_I, t_I) ; (0, - dL_fD, t_F) ],
#	that were typically output by the Perl script, 'cmpt_mu_Adel.pl',
#	and outputs the results in the format different from that of 'readin_mu_Adel' above,
#	(and almost identical to that of 'readin_2B_graft_convoluted_1st').
#	It was created by slightly modifying the predecessor, 'readin_2B_graft_convoluted_1st'.
#	[CREATED on 2020/02/17.]
#	
# readin_2B_graft_convoluted_1st:
#	This subroutine reads in the 1st set of multiplication factors,
#	muP_{1st} [ (L_F, t_F)  | (L_I, t_I) ],
#	which will be graft-convoluted.
#	Here, L_I and L_F (both in {1, 2, ..., $ub_ct_sites)) are the numbers of sites
#	at the initial time (=t_I) and the final time (=t_F), respectively.
#	It was created by modifying the predecessor, 'readin_muP0s'.
#	[CREATED on 2020/02/15.]
#	
# readin_muP0s:
#	This subroutine reads in the multiplication factors, mu_{P0} [ (L, t) | (L_I, t_I) ],
#	that were output by 'primt_muP0s'.
#	[CREATED on 2020/02/05.]
#
# INPUT: (\$infile, $ct_sb_intvls, $ub_ct_sites), where
#
#	$infile is the path to the input file;
#
#	$ct_sb_intvls is the number of sub-intervals into which the whole time-interval is split;
#	$ub_ct_sites is the upper-bound (inclusive) of the number of sites in between the PASs
#		up to which the multiplication factors are computed.
#
# OUTPUTS: (\@mu_Adel_2nd), where
#
#	[ $mu_Adel_1st[$LI]->[$dL_fD][$jt] (w/ 0 < $LI <= $ub_ct_sites, & 0 <= dL_fD <= {upper-bound specified by the input} & 0 <= $jt <= $ct_sb_intvls) ] ... for the predecessor, 'readin_mu_Adel_1st'.
#
#	$mu_Adel_2nd[$dL_fD]->[$LI][$jt] (w/ 0 < $LI <= $ub_ct_sites, & 0 <= dL_fD <= {upper-bound specified by the input} & 0 <= $jt <= $ct_sb_intvls)
#
#	is the multiplication factor, mu_{Adel} [ ($LI, t_I) ; (0, - $dL_fD, t_F) ],
#	that an "A"-region (with $LI initial sites) was deleted
#	at the $jt th time-lapse (actually, between $jt-1 and $jt),
#	in conjunction with $dL_fD sites in the flanking "D"-region(s).
#
sub readin_mu_Adel_2nd (\$$$) {

	my ($rf_infile, $ct_sb_intvls, $ub_ct_sites) = @_;

	local $/ = "\n"; # Use a newline character as the input record separater. #

	my $infile = ${$rf_infile};
	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_mu_Adel_2nd'] The subroutine, 'open_input_file2', FAILED to open a file-handle for the file \"$infile\": $! !\n";
	    return (undef);
	}

	my $cap_ct_sites = $ub_ct_sites +1;
	my $ct_tlapses = $ct_sb_intvls +1;

	    # Initialize the input 3D array, @mu_Adel_2nd. #
	    # (Modified in creation of 'readin_mu_Adel_2nd'.) (1) #
	my @mu_Adel_2nd = ();
=For the predecessor, 'readin_mu_Adel_1st'. (1) #
	my @mu_Adel_1st = ([]);
	for (my $LI = 1; $LI < $cap_ct_sites; $LI++) {
	    my @array_2D = ();
##=For the predecessor, 'readin_2B_graft_convoluted_1st'. (1)
##	    my @array_2D = ([]);
##	    for (my $LF = 1; $LF < $cap_ct_sites; $LF++) { $array_2D[$LF] = []; }
##=cut # End of "For the predecessor, 'readin_2B_graft_convoluted_1st'. (1)"
##=For the predecessor, 'readin_muP0s'. (1)
##	    my @array_2D = ();
##	    for (my $jt = 0; $jt < $ct_tlapses; $jt++) { $array_2D[$jt] = [0]; } # 
##=cut # End of "For the predecessor, 'readin_muP0s'. (1)"
	    $mu_Adel_1st[$LI] = \@array_2D;
	}
=cut
	    # End of "(Modified in creation of 'readin_mu_Adel_2nd'.) (1)" #


	while (my $line = <$fh_in>) { # Main outer while-loop (over input lines). #

	   ##my ($LI, $LF) = ($line =~ m/^(\d+)\:(\d+)\t/);
	   my ($LI, $dL_fD) = ($line =~ m/^(\d+)\:(\d+)\t/);

	   unless (defined $LI) { next; } # Skip lines without multiplication factors. #

	   unless ((0 < $LI) and ($LI < $cap_ct_sites) and ($dL_fD >= 0)) {
		print STDOUT "[CAUTION 'readin_mu_Adel_2nd'] Found a row whose (\$LI, \$dL_fD) = ($LI, $dL_fD) is NOT WITHIN the specified range, [1, $ub_ct_sites] x [0, {input-specified}]: $! ! ... I will SKIP this row ...\n\n";
		next;
	   }
=For the predecessor, 'readin_2B_graft_convoluted_1st'. (2)
	   unless ((0 < $LI) and ($LI < $cap_ct_sites) and (0 < $LF) and ($LF < $cap_ct_sites)) {
		print STDOUT "[CAUTION 'readin_mu_Adel_1st'] Found a row whose (\$LI, \$LF) = ($LI, $LF) is NOT WITHIN the specified range, [1, $ub_ct_sites] x [1, $ub_ct_sites]: $! ! ... I will SKIP this row ...\n\n";
		next;
	   }
=cut # End of "For the predecessor, 'readin_2B_graft_convoluted_1st'. (2)"

	   ##$line =~ s/\n$//;
	   $line =~ s/\s+$//; # Modified on 2020/02/08. #

	   my @fields = split /\t/, $line;
	   my $label = shift @fields; 
	   unless (@fields == $ct_tlapses) {
		my $ct_flds = @fields;
		print STDOUT "[CAUTION 'readin_mu_Adel_2nd'] In the row labelled '$label', the number of fields (=$ct_flds) is NOT EQUAL to the specified number of Time-Lapses (=$ct_tlapses): $! ! ... I will SKIP this row ...\n\n";
		next;
	   }

			# Modified in creation of 'readin_mu_Adel_2nd'. (2) #
	   unless (defined $mu_Adel_2nd[$dL_fD]) { $mu_Adel_2nd[$dL_fD] = [[]]; }

	   $mu_Adel_2nd[$dL_fD]->[$LI] = \@fields;
	   ##$mu_Adel_1st[$LI]->[$dL_fD] = \@fields; # For the predecessor, 'readin_mu_Adel_1st'. #
			# End of "Modified in creation of 'readin_mu_Adel_2nd'. (2)" #

=For the predecessor, 'readin_muP0s'. (2)
	   for (my $jt=0; $jt < $ct_tlapses; $jt++) { # Inner for-loop (over data fields). #
		my $muP0 = $fields[$jt];
		$muP0s[$LI]->[$jt][$l] = $muP0;
	   }	# End of the inner for-loop (over data fields). #
=cut # End of "For the predecessor, 'readin_muP0s'. (2)"

	}	# End of the main outer while-loop (over input lines). #

	close ($fh_in);

			# Check the consistency of the output. #
			# (Modified in creation of 'readin_mu_Adel_2nd'.) (3) #
	my $cap_dL_fD = @mu_Adel_2nd;
	for (my $dL_fD =0; $dL_fD < $cap_dL_fD; $dL_fD++) {

	    my $mu_Adel_fxd_dL_fD = $mu_Adel_2nd[$dL_fD] ;

	    unless (defined $mu_Adel_fxd_dL_fD) {
		print STDOUT "[ERROR 'readin_mu_Adel_2nd'] For \$dL_fD = $dL_fD, ", '@{$mu_Adel_2nd[$dL_fD]}', " is NOT DEFINED : $! !!\n";
		return undef;
	    }

	    my $size_vec = @{$mu_Adel_fxd_dL_fD} ;

	    unless ($size_vec == $cap_ct_sites) {
		print STDOUT "[ERROR 'readin_mu_Adel_2nd'] For \$dL_fD = $dL_fD, the size of ", '@{$mu_Adel_2nd[$dL_fD]}', " (=$size_vec) CLASHES with \$cap_ct_sites (=$cap_ct_sites) : $! !!\n";
		return undef;
	    }
	}
			# End of "(Modified in creation of 'readin_mu_Adel_2nd'.) (3)" #

	return (\@mu_Adel_2nd);

}	# End of the subroutine, 'readin_mu_Adel_2nd'. #



1;		# a "spell" which makes the module "use"-able and/or "require"-able #


